/******************************************************************
 *
 *  File:   fiona_sniffer.c
 *
 *  Author: Nick Vaccaro <nvaccaro@lab126.com>
 *
 *  Date:   03/29/07
 *
 *  Copyright 2007, Lab126, Inc.  All rights reserved.
 *
 *  Description:
 *      Queue sniffer support routines.  These routines are used to
 *  deal with events in the event queue that entered after the event
 *  processing engine has started processing an event that's going to
 *  put us to sleep.  In this case, we have to examine queue contents
 *  to see if any of the events would have effected device behavior
 *  had we been able to sleep "immediately".  Because we can't sleep
 *  "immediately", we have to treat some events in the event queue
 *  special based on what type of sleep state we were heading to (i.e.
 *  what are the wake sources and are any events that would have been
 *  a wake source in the queue prior to sleep, etc).
 *
 ******************************************************************/
#include  <linux/autoconf.h>

#include <linux/input.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/proc_fs.h>
#include <linux/kbd_kern.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/poll.h>
#include <linux/pm.h>
#include <linux/delay.h>
#include <asm/arch/pxa-regs.h>
#include <asm/delay.h>
#include <asm/io.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fpow.h>
#include <asm/arch/fiona_sniffer.h>
#include <asm/arch/fiona_eventmap.h>
#include <asm/arch/fiona_keymap.h>

//
// Local Prototypes
//
static int __post_sniff_strip_queue_for_power_off(void);
static int __post_sniff_strip_queue_for_power_on(void);
static int  find_remove_events_from_queue(int event);

//
// External Prototypes
//
extern void print_queue(char *str, int forcePrint);
extern void __remove_event_index_from_queue(int index_to_remove);
extern int  __remove_events_in_queue(int event);
extern int __find_event_in_queue(FPOW_EVENT event);
extern int has_new_data_arrived(void);
extern void clear_data_queued_flag(void);
extern void set_data_queued_flag(void);
extern int fiona_event_queue_empty(void);

//
// External Globals
//
extern int g_fiona_sniff_debug;
extern int g_fiona_ext_sniff_debug;
extern struct kbd_queue *queue;    /* Keyboard data buffer. */
extern spinlock_t g_kbd_lock;
extern int qSizeWrap;

asmlinkage int hextlog(const char *fmt, ...)
{
    va_list args;
    int r = 0;
    if (g_fiona_ext_sniff_debug)  {
        va_start(args, fmt);
        r = vprintk(fmt, args);
        va_end(args);
    }
    return r;
}

asmlinkage int hlog(const char *fmt, ...)
{
    va_list args;
    int r = 0;
    if (g_fiona_sniff_debug)  {
        va_start(args, fmt);
        r = vprintk(fmt, args);
        va_end(args);
    }
    return r;
}


int postSniffFixQueue(int state)
{
    int fixed = 0;

    switch(state)  {
        case OFF_STATE_SNIFF:     // Power Switch On in queue
            hextlog("postSniffFixQueue(OFF) calling post_sniff_remove_events_from_queue(ON)\n");
            // remove all but one event
            fixed = post_sniff_remove_events_from_queue(OFF_STATE_SNIFF);
            if (fixed)
                set_remembered_wake_source(POWER_ON_WAKE);
            break;

        case ON_STATE_SNIFF:     // Power Switch Off in queue
            hextlog("postSniffFixQueue(%d) doesn't support ON_STATE_SNIFF yet...\n");
            break;

        case POWER_LOCKOUT_STATE_SNIFF: // Power Lockout
            strip_queue_for_power_lockout();
            hextlog("postSniffFixQueue(POWER_LOCKOUT) set remebered wake source to AC_INSERTED\n");
            set_remembered_wake_source(convert_passed_event_to_wake_src(EVENT_AC_INSERTED));
            break;

        case IDLE_SLEEP_OFF_FOUND:
            hextlog("postSniffFixQueue(IDLE_SLEEP_OFF_FOUND) set remebered wake source to POWER_OFF_WAKE\n");
            set_remembered_wake_source(POWER_OFF_WAKE);
            break;

        default:
            hextlog("postSniffFixQueue(%d) doing nothing for this case.\n",state);
            break;
    }
    return(fixed);
}


// Parms    0 - Fix for Power Switch On
//          2 - Fix for WAN ON
//          3 - Fix for WAN OFF
// Returns  0 - No on's left in q
//          1 - power on in queue
int fixEventQueue(int event)
{
    int onInQ = 0;

    switch(event)  {
        case WAN_SWITCH_ON_EVENT:       // WAN Switch On Event
            hextlog("fixEventQueue() simulating WAN SWITCH ON wake, set remebered wake source to %s)\n",fpow_wake_src_to_str(convert_passed_event_to_wake_src(EVENT_WAN_SWITCH_ON)));
            set_remembered_wake_source(convert_passed_event_to_wake_src(EVENT_WAN_SWITCH_ON));
            break;

        case WAN_SWITCH_OFF_EVENT:      // WAN Switch Off Event
            hextlog("fixEventQueue() simulating WAN SWITCH OFF wake, set remebered wake source to %s)\n",fpow_wake_src_to_str(convert_passed_event_to_wake_src(EVENT_WAN_SWITCH_OFF)));
            set_remembered_wake_source(convert_passed_event_to_wake_src(EVENT_WAN_SWITCH_OFF));
            break;

        case POWER_SWITCH_ON_EVENT:     // Power Switch On
            hextlog("fixEventQueue() calling find_remove_events_from_queue(%s)\n",fpow_passed_event_to_str(event));
            onInQ = find_remove_events_from_queue(EVENT_POWER_SWITCH_ON);
            break;

        case POWER_SWITCH_OFF_EVENT:     // Power Switch Off in event queue, make it the wake source
            hextlog("fixEventQueue() simulating POWER SWITCH OFF wake, set remebered wake source to %s)\n",fpow_wake_src_to_str(convert_passed_event_to_wake_src(EVENT_POWER_SWITCH_OFF)));
            set_remembered_wake_source(convert_passed_event_to_wake_src(EVENT_POWER_SWITCH_OFF));
            break;

        case CRITICAL_BATTERY_EVENT:
            hextlog("fixEventQueue(CRITICAL_BATTERY_EVENT ) simulating AC IN wake and sniff-stripping queue.");
            strip_queue_for_power_lockout();
            hlog("fixEventQueue(CRITICAL_BATTERY_EVENT ) set remebered wake source to %s)\n",fpow_wake_src_to_str(convert_passed_event_to_wake_src(EVENT_AC_INSERTED)));
            set_remembered_wake_source(convert_passed_event_to_wake_src(EVENT_AC_INSERTED));
            break;

        default:
            hextlog("fixEventQueue(%s) doing nothing for this case.\n",fpow_passed_event_to_str(event));
            break;
    }
    return(onInQ);
}


// Returns  0 - No POWER_SWITCH_ON events in queue
//          1 - POWER_SWITCH_ON event in queue
int postSniffFixUnstallQueue(int state)
{
    int onInQ = 0;
    int unstall = 0;

    switch(state) {
        case ON_STATE_SNIFF:
            printk("postSniffFixUnstallQueue(ON)\n");
            if (postSniffFixQueue(OFF_STATE_SNIFF))
            printk("postSniffFixUnstallQueue(ON) not yet supported\n");
            return(0);

        case OFF_STATE_SNIFF:
            printk("postSniffFixUnstallQueue(OFF)\n");
            if (postSniffFixQueue(OFF_STATE_SNIFF))
                unstall = 1;
            else
                printk("postSniffFixUnstallQueue(OFF) not unstalling queue, no ON in queue\n");
            break;

        case POWER_LOCKOUT_STATE_SNIFF:
            printk("postSniffFixUnstallQueue(POWER_LOCKOUT) not yet supported\n");
            return(0);

        case IDLE_SLEEP_STATE_SNIFF:
            printk("postSniffFixUnstallQueue(IDLE_SLEEP_STATE_SNIFF) not yet supported\n");
            return(0);

        case IDLE_SLEEP_OFF_FOUND:
            printk("postSniffFixUnstallQueue(IDLE_SLEEP_OFF_FOUND)\n");
            postSniffFixQueue(IDLE_SLEEP_OFF_FOUND);
            unstall = 1;
            break;
            
        default:
            printk("postSniffFixUnstallQueue(%d) not yet supported\n",state);
            return(0);
    }
    if (unstall) {
        printk("postSniffFixUnstallQueue() unstalling queue()\n");
        unstallEventQueue();
    }
    else
        printk("postSniffFixUnstallQueue() NOT unstalling queue()\n");
    return(onInQ);
}


// Returns  0 - No POWER_SWITCH_ON events in queue
//          1 - POWER_SWITCH_ON event in queue
int fixUnstallEventQueue(int event)
{
    int onInQ = 0;
    onInQ = fixEventQueue(event);

    // If we fixed the queue and there are no on's in the queue, don't unstall it
    hlog("fixUnstallEventQueue(%s)\n",fpow_passed_event_to_str(event));
    switch(event)  {
        case POWER_SWITCH_ON_EVENT:
            if (onInQ == 0)
                unstallEventQueue();
            else
                printk("fixUnstallEventQueue() not unstalling queue, no ON in queue\n");
            break;
        default:
            unstallEventQueue();
            break;
    }
    return(onInQ);
}

/***************************************************************************
 * Routine:     strip_queue_for_power_lockout
 * Purpose:     Parse the event queue, removing problematic events from the
 *              event queue based on having gone into a power lockout.
 *              This should be called in response to waking/seeing an AC_INSERT
 *              event from STATE_V (or while processing STATE_V)
 * Paramters:   none
 * Returns:     none
 * Assumptions: AC adapter has been inserted, so it's OK to clear battery-related
 *              events...
 * Side Effects:    Removed battery events from the queue.
 ***************************************************************************/
int strip_queue_for_power_lockout(void)
{
    int cur = 0;
    int updateCur = 1;
    int result = 0;
    int eventsStripped = 0;
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);

    // print the queue
    print_queue("__strip_queue_for_power_lockout() START QUEUE:\n",0);

    // We processed a power lockout, now clear the queue from battery events, 
    // key/button/scroll activity, etc...
    cur = queue->tail;
    while (cur != queue->head) {
        hlog("__strip_queue_for_power_lockout() checking event %x\n",queue->buf[cur]);

        // If it's an event that doesn't make sense in context, remove it...
        switch(queue->buf[cur])  {
            case EVENT_TIMEOUT_1:
            case EVENT_TIMEOUT_2:
            case EVENT_TIMEOUT_3:
            case EVENT_BATTERY_NO_WAN:
            case EVENT_BATTERY_LOW:
            case EVENT_BATTERY_CRITICAL:
                hlog("__strip_queue_for_power_lockout() found %s, REMOVING...\n",fpow_passed_event_to_str(queue->buf[cur]));
                __remove_event_index_from_queue(cur);
                updateCur = 0;  // don't update cur, we've backed up one...
                break;

            default:
                // If it's a key, button, or scroll wheel event, get rid of it
                result = keycode_to_rawcode(queue->buf[cur]);
                if ((result != -1) && (result < FIONA_KEYMAP_SIZE)) {
                    // It's an IOC KBS event, get rid of it...
                    hlog("__strip_queue_for_power_lockout() found KBS 0x%x, REMOVING...\n",result);
                    __remove_event_index_from_queue(cur);
                    updateCur = 0;  // don't update cur, we've backed up one...
                }
                break;
        }
        if (updateCur)
            cur = ((cur + 1) & qSizeWrap);
        else
            updateCur = 1;
    }

    hlog("strip_queue_for_wan_pairs(): stripped %d events, EXIT QUEUE:\n",eventsStripped);
    print_queue(NULL,0);
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(0);
}


// Returns  Number of events removed
int strip_queue_for_wan_events(int on_count, int off_count)
{
    int cur = 0;
    int updateCur = 1;
    int eventsStripped = 0;
    int items_to_remove = (MIN(on_count,off_count) * 2);
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);

    if ((0 == items_to_remove) || fiona_event_queue_empty())
        goto unlockExit;

    // print the queue
    print_queue("strip_queue_for_wan_events() START QUEUE:\n",0);

    // We have the event in the queue, make sure to remove any
    // events that don't make sense given the context (typing after sleep, etc)
    cur = queue->tail;
    while (items_to_remove && (cur != queue->head)) {
        hlog("strip_queue_for_wan_events() checking event %x\n",queue->buf[cur]);
        // If it's an event that doesn't make sense in context, remove it...
        switch(queue->buf[cur])  {
            case EVENT_WAN_SWITCH_ON:
            case EVENT_WAN_SWITCH_OFF:
                hlog("strip_queue_for_wan_events() found WAN_SWITCH_O%s\n",(queue->buf[cur] == EVENT_WAN_SWITCH_ON)?"N":"FF");
                __remove_event_index_from_queue(cur);
                updateCur = 0;  // don't update cur, we've backed up one...
                eventsStripped++;
                items_to_remove--;
                break;

            default:
                break;
        }
        if (updateCur)
            cur = (cur + 1) & qSizeWrap;
        else
            updateCur = 1;
    }

unlockExit:
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    hlog("strip_queue_for_wan_pairs(): stripped %d events, EXIT QUEUE:\n",eventsStripped);
    print_queue(NULL,0);
    return(eventsStripped);
}


//
// Returns  INVALID_EVENT - NO wake event was found in the event queue
//          WAN_OFF_WAKE - A WAN OFF switch wake event was found in the event queue
//          WAN_ON_WAKE - A WAN OFF switch wake event was found in the event queue
//          ALT_FONT_ - An ALT+FONT was found in the queue
//          3 - An AC Insert event was found
//
int idle_wake_event_in_event_queue(void)
{
    int cur = 0;
    int event = INVALID_EVENT;
    int alt_press_seen = 0;
    int alt_release_seen = 0;
    int font_press_seen = 0;
    int font_release_seen = 0;
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty() || (0 == has_new_data_arrived())) {
        event = INVALID_EVENT;
        goto unlockExit;
    }

    cur = queue->tail;
    while (cur != queue->head) {
        switch(queue->buf[cur])  {
            case EVENT_WAN_SWITCH_OFF:
                hlog("idle_wake_event_in_event_queue() found %s, returning TRUE\n",fpow_passed_event_to_str(queue->buf[cur]));
                event = WAN_OFF_WAKE;
                goto unlockExit;

            case EVENT_WAN_SWITCH_ON:
                hlog("idle_wake_event_in_event_queue() found %s, returning TRUE\n",fpow_passed_event_to_str(queue->buf[cur]));
                event = WAN_ON_WAKE;
                goto unlockExit;

            case ALT_KEY_PRESSED: 
                alt_press_seen = 1;
                break;

            case ALT_KEY_RELEASED: 
                alt_release_seen = 1;
                break;

            case FONT_KEY_PRESSED: 
                font_press_seen = 1;
                break;

            case FONT_KEY_RELEASED: 
                font_release_seen = 1;
                break;

            // Although AC insert is a wake event, it's one that puts the device right back to
            // sleep, so we don't treat it like a "wake event" in this case...
            case EVENT_AC_INSERTED:
            default:
                break;
        }

        // If we found what we're looking for, head out...
        // WARNING: We're going to allow "press only" events to count
        if (alt_press_seen && font_press_seen) {
            event = ALT_FONT_KEY_COMBINATION;
            hlog("idle_wake_event_in_event_queue() found ALT_FONT_KEY_COMBINATION, returning TRUE\n");
            break;
        }
        else
            cur = ((cur + 1) & qSizeWrap);
    }

    // If we didn't find anything this round, clear until more data arrives
    if (event == INVALID_EVENT)
        clear_data_queued_flag();
    else
        hlog("idle_wake_event_in_event_queue() returning %s\n",fpow_wake_src_to_str(event));

unlockExit:
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(event);
}


//
// Purpose  Remove all events from the queue that don't make sense in a wakeup
//          context (i.e. stale events that entered the queue prior to sleep)
//
int __remove_events_for_wake(void)
{
    int cur = queue->tail;
    int head = queue->head;
    int remove_event = 0;
    int events_removed = 0;
    int update = 1;
    int raw = -1;

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty()) {
        hlog("__remove_events_for_wake() found EMPTY QUEUE , returning 0\n");
        return(0);
    }

    // Search the queue, looking for an ALT+FONT event...
    while (cur != head) {
        hlog("__remove_events_for_wake(): Checking stale event %s\n",fpow_passed_event_to_str(queue->buf[cur])); switch(queue->buf[cur])  { case BATTERY_LOW_EVENTCODE:
            case BATTERY_NO_WAN_EVENTCODE:
            case BATTERY_CRITICAL_EVENTCODE:
            case PRIME_PUMP_EVENTCODE:
            case TIMEOUT1_EVENTCODE:
            case TIMEOUT2_EVENTCODE:
            case TIMEOUT3_EVENTCODE:
            case TIMEOUT_USB_WAKE_EVENTCODE:
            case TPH_COMPLETE_EVENTCODE:
            case ALT_KEY_PRESSED: 
            case ALT_KEY_RELEASED: 
            case FONT_KEY_PRESSED: 
            case FONT_KEY_RELEASED: 
                remove_event = 1;
                break;

            case WAN_RING_EVENTCODE:
            case POWER_SW_ON_EVENTCODE:
            case POWER_SW_OFF_EVENTCODE:
            case WATERMARK_CHECK_EVENTCODE:
            default:
                // If it's a key, button, or scroll wheel event, get rid of it
                raw = keycode_to_rawcode(queue->buf[cur]);
                if ((raw != -1) && (raw < FIONA_KEYMAP_SIZE)) {
                    // It's an IOC KBS event, get rid of it...
                    remove_event = 1;
                }
                else
                    remove_event = 0;
                break;
        }

        if (remove_event)  {
            hlog("__remove_events_for_wake() REMOVING stale event %s\n",fpow_passed_event_to_str(queue->buf[cur]));
            __remove_event_index_from_queue(cur);
            // don't update cur, we've deleted one...
            head--;  if (head < 0) head = qSizeWrap;
            update = 0;
            events_removed = 1;
            print_queue(NULL,0);
        }
        if (update)
            cur = ((cur + 1) & qSizeWrap);
        else
            update = 1;
    }


    // If we stripped any data, reset "data arrived" so another search can happen
    if (events_removed == 1)
        set_data_queued_flag();     

    return(events_removed);
}


//
// Purpose  Remove the ALT+FONT key combo from the queue, replace with EVENT_KEY_WAKE
// Returns  0 - Key strokes representing ALT+FONT removed from event queue
//          1 - Couldn't remove event, queue left unaltered...
//
int remove_alt_font_keys_from_event_queue(void)
{
    int cur = 0;
    int remove_event = 0;
    int no_events_removed = 1;
    int update = 1;
    int first = 1;
    unsigned long flags;

    spin_lock_irqsave(&g_kbd_lock, flags);

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty()) {
        hlog("remove_alt_font_keys_from_event_queue() found EMPTY QUEUE , returning 0\n");
        no_events_removed = 0;
        goto unlockExit;
    }

    // Search the queue, looking for an ALT+FONT event...
    cur = queue->tail;
    while (cur != queue->head) {
        switch(queue->buf[cur])  {
            case ALT_KEY_PRESSED: 
            case ALT_KEY_RELEASED: 
            case FONT_KEY_PRESSED: 
            case FONT_KEY_RELEASED: 
                remove_event = 1;
                break;
            default:
                remove_event = 0;
                break;
        }

        if (remove_event)  {
            if (first)  {
                first = 0;
                // Swap the event for EVENT_KEY_WAKE
                queue->buf[cur] = EVENT_KEY_WAKE;
            }
            else {
                __remove_event_index_from_queue(cur);
                // don't update cur, we've deleted one...
                update = 0;
            }
            no_events_removed = 0;
            print_queue(NULL,0);
        }
        if (update)
            cur = ((cur + 1) & qSizeWrap);
        else
            update = 1;
    }

    hlog("remove_alt_font_keys_from_event_queue() returning %d,%sALT+FONT event found in queue\n",no_events_removed,((no_events_removed==1)?" ":" no "));


    // If we stripped any data, reset "data arrived" so another search can happen
    if (no_events_removed == 0)
        set_data_queued_flag();     

unlockExit:
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(no_events_removed);
}



//
// Returns  1 - Power Switch Off is in the queue
//          0 - NO Power Switch off events are in the queue

/***************************************************************************
 * Routine:     __sniff_strip_queue_for_power_off
 * Purpose:     Seach the event queue, removing any WAN switch events
 *              and any paired On/OFF events.
 * Paramters:   none
 * Returns:     0 - No Power Switch Off events in the queue
 *              1 - A Power Switch Off event is in the queue 
 * Assumptions: none
 * Side Effects:    Removes WAN switch events and paired power switch events
 ***************************************************************************/
static int __sniff_strip_queue_for_power_off(void)
{
    int cur = queue->tail;
    int head = queue->head;
    int updateCur = 1;

    int stripped = 0;
    int offInQ = 0;
    int onSeen = 0;
    int offSeen = 0;
    int do_remove = 0;

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty() || (0 == has_new_data_arrived()))
        return(0);

    // Make sure the queue is stalled...
    if (0 == isEventQueueStalled()) {
        hlog("__sniff_strip_queue_for_power_off() found Queue NOT STALLED, aborting\n");
        return(0);
    }

    // print the queue
    hlog("\n");
    hlog("***************************************\n");
    hlog("__sniff_strip_queue_for_power_off() START QUEUE:\n");
    print_queue(NULL,0);
    hlog("***************************************\n");

    // PASS 1 - Strip all WAN_SWITCH events
    //      Since we check WAN switch when we Power On,
    //      and we wouldn't be here if we weren't in a Power Off,
    //      we can remove all WAN switch events, cuz we're guaranteed
    //      to check it after this point when we perform the Powering back up....
    //      Count the relevant events while we're there
    //
    while (cur != head) {
        switch(queue->buf[cur])  {
            case EVENT_POWER_SWITCH_ON:
                onSeen++;
                break;
            case EVENT_POWER_SWITCH_OFF:
                offSeen++;
                break;
            case EVENT_WAN_SWITCH_ON:
            case EVENT_WAN_SWITCH_OFF:
                hlog("__sniff_strip_queue_for_power_off() found %s, REMOVING...\n", fpow_passed_event_to_str(queue->buf[cur]));
                __remove_event_index_from_queue(cur);
                stripped++;
                head--;  if (head < 0) head = qSizeWrap;
                print_queue(NULL,0);
                updateCur = 0;  // don't update cur, we've deleted one...
                break;
            default:
                break;
        }
        if (updateCur)
            cur = ((cur + 1) & qSizeWrap);
        else
            updateCur = 1;
    }

    // print the queue
    hlog("\n");
    hlog("***************************************\n");
    hlog("__sniff_strip_queue_for_power_off() QUEUE after WAN strip:\n");
    print_queue(NULL,0);
    hlog("***************************************\n\n");

    //
    // If we don't have an evenly paired set of OFF/ON events, head back now
    //
    if (!(onSeen && offSeen)) {
        if (offSeen) {
            hlog("__sniff_strip_queue_for_power_off() returning POWER_SWITCH_OFF SEEN\n");
            offInQ = 1;
        }
        goto exit;
    }

    //
    // Pass 2 - If we have an evenly paired set of OFF/ON events, remove them
    //          now in pairs.
    //
    if (offSeen != onSeen) {
        if (offSeen > onSeen) {
            // We'll have at least one left when we're done
            offInQ = 1;
            offSeen = onSeen;
        }
        else
            onSeen = offSeen;
    }

    // Reset cur to point to end of queue again
    cur = queue->tail;
    while (cur != head) {
        switch(queue->buf[cur])  {
            case EVENT_POWER_SWITCH_ON:
                if (onSeen)  {
                    onSeen--;
                    hlog("__sniff_strip_queue_for_power_off() REMOVING POWER_SWITCH_ON/n");
                    do_remove = 1;
                }
                break;
            case EVENT_POWER_SWITCH_OFF:
                if (offSeen)  {
                    offSeen--;
                    hlog("__sniff_strip_queue_for_power_off() REMOVING POWER_SWITCH_OFF/n");
                    do_remove = 1;
                }
                break;

            default:
                do_remove = 0;
                break;
        }
        if (do_remove)  {
            do_remove = 0;
            __remove_event_index_from_queue(cur);
            stripped++;
            head--;  if (head < 0) head = qSizeWrap;
            print_queue(NULL,0);
            updateCur = 0;  // don't update cur, we've deleted one...
        }
        if (updateCur)
            cur = ((cur + 1) & qSizeWrap);
        else
            updateCur = 1;
    }

exit:

    // print the queue
    hlog("***************************************\n");
    print_queue("__sniff_strip_queue_for_power_off() EXIT QUEUE:\n",0);
    hlog("***************************************\n\n");

    // If we didn't strip any data, clear "data arrived" so we don't recheck unless
    // new data arrives....
    if (stripped == 0)
        clear_data_queued_flag();

    return(offInQ);
}


//
// Returns  1 - EVEN number of paired events found and removed this pass
//          0 - NO events stripped found this pass
static int __post_sniff_strip_queue_for_power_on(void)
{
    int onInQ = 0;

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty())
        return(0);

    // Make sure the queue is stalled...
    if (0 == isEventQueueStalled()) {
        hlog("__post_sniff_strip_queue_for_power_on() found Queue NOT STALLED, aborting\n");
        return(0);
    }

    onInQ = __find_event_in_queue(POWER_SWITCH_ON_EVENT);

    // print the queue
    hlog("\n");
    hlog("***************************************\n");
    hlog("__post_sniff_strip_queue_for_power_on() END QUEUE:\n");
    print_queue(NULL,0);
    hlog("***************************************\n");

    return(onInQ);
}


//
// Returns  1 - A power off is in the queue
//          0 - NO Power Switch Off event in the queue
static int __post_sniff_strip_queue_for_power_off(void)
{
    int offInQ = 0;

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty())
        return(0);

    // Make sure the queue is stalled...
    if (0 == isEventQueueStalled()) {
        hlog("__post_sniff_strip_queue_for_power_off() found Queue NOT STALLED, aborting\n");
        return(0);
    }

    offInQ = __find_event_in_queue(POWER_SWITCH_OFF_EVENT);

    // print the queue
    hlog("\n");
    hlog("***************************************\n");
    hlog("__post_sniff_strip_queue_for_power_on() END QUEUE:\n");
    print_queue(NULL,0);
    hlog("***************************************\n");

    return(offInQ);
}


//
// Returns  1 - EVEN number of paired events found and removed this pass
//          0 - NO events stripped found this pass
static int __strip_queue_for_power_on(int *onInQueue, int *wakeInQ, int *totalStripped)
{
    int cur = queue->tail;
    int updateCur = 1;
    int eventsStripped = 0;
    int nonEventsStripped = 0;
    int onSeen = 0;
    int onInQ = 0;
    int offInQ = 0;
    int onIndex = 0;
    int offIndex = 0;

    if ((NULL == onInQueue) || (NULL == wakeInQ) || (NULL == totalStripped))
        hlog("ERROR: __strip_queue_for_power_on() passed NULL!\n");

    *wakeInQ = INVALID_EVENT;
    *onInQueue = 0;
    *totalStripped = 0;

    // If there's no (new) data to process, bail out...
    if (fiona_event_queue_empty() || (0 == has_new_data_arrived()))
        return(0);

    // Make sure the queue is stalled...
    if (0 == isEventQueueStalled()) {
        hlog("__strip_queue_for_power_on() found Queue NOT STALLED, aborting\n");
        return(0);
    }

    // print the queue
    hlog("\n");
    hlog("***************************************\n");
    hlog("__strip_queue_for_power_on() START QUEUE:\n");
    print_queue(NULL,0);
    hlog("***************************************\n");

    // PASS 1 - Strip all WAN_SWITCH events
    //      Since we check WAN switch when we Power On,
    //      and we wouldn't be here if we weren't in a Power Off,
    //      we can remove all WAN switch events, cuz we're guaranteed
    //      to check it after this point when we perform the Powering back up....
    //
    while (cur != queue->head) {
        //hextlog("__strip_queue_for_power_on() checking event %s\n",fpow_passed_event_to_str(queue->buf[cur]));
        switch(queue->buf[cur])  {
            case EVENT_WAN_SWITCH_ON:
            case EVENT_WAN_SWITCH_OFF:
                hlog("__strip_queue_for_power_on() found %s, REMOVING...\n", fpow_passed_event_to_str(queue->buf[cur]));
                __remove_event_index_from_queue(cur);
                print_queue(NULL,0);
                updateCur = 0;  // don't update cur, we've deleted one...
                eventsStripped++;
                break;
            default:
                break;
        }
        if (updateCur)
            cur = ((cur + 1) & qSizeWrap);
        else
            updateCur = 1;
    }

    // print the queue
    hlog("\n");
    hlog("***************************************\n");
    hlog("__strip_queue_for_power_on() QUEUE after WAN strip:\n");
    print_queue(NULL,0);
    hlog("***************************************\n\n");

    // PASS 1 - Strip all WAN_SWITCH events
    // Reset cur to point to end of queue again
    cur = queue->tail;
                 
    //
    // Pass 2 - Strip all events between any POWER_OFF and associated POWER_ON.
    //
    // We've seen and passed up a Power Off already, or the Q wouldn't be stalled.
    // The start of the queue represents all events AFTER the POWER_OFF was received.
    //
    // We want to strip any event that happen between the start of the queue and the next
    //  power on event, as well as all relevant events between POWER_OFF and it's associated
    //  POWER_ON.
    //
    while (cur != queue->head) {

        if (onInQ == 1)  {
            // If we have offInQ && onInQ, then we have a matching set in queue that
            // need to get removed...
            if (offInQ)  {
                hextlog("__strip_queue_for_power_on() found matching ON/OFF's, REMOVING index %d and %d...\n",onIndex,offIndex);
                // print the queue
                hlog("\n");
                hlog("***************************************\n");
                hlog("__strip_queue_for_power_on() stripping ON/OFF pair at %d,%d:\n",onIndex,offIndex);
                print_queue(NULL,0);
                hlog("***************************************\n");


                // Adjust for the fact that after deleting the POWER_SWITCH_ON,
                // the index of the POWER_SWITCH_OFF changes....
                __remove_event_index_from_queue(onIndex);            // Remove POWER_SWITCH_ON
                offIndex--;
                if (offIndex<0)
                    offIndex = qSizeWrap;
                __remove_event_index_from_queue(offIndex);           // Remove POWER_SWITCH_OFF

                // print the queue
                hlog("\n");
                hlog("***************************************\n");
                hlog("__strip_queue_for_power_on() REMOVED ON/OFF pair\n");
                print_queue(NULL,0);
                hlog("***************************************\n\n");

                cur = offIndex;
                updateCur = 0;      // don't update cur, we've deleted one
                eventsStripped++;

                // On and OFF no longer in queue
                onInQ = offInQ = 0;
                onSeen = 0;
            }
            else  {
                // If we've seen an on, continue parsing until we see an off event...
                hextlog("__strip_queue_for_power_on() ON, NO OFF seen checking event %s (cur=%d, h=%d)\n",fpow_passed_event_to_str(queue->buf[cur]),cur,(int) queue->head);
                switch(queue->buf[cur]) {
                    case EVENT_POWER_SWITCH_OFF: 
                        hextlog("__strip_queue_for_power_on() found matching POWER_SWITCH_OFF at index %d (on at %d)\n",cur,onIndex);
                        // We've seen an ON and now have the matching OFF, 
                        offIndex = cur;
                        offInQ = 1;
                        updateCur = 0;  // don't update cur in case we're at the end of queue.
                                        // instead, we'll strip the pair next loop and fix cur then...
                        break;
                    case EVENT_POWER_SWITCH_ON: 
                        // A second ON was found without an OFF in between....  Nuke it...
                        hlog("__strip_queue_for_power_on() found DUPLICATE POWER_SWITCH_ON at index %d, REMOVING...\n",cur);
                        __remove_event_index_from_queue(cur);
                        print_queue(NULL,0);
                        updateCur = 0;  // don't update cur, we've deleted one...
                        eventsStripped++;
                        break;
                    default:
                        break; 
                }
            }
        }
        else  {
            hextlog("__strip_queue_for_power_on() checking event %s in OFF seen code...\n",fpow_passed_event_to_str(queue->buf[cur]));
            switch(queue->buf[cur])  {
                // Strip all events until the next Power On
                case EVENT_WAN_RING:
                case EVENT_KEY_WAKE:
                case EVENT_WAN_SWITCH_ON:
                case EVENT_WAN_SWITCH_OFF:
                case EVENT_USB_PLUG_INSERTED:
                case EVENT_USB_PLUG_REMOVED:
                case EVENT_BATTERY_NO_WAN:
                case EVENT_BATTERY_LOW:
                case EVENT_BATTERY_CRITICAL:
                // If we see another OFF without seeing the ON, get rid of it
                case EVENT_POWER_SWITCH_OFF:
                    hlog("__strip_queue_for_power_on() found %s, REMOVING...\n", fpow_passed_event_to_str(queue->buf[cur]));
                    print_queue(NULL,0);
                    __remove_event_index_from_queue(cur);
                    print_queue(NULL,0);
                    updateCur = 0;  // don't update cur, we've deleted one...
                    eventsStripped++;
                    break;

                // Don't delete these cases, they're relevant
                case EVENT_AC_REMOVED:
                case EVENT_AC_INSERTED:
                case EVENT_HEADPHONES_INSERTED:
                case EVENT_HEADPHONES_REMOVED:
                case EVENT_AUDIO_ON:
                case EVENT_AUDIO_OFF:
                case EVENT_WAN_SWITCH_SWEEPER:
                case EVENT_TWO_FINGER_SWEEPER:
                case EVENT_TPH_COMPLETE:
                    break;

                case EVENT_POWER_SWITCH_ON:
                    onSeen = 1;
                    onIndex = cur;
                    onInQ = 1;
                    hextlog("__strip_queue_for_power_on() saw POWER_ON at index %d\n",onIndex);
                    break;

                case EVENT_TIMEOUT_1:
                case EVENT_TIMEOUT_2:
                case EVENT_TIMEOUT_3:
                case EVENT_TIMEOUT_USB_WAKE:
                case EVENT_USB_WAKE:
                case EVENT_VALID_PASSWORD:
                case EVENT_INVALID_PASSWORD:
                case EVENT_PRIME_PUMP:
                default:
                    // Must be a character, remove it....
                    hlog("__strip_queue_for_power_on() removing irrelevant event %x\n",queue->buf[cur]);
                    __remove_event_index_from_queue(cur);
                    updateCur = 0;  // don't update cur, we've backed up one...
                    nonEventsStripped++;
                    break;
            }
        }

        if (updateCur)
            cur = (cur + 1) & qSizeWrap;
        else
            updateCur = 1;
    } // End while

    // print the queue
    hlog("***************************************\n");
    print_queue("__strip_queue_for_power_on() EXIT QUEUE:\n",0);
    hlog("***************************************\n\n");

    // Return whether there's currently an ON in the queue
    if (NULL != onInQueue) {
        hlog("__strip_queue_for_power_on() setting onInQueue to %d\n",onSeen);
        *onInQueue = onInQ;
        if (onSeen && (NULL != wakeInQ)) {
            hlog("__strip_queue_for_power_on() setting wakeEventInQ to EVENT_POWER_SWITCH_ON\n");
            *wakeInQ = EVENT_POWER_SWITCH_ON;
        }
    }

    // Pass back total number of events seen
    if (NULL != totalStripped)
        *totalStripped = (eventsStripped+nonEventsStripped);

    // If we didn't strip any data, clear "data arrived" so we don't recheck unless
    // new data arrives....
    if (eventsStripped == 0)
        clear_data_queued_flag();

    return(eventsStripped?1:0);
}


int remove_paired_wan_events_in_queue(void)
{
    unsigned long flags;
    int found = 0;
    int first = 0;
    int cur = 0;
    int done = 0;
    int onIndex = 0, offIndex = 0;
    int onFound = 0, offFound = 0;

    spin_lock_irqsave(&g_kbd_lock, flags);

    // Nothing to search if queue is empty
    if (fiona_event_queue_empty()) 
        goto exit;

    // Print starting queue
    print_queue("remove_paired_wan_events_in_queue() starting queue:",0);
    cur = queue->tail;
    do {
        if (queue->buf[cur] == EVENT_WAN_SWITCH_ON) {
            if (first == 0) {
                first = 2;      // 2 = WAN ON
                hlog("remove_paired_wan_events_in_queue() found WAN_ON first...\n");
            }
            else
                hlog("remove_paired_wan_events_in_queue() found WAN_ON at %d\n",cur);
            onIndex = cur;
            onFound = 1;
        }
        else if (queue->buf[cur] == EVENT_WAN_SWITCH_OFF) {
            if (first == 0) {
                first = 3;      // 3 = WAN OFF
                hlog("remove_paired_wan_events_in_queue() found WAN_OFF first at %d\n",cur);
            }
            else
                hlog("remove_paired_wan_events_in_queue() found WAN_OFF at %d",cur);
            offIndex = cur;
            offFound = 1;
        }
        if (onFound && offFound) {
            // Adjust for the fact that after deleting the first one,
            // the index of the second one changes....
            if (first == 3)  {  // WAN Off is first...
                onIndex--;
                if (onIndex<0)
                    onIndex = qSizeWrap;
            }
            else {
                offIndex--;
                if (offIndex<0)
                    offIndex = qSizeWrap;
            }
            hlog("remove_paired_wan_events_in_queue() removing event %s\n",fpow_passed_event_to_str(queue->buf[offIndex]));
            __remove_event_index_from_queue(offIndex);
            hlog("remove_paired_wan_events_in_queue() removing event %s\n",fpow_passed_event_to_str(queue->buf[onIndex]));
            __remove_event_index_from_queue(onIndex);
            done = 1;
            found = 1;
        }
        else
            cur = (cur + 1) & qSizeWrap;

        if (cur == queue->head)
            done = 1;
    } while (0 == done);

exit:
    print_queue("remove_paired_wan_events_in_queue() ending queue:",0);
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(found);
}



int
sniff_remove_events_from_queue(int state)
{
    unsigned long flags;
    int stripped = 0;
    int wakeInQ = INVALID_EVENT;
    int eventInQ = 0;
    int totalStripped = 0;

    // Range-check the event, make sure it's valid
    // It's zero based, so this works
    spin_lock_irqsave(&g_kbd_lock, flags);

    // Right now, we only allow stripping for POWER_SWITCH_ON_EVENT
    switch(state) {
        case OFF_STATE_SNIFF:
            hlog("sniff_remove_events_from_queue(POWER_SWITCH_ON) calling __strip_queue_for_power_on()\n");
            stripped = __strip_queue_for_power_on(&eventInQ, &wakeInQ, &totalStripped);
            if (wakeInQ != INVALID_EVENT) {
                hlog("sniff_remove_events_from_queue() found wake event %s (0x%x)\n",fpow_passed_event_to_str(wakeInQ),wakeInQ);
                hlog("sniff_remove_events_from_queue() found onInQ = %d\n",eventInQ);
            }
            break;

        case ON_STATE_SNIFF:
            hlog("sniff_remove_events_from_queue(POWER_SWITCH_OFF) calling __strip_queue_for_power_off()\n");
            eventInQ = __sniff_strip_queue_for_power_off();
            if (eventInQ) hlog("sniff_remove_events_from_queue(POWER_SWITCH_OFF) found OFF left in queue()\n");
            break;

        case POWER_LOCKOUT_STATE_SNIFF:
            eventInQ = find_event_in_queue(AC_PLUG_INSERTED_EVENT);
            //stripped = __remove_events_in_queue(AC_PLUG_INSERTED_EVENT);
            break;

        case IDLE_SLEEP_STATE_SNIFF:
        default:
            hlog("ERROR: sniff_remove_events_from_queue(0x%x) only works with ON,OFF,POWER_LOCKOUT\n",state);
            break;
    }
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(eventInQ);
}



// Leave at least one wake event for state
int
post_sniff_remove_events_from_queue(FPOW_SNIFF_STATES state)
{
    unsigned long flags;
    int stripped = 0;
    int eventInQ = 0;

    // Range-check the event, make sure it's valid
    // It's zero based, so this works
    spin_lock_irqsave(&g_kbd_lock, flags);

    // Right now, we only allow stripping for POWER_SWITCH_ON_EVENT
    switch(state) {
        case OFF_STATE_SNIFF:
            hlog("post_sniff_remove_events_from_queue(POWER_SWITCH_ON) calling __strip_queue_for_power_on()\n");
            eventInQ = __post_sniff_strip_queue_for_power_on();
            if (eventInQ != INVALID_EVENT) {
                hlog("post_sniff_remove_events_from_queue() found onInQ = %d\n",eventInQ);
            }
            break;

        case ON_STATE_SNIFF: 
            hlog("post_sniff_remove_events_from_queue(POWER_SWITCH_OFF) calling __strip_queue_for_power_off()\n");
            eventInQ = __post_sniff_strip_queue_for_power_off();
            if (eventInQ) hlog("post_sniff_remove_events_from_queue(POWER_SWITCH_OFF) found OFF left in queue()\n");
            break;

        case IDLE_SLEEP_STATE_SNIFF:
            break;

        case POWER_LOCKOUT_STATE_SNIFF:
            stripped = __remove_events_in_queue(AC_PLUG_INSERTED_EVENT);
            break;

        case WAKE_FROM_SLEEP_STATE_SNIFF:
            hlog("post_sniff_remove_events_from_queue(WAKE_FROM_SLEEP) checking for stale events...\n");
            __remove_events_for_wake();
            break;

        default:
            hlog("ERROR: post_sniff_remove_events_from_queue(0x%x) - state not supported\n",state);
            break;
    }
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(eventInQ);
}



static int
find_remove_events_from_queue(int event)
{
    unsigned long flags;
    int stripped = 0;
    int wakeInQ = INVALID_EVENT;
    int eventInQ = 0;
    int totalStripped = 0;

    // Range-check the event, make sure it's valid
    // It's zero based, so this works
    spin_lock_irqsave(&g_kbd_lock, flags);

    // Right now, we only allow stripping for POWER_SWITCH_ON_EVENT
    switch(event) {
        case EVENT_POWER_SWITCH_ON: 
            hlog("find_remove_events_from_queue(POWER_SWITCH_ON) calling __strip_queue_for_power_on()\n");
            stripped = __strip_queue_for_power_on(&eventInQ, &wakeInQ, &totalStripped);
            if (wakeInQ != INVALID_EVENT) {
                hlog("find_remove_events_from_queue() found wake event %s (0x%x)\n",fpow_passed_event_to_str(wakeInQ),wakeInQ);
                hlog("find_remove_events_from_queue() found onInQ = %d\n",eventInQ);
            }
            break;

        case EVENT_POWER_SWITCH_OFF: 
            hlog("find_remove_events_from_queue(POWER_SWITCH_OFF) calling __strip_queue_for_power_off()\n");
            eventInQ = __sniff_strip_queue_for_power_off();
            if (eventInQ) hlog("find_remove_events_from_queue(POWER_SWITCH_OFF) found OFF left in queue()\n");
            break;

        case EVENT_KEY_WAKE:
            break;

        case EVENT_AC_INSERTED:
            stripped = __remove_events_in_queue(event);
            break;

        case EVENT_WAN_SWITCH:
            break;

        default:
            hlog("ERROR: find_remove_events_from_queue(0x%x) only works with POWER_ON and AC_INSERTED\n",event);
            break;
    }
    spin_unlock_irqrestore(&g_kbd_lock, flags);
    return(eventInQ);
}

