/*
 * IOC PNLCD component support for the
 * Fiona Power Management System (fpow)
 * 
 * Copyright (C) 2006 Lab126, Inc.  All rights reserved. <nvaccaro@lab126.com>
 */

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/completion.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/syscalls.h>
#include <asm/uaccess.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/suspend.h>
#include <asm/system.h>
#include <asm/arch/system.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fpow.h>
#include <linux/ioc.h>

#ifdef CONFIG_FIONA_PM_PNLCD

extern FPOW_POWER_MODE  g_ioc_power_mode;
extern int IOC_Enable_PNLCD(unsigned short enable);

// Holds current PNLCD power mode status
FPOW_POWER_MODE  g_pnlcd_power_mode = FPOW_MODE_ON;

// Holds fpow device entry returned from fpow_register()
struct fpow_component *g_pnlcd_fpow_component_ptr = NULL;



/*********************************************************
 * Routine:     pnlcd_fpow_setmode
 *
 * Purpose:     Handled Fiona Power Management callbacks
 *              for doing IOC PNLCD power management.
 *
 * Paramters:   private - value passed at registration - ignored
 *              state   - ignored
 *              mode    - Power mode to switch to
 *                        Currently supported Modes:
 *                          FPOW_MODE_ON
 *                          FPOW_MODE_OFF
 *
 * Returns:     
 *  FPOW_NO_ERR                     - Transitioned to mode with no errors.
 *  -FPOW_NOT_ENOUGH_POWER_ERR      - not enough power to transition
 *  -FPOW_COMPONENT_TRANSITION_ERR  - Failed to transition to mode
 *  -EINVAL                         - Invalid mode
 *
 * Assumptions: none
 *********************************************************/

FPOW_ERR
pnlcd_fpow_setmode(void *private, u32 state, u32 mode)
{
	int	retVal = FPOW_NO_ERR;

   	powdebug("pnlcd_fpow_setmode(%s)\n",fpow_mode_to_str(mode));

	// Switch to the appropriate power level
	if (mode == g_pnlcd_power_mode)  {
        powdebug("pnlcd_fpow_setmode() found mode same as current (%s)\n",fpow_mode_to_str(mode));
		return(retVal);
    }

	// Perform the power transition
	switch (mode)  {
		case FPOW_MODE_ON:
			if (IOC_Enable_PNLCD(ENABLE) == 0)
				g_pnlcd_power_mode = mode;
            else powdebug("IOC_Enable_PNLCD(ENABLE)  returned error !!\n");
			break;

		case FPOW_MODE_OFF:
			if (IOC_Enable_PNLCD(DISABLE) == 0)
				g_pnlcd_power_mode = mode;
            else powdebug("IOC_Enable_PNLCD(DISABLE)  returned error !!\n");
			break;

		default:
            powdebug("pnlcd_fpow_setmode() hit default mode !!\n");
			retVal = -EINVAL;
			break;
	}

	powdebug("PNLCD: driver now in %s mode (retVal=%d)\n",fpow_mode_to_str(g_pnlcd_power_mode),retVal);
    return(retVal);
}

static int
pnlcd_fpow_getmode(void *private)
{
    extpowdebug("pnlcd_fpow_getmode(0x%x) returning %s\n",(int)private,fpow_mode_to_str(g_pnlcd_power_mode));
	return (g_pnlcd_power_mode);
}

int 
pnlcd_register_with_fpow(struct device *dev, struct fpow_component **fpow_dev)
{
    struct fpow_registration_rec   reg;
    int rc = 0;

	powdebug("pnlcd_register_with_fpow(0x%x, 0x%x)\n",(int)dev,(int)fpow_dev);
    memclr(&reg,sizeof(reg));

    strcpy(reg.name,"pnlcd");
    reg.device_class = FPOW_CLASS_VIDEO_PNLCD;
    reg.supported_modes = ( FPOW_MODE_ON_SUPPORTED | \
                            FPOW_MODE_OFF_SUPPORTED);
    reg.private = (void*) dev;
    reg.getmode = pnlcd_fpow_getmode;
    reg.setmode = pnlcd_fpow_setmode;
    *fpow_dev = fpow_register(&reg);
    if (*fpow_dev == NULL) {
        printk("PNLCD: Failed to register with fpow !!\n");
        rc = -1;
    }
    return(rc);
}

void
pnlcd_unregister_with_fpow(struct fpow_component *fpow_dev)
{
	fpow_unregister(fpow_dev);
}

int
pnlcd_awake(void)
{
    switch(pnlcd_fpow_getmode(NULL))  {
        case FPOW_MODE_ON:
            return(1);
        case FPOW_MODE_SLEEP:
        case FPOW_MODE_OFF:
        default:
            return(0);
    }
}

#endif
