/*
 *	Watchdog driver for the SA11x0/PXA2xx
 *
 *      (c) Copyright 2000 Oleg Drokin <green@crimea.edu>
 *          Based on SoftDog driver by Alan Cox <alan@redhat.com>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Neither Oleg Drokin nor iXcelerator.com admit liability nor provide
 *	warranty for any of this software. This material is provided
 *	"AS-IS" and at no charge.
 *
 *	(c) Copyright 2000           Oleg Drokin <green@crimea.edu>
 *
 *      27/11/2000 Initial release
 * 
 *
 *  Lab126 changes copyright (C) 2007 Lab126, Inc.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/init.h>

#ifdef CONFIG_ARCH_PXA
#include <asm/arch/pxa-regs.h>
#endif

#include <asm/hardware.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>

#if defined(CONFIG_ARCH_LAB126) && defined(CONFIG_PM)
#include <linux/suspend.h>
#endif

#define OSCR_FREQ		CLOCK_TICK_RATE

static unsigned long sa1100wdt_users;
static int pre_margin;
static int boot_status;

#if defined(CONFIG_ARCH_LAB126) && defined(CONFIG_PM)
static wait_queue_head_t pm_event_wq;
static pid_t pm_handler_thread_pid = 0;
static DECLARE_COMPLETION(pm_handler_thread_exited);

static void init_wdt(void)
{
	OSMR3 = OSCR + pre_margin;
	OSSR = OSSR_M3;
	OWER = OWER_WME;
	OIER |= OIER_E3;
}

static int pm_handler_thread(void *unused)
{
	daemonize("wdtpmd");
	allow_signal(SIGKILL);

	while (1) {
		wait_event_interruptible(pm_event_wq, 0);

		if (current->flags & PF_FREEZE) {
			refrigerator(PF_FREEZE);

			init_wdt();
		}

        if (signal_pending(current)) {
            break;
        }
	}

    complete_and_exit(&pm_handler_thread_exited, 0);

	return 0;
}

static void start_pm_thread(void)
{
	init_waitqueue_head(&pm_event_wq);

	pm_handler_thread_pid = kernel_thread(pm_handler_thread, NULL, CLONE_KERNEL);
}

static void
stop_pm_thread(
    void)
{
    int ret;

    if (pm_handler_thread_pid > 0) {
        ret = kill_proc(pm_handler_thread_pid, SIGKILL, 1);
        if (ret == 0) {
            wait_for_completion(&pm_handler_thread_exited);
        }
    }
}


static wait_queue_head_t	d_state_event_wq;
static pid_t			d_state_thread_pid = 0;
static DECLARE_COMPLETION(d_state_thread_exited);

/* This thread is intended to help detect processes that are stuck in an
   uninterruptible 'D' state.  It wakes up once a second and checks if there
   are any processes in the 'D' state.  If it finds processes in this state
   for "limit" consecutive sampling periods, it dumps the processes that are
   in the 'D' state at that moment.  With a "limit" value of 20, for example,
   it is unlikely that anything other than a stuck process would cause a dump
   to occur.  If a process remains stuck, this code will continue to dump
   every "limit" samples.  The work that this thread does is very lightweight,
   so it currently samples once a second, but this could be stretched, if
   necessary.
 */

static int d_state_thread(void *unused)
{
	unsigned long	limit	    = 20;
	unsigned long	show_all    = 0;  // Set nonzero for more verbosity
	unsigned long	iteration   = 0;
	unsigned long 	count	    = 0;
	unsigned long	nr_d;
	
	daemonize("dstated");
	allow_signal(SIGKILL);

	while (1) {
		wait_event_interruptible_timeout(d_state_event_wq, 0, 1 * HZ);

		if (current->flags & PF_FREEZE) {
			refrigerator(PF_FREEZE);
		}

		if (signal_pending(current)) {
			break;
		}

		iteration++;
		
		/* Get total number of uninterruptible 'D' state processes */
		
		nr_d = nr_uninterruptible();

		/* If no processes are uninterruptible, reset counter */

		if (nr_d == 0) {
			count = 0;
			continue;
		}

		/* At least one process is uninterruptible, decide whether or
		   not to dump process information.  If there are
		   uninterruptible processes for 'limit'  consecutive samples,
		   dump the information.
		   We reset the counter after dumping, so if processes remain
		   in this state for another 'limit' samples, they will be
		   dumped again.
		 */

		if (show_all) {
			printk(KERN_INFO "dstated: %ld  (%ld%s)\n",
					nr_d, iteration,
					count ? "" : "*");
		}

		count++;
		
		if (count == limit) {
			if (!show_all)
				printk(KERN_INFO "dstated: %ld process%s in "
					"'D' state  (timestamp: %ld)\n", nr_d,
					(nr_d > 1 ? "es" : ""), iteration);
			show_state();
			count = 0;
		}
	}

	complete_and_exit(&d_state_thread_exited, 0);

	return 0;
}

static void start_d_state_thread(void)
{
	init_waitqueue_head(&d_state_event_wq);

	d_state_thread_pid = kernel_thread(d_state_thread, NULL, CLONE_KERNEL);
}

static void stop_d_state_thread(void)
{
	int ret;
	
	if (d_state_thread_pid > 0) {
		ret = kill_proc(d_state_thread_pid, SIGKILL, 1);
		if (ret == 0) {
			wait_for_completion(&d_state_thread_exited);
		}
	}
}
#endif

/*
 *	Allow only one person to hold it open
 */
static int sa1100dog_open(struct inode *inode, struct file *file)
{
	nonseekable_open(inode, file);
	if (test_and_set_bit(1,&sa1100wdt_users))
		return -EBUSY;

	/* Activate SA1100 Watchdog timer */
#if defined(CONFIG_ARCH_LAB126) && defined(CONFIG_PM)
	init_wdt();
#else
	OSMR3 = OSCR + pre_margin;
	OSSR = OSSR_M3;
	OWER = OWER_WME;
	OIER |= OIER_E3;
#endif

	return 0;
}

/*
 * The watchdog cannot be disabled.
 *
 * Previous comments suggested that turning off the interrupt by
 * clearing OIER[E3] would prevent the watchdog timing out but this
 * does not appear to be true (at least on the PXA255).
 */
static int sa1100dog_release(struct inode *inode, struct file *file)
{
	printk(KERN_CRIT "WATCHDOG: Device closed - timer will not stop\n");

	clear_bit(1, &sa1100wdt_users);

	return 0;
}

static ssize_t sa1100dog_write(struct file *file, const char __user *data, size_t len, loff_t *ppos)
{
	if (len)
		/* Refresh OSMR3 timer. */
		OSMR3 = OSCR + pre_margin;

	return len;
}

static struct watchdog_info ident = {
	.options	= WDIOF_CARDRESET | WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
	.identity	= "SA1100/PXA255 Watchdog",
};

static int sa1100dog_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	int ret = -ENOIOCTLCMD;
	int time;
	void __user *argp = (void __user *)arg;
	int __user *p = argp;

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		ret = copy_to_user(argp, &ident,
				   sizeof(ident)) ? -EFAULT : 0;
		break;

	case WDIOC_GETSTATUS:
		ret = put_user(0, p);
		break;

	case WDIOC_GETBOOTSTATUS:
		ret = put_user(boot_status, p);
		break;

	case WDIOC_SETTIMEOUT:
		ret = get_user(time, p);
		if (ret)
			break;

		if (time <= 0 || time > 255) {
			ret = -EINVAL;
			break;
		}

		pre_margin = OSCR_FREQ * time;
		OSMR3 = OSCR + pre_margin;
		/*fall through*/

	case WDIOC_GETTIMEOUT:
		ret = put_user(pre_margin / OSCR_FREQ, p);
		break;

	case WDIOC_KEEPALIVE:
		OSMR3 = OSCR + pre_margin;
		ret = 0;
		break;
	}
	return ret;
}

static const struct file_operations sa1100dog_fops =
{
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= sa1100dog_write,
	.ioctl		= sa1100dog_ioctl,
	.open		= sa1100dog_open,
	.release	= sa1100dog_release,
};

static struct miscdevice sa1100dog_miscdev =
{
	.minor		= WATCHDOG_MINOR,
	.name		= "watchdog",
	.fops		= (struct file_operations *)&sa1100dog_fops,
};

static int margin __initdata = 60;		/* (secs) Default is 1 minute */

static int __init sa1100dog_init(void)
{
	int ret;

#if defined(CONFIG_ARCH_LAB126) && defined(CONFIG_PM)
	start_pm_thread();
	start_d_state_thread();
#endif	

	/*
	 * Read the reset status, and save it for later.  If
	 * we suspend, RCSR will be cleared, and the watchdog
	 * reset reason will be lost.
	 */
	boot_status = (RCSR & RCSR_WDR) ? WDIOF_CARDRESET : 0;
	pre_margin = OSCR_FREQ * margin;

	ret = misc_register(&sa1100dog_miscdev);
	if (ret == 0)
		printk("SA1100/PXA2xx Watchdog Timer: timer margin %d sec\n",
		       margin);
	return ret;
}

static void __exit sa1100dog_exit(void)
{
#if defined(CONFIG_ARCH_LAB126) && defined(CONFIG_PM)
	stop_pm_thread();
	stop_d_state_thread();
#endif  
    
    misc_deregister(&sa1100dog_miscdev);
}

module_init(sa1100dog_init);
module_exit(sa1100dog_exit);

MODULE_AUTHOR("Oleg Drokin <green@crimea.edu>");
MODULE_DESCRIPTION("SA1100/PXA2xx Watchdog");

module_param(margin, int, 0);
MODULE_PARM_DESC(margin, "Watchdog margin in seconds (default 60s)");

MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
