
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/version.h>
#include <linux/sched.h>

#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/hardware.h>
#include <asm/io.h>

#include "apollo.h"




#define do_debug dummy_print 

static int dummy_print(char *fmt, ...) {
  return 0;
}

static char* apollo_base_map = NULL;

static reg_addr_t reg_GPSR1 = NULL;
static reg_addr_t reg_GPSR2 = NULL;

static reg_addr_t reg_GPCR1 = NULL;
static reg_addr_t reg_GPCR2 = NULL;

static reg_addr_t reg_GPLR1 = NULL;
static reg_addr_t reg_GPLR2 = NULL;

static reg_addr_t reg_GPDR1 = NULL;
static reg_addr_t reg_GPDR2 = NULL;


volatile int ack_received = 0;
wait_queue_head_t apollo_ack_wq;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
static void eink_irq_hdl(int irq, void *data, struct pt_regs* regs) {
#else
irqreturn_t eink_irq_hdl(int irq, void *data, struct pt_regs* regs) {
#endif
  wake_up(&apollo_ack_wq);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,23)
	return IRQ_HANDLED;
#endif
}


void apollo_init_irq(void) {
  int rqstatus;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
  set_irq_type(IRQ_GPIO(83), IRQT_FALLING);
#else
  set_GPIO_IRQ_edge(83, GPIO_FALLING_EDGE);
  printk("SET FALLING EDGE 2.4 style done\n");
#endif

  rqstatus = request_irq(IRQ_GPIO(83), eink_irq_hdl, SA_INTERRUPT, "einkfb", NULL);
  printk("rqstatus %d irq %d\n", rqstatus, IRQ_GPIO(83));

  disable_irq(IRQ_GPIO(83)); /* initially disable irq.  we will use it 
				only for longer waits to avoid prolonged 
				spinning */
  printk("DISABLE_IRQ done\n");
}

void apollo_shutdown_irq(void) {
  free_irq(IRQ_GPIO(83), NULL);
}



int apollo_comm_init(void) {
  ack_received = 0;

  init_waitqueue_head(&apollo_ack_wq);

  apollo_base_map = ioremap_nocache(BASE_ADDR, MAP_SIZE);
  reg_GPSR1 = (reg_addr_t) (apollo_base_map + (GPSR1_ADDR & MAP_MASK));
  reg_GPSR2 = (reg_addr_t) (apollo_base_map + (GPSR2_ADDR & MAP_MASK));

  reg_GPCR1 = (reg_addr_t) (apollo_base_map + (GPCR1_ADDR & MAP_MASK));
  reg_GPCR2 = (reg_addr_t) (apollo_base_map + (GPCR2_ADDR & MAP_MASK));

  reg_GPLR1 = (reg_addr_t) (apollo_base_map + (GPLR1_ADDR & MAP_MASK));
  reg_GPLR2 = (reg_addr_t) (apollo_base_map + (GPLR2_ADDR & MAP_MASK));

  reg_GPDR1 = (reg_addr_t) (apollo_base_map + (GPDR1_ADDR & MAP_MASK));
  reg_GPDR2 = (reg_addr_t) (apollo_base_map + (GPDR2_ADDR & MAP_MASK));

  apollo_init_irq();

  return 0;
}

int apollo_comm_shutdown(void) {
  apollo_shutdown_irq();

  iounmap(apollo_base_map);
  apollo_base_map = NULL;
  reg_GPSR1 = NULL;
  reg_GPSR2 = NULL;

  reg_GPCR1 = NULL;
  reg_GPCR2 = NULL;

  reg_GPLR1 = NULL;
  reg_GPLR2 = NULL;

  reg_GPDR1 = NULL;
  reg_GPDR2 = NULL;
 

  return 0;
}

#define MAX_TIMEOUT (HZ * 4)
#define FAST_TIMEOUT 1000


inline int wait_for_ack_irq(void) {
  enable_irq(IRQ_GPIO(83));
  wait_event_interruptible_timeout(apollo_ack_wq, ((apollo_read_reg(reg_GPLR2) & (1 << 19)) == 0), MAX_TIMEOUT);
  disable_irq(IRQ_GPIO(83));
  return 1;
}


inline int wait_for_ack(void) {
  register int i=0;

  /* wait for nACK to go low (GPLR2_83) */

  for(i = 0; i < FAST_TIMEOUT; i++) {
      if((apollo_read_reg(reg_GPLR2) & (1 << 19)) == 0) {
	return (1);
      } 
  }

  /*  do_debug("going on to longer wait i %d\n", i); */

  return wait_for_ack_irq();
}

void send_command(unsigned char data)
{
 
  if(apollo_base_map == NULL) return;
  do_debug("send command: 0x%x\n",data);

  apollo_write_reg(reg_GPSR2, 1 << 3);
  apollo_write_reg(reg_GPSR1, data << 26);
  apollo_write_reg(reg_GPSR2, data >> 6);
  apollo_write_reg(reg_GPCR1, (0x000000ff & ~data) << 26);
  apollo_write_reg(reg_GPCR2, (0x000000ff & ~data) >> 6);
  /* now give a DS */
  apollo_write_reg(reg_GPCR2, 1 << 2);
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
  /* set C/nD back to D*/
  apollo_write_reg(reg_GPCR2, 1 << 3);
  return;
}


inline void send_data(register unsigned char data){
  register unsigned int gpsr1, gpsr2, gpcr1, gpcr2;

  gpsr1 = data << 26;
  gpsr2 = data >> 6;
  gpcr1 = (0x000000ff & ~data) << 26;
  gpcr2 = (0x000000ff & ~data) >> 6;


  /* first setup data on LDD7:0 */
  apollo_write_reg(reg_GPSR1, gpsr1);
  apollo_write_reg(reg_GPSR2, gpsr2);
  apollo_write_reg(reg_GPCR1, gpcr1);
  apollo_write_reg(reg_GPCR2, gpcr2);
  /* now give a DS */
  apollo_write_reg(reg_GPCR2, 1 << 2);
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
  return;
}

inline void preack_init(register unsigned char data) {
  register unsigned int gpsr1, gpsr2, gpcr1, gpcr2;

  gpsr1 = data << 26;
  gpsr2 = data >> 6;
  gpcr1 = (0x000000ff & ~data) << 26;
  gpcr2 = (0x000000ff & ~data) >> 6;


  /* first setup data on LDD7:0 */
  apollo_write_reg(reg_GPSR1, gpsr1);
  apollo_write_reg(reg_GPSR2, gpsr2);
  apollo_write_reg(reg_GPCR1, gpcr1);
  apollo_write_reg(reg_GPCR2, gpcr2);
  /* now give a DS */
  apollo_write_reg(reg_GPCR2, 1 << 2);
}


inline void preack_end(void) {
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
}

inline void send_data_preack(register unsigned char data) {
  register unsigned int gpsr1, gpsr2, gpcr1, gpcr2;
  gpsr1 = data << 26;
  gpsr2 = data >> 6;
  gpcr1 = (0x000000ff & ~data) << 26;
  gpcr2 = (0x000000ff & ~data) >> 6;
  wait_for_ack();
  apollo_write_reg(reg_GPSR2, 1 << 2);
  apollo_write_reg(reg_GPSR1, gpsr1);
  apollo_write_reg(reg_GPSR2, gpsr2);
  apollo_write_reg(reg_GPCR1, gpcr1);
  apollo_write_reg(reg_GPCR2, gpcr2);
  apollo_write_reg(reg_GPCR2, 1 << 2);
  return;
}


void data2highz(void)
{
  u32 r1, r2;

  if(apollo_base_map == NULL) return;
  /* The Data bus is mapped on GPIO 58 - 65
  *  IO 58 = bit 26 of register 1.
  *  IO 63 = bit 32 (MSB of register 1.)
  *  IO 64 = bit 0 of register 2.
  *  The GPIO can be set to input by writting 0 in the GPDRx register. */

  r1 = apollo_read_reg(reg_GPDR1);
  r2 = apollo_read_reg(reg_GPDR2);

  apollo_write_reg(reg_GPDR1, (r1 & 0x03ffffff)); /*clear upper 6 bits*/
  apollo_write_reg(reg_GPDR2, (r2 & 0xfffffffc)); /*clear lower 2 bits*/  
}

void data2out(void)
{
  u32 r1, r2;
  if(apollo_base_map == NULL) return;
  /*   The Data bus is mapped on GPIO 58 - 65
  *    IO 58 = bit 26 of register 1.
  *    IO 63 = bit 32 (MSB of register 1.)
  *    IO 64 = bit 0 of register 2.
  *    The GPIO can be set to output by writting 1 in the GPDRx register.  */

  r1 = apollo_read_reg(reg_GPDR1);
  r2 = apollo_read_reg(reg_GPDR2);

  apollo_write_reg(reg_GPDR1, (r1 | 0xfc000000)); /*set upper 6 bits*/
  apollo_write_reg(reg_GPDR2, (r2 | 0x00000003)); /*set lower 2 bits*/  
}


u32 apollo_read(void)
{
  /* define local variable*/
  u32 temp1, temp2;
  
  if(apollo_base_map == NULL) return 0;

  data2highz();
  /* Set RW high = GPIO69 = GPSR2 bit 5 */
  apollo_write_reg(reg_GPSR2, 1 << 5);
  /* Set CD low = GPIO67 = GPCR2 bit 3 */
  apollo_write_reg(reg_GPCR2, 1 << 3);

  /* now give a DS (to zero) = GPIO66 = GPCR2 bit 2*/
  apollo_write_reg(reg_GPCR2, 1 << 2);
  wait_for_ack();
  /* read data*/
  
  /*read upper 6 bits*/
  temp1 = (apollo_read_reg(reg_GPLR1) & 0xFC000000)  >> 26;
  
  /*read lower 2 bits*/
  temp2 = (apollo_read_reg(reg_GPLR2) & 0x00000003)  << 6;
  temp2 = temp2 | temp1;    
  
  /* set DS back high = GPIO66 = GPSR2 bit 2*/
  apollo_write_reg(reg_GPSR2, 1 << 2);
  /* Set RW low = GPIO69 = GPCR2 bit 5 */
  apollo_write_reg(reg_GPCR2, 1 << 5);
  data2out();
  return temp2;
}


u32 apollo_get_status() {
  send_command(APOLLO_GET_STATUS_CMD);
  return apollo_read();
}

u32 apollo_erase_screen(unsigned char color) {
  u32 status;
  unsigned long start_time = jiffies;
  send_command(APOLLO_ERASE_SCRN_CMD);
  send_data(color);
  status = apollo_get_status();
  do_debug("status 0x%x time %d\n", status, jiffies - start_time);
  return status;
}


u32 apollo_load_image_data(unsigned char *buffer, int bufsize) {
  int i;

  send_command(APOLLO_LOAD_IMAGE_CMD);
  preack_init(buffer[0]);

  for(i = 1; i < bufsize; i++) {
    send_data_preack(buffer[i]);
  }

  preack_end();

  send_command(APOLLO_STOP_IMGLD_CMD);

  return 0;
}

u32 apollo_display(void) {
  u32 status;
  send_command(APOLLO_DISP_IMAGE_CMD);
  status = apollo_get_status();
  return status;
}

u32 apollo_show_data(unsigned char *buffer, int bufsize) {
  apollo_load_image_data(buffer, bufsize);
  return apollo_display();
}


u32 apollo_show_partial_data(unsigned char *buffer, int bufsize, 
			     u16 x1, u16 y1, u16 x2, u16 y2) {
  int i;
  u32 status;
  unsigned long j1, j2, j3;
  j1 = jiffies;

  send_command(APOLLO_LOAD_PARTIAL_IMAGE_CMD);

  send_data((x1 >> 8) & 0xff);
  send_data(x1 & 0xff);

  send_data((y1 >> 8) & 0xff);
  send_data(y1 & 0xff);

  send_data((x2 >> 8) & 0xff);
  send_data(x2 & 0xff);

  send_data((y2 >> 8) & 0xff);
  send_data(y2 & 0xff);

  for(i = 0; i < bufsize; i++) {
    send_data(buffer[i]);
  }

  send_command(APOLLO_STOP_IMGLD_CMD);
  j2 = jiffies;
  send_command(APOLLO_DISP_PARTIAL_IMAGE_CMD);
  status = apollo_get_status();
  j3 = jiffies;
  /*  printk("load %ld display %ld\n", j2 - j1, j3 - j2); */
  return status;
}


int WriteToFlash (unsigned long Address, char Data) {

  int Check = 1;

  send_command (APOLLO_FLASH_WRITE);
  send_data((unsigned char)((Address & 0x00FF0000) >> 16));
  send_data((unsigned char)((Address & 0x0000FF00) >> 8));
  send_data((unsigned char)(Address & 0x000000FF));
  send_data(Data);

  return(Check);

}




int ReadFromFlash (unsigned long Address, char *Data) {

  int Check = 1;

  send_command (APOLLO_FLASH_READ);

  send_data((unsigned char)((Address & 0x00FF0000) >> 16));
  send_data((unsigned char)((Address & 0x0000FF00) >> 8));
  send_data((unsigned char)(Address & 0x000000FF));

  *Data = apollo_read();



  return(Check);

}

#define APOLLO_WRITE_REGISTER 0x10
#define APOLLO_READ_REGISTER 0x11

int apollo_write_register(unsigned char regAddress, unsigned char data) {
  send_command(APOLLO_WRITE_REGISTER);
  send_data(regAddress);
  send_data(data);
  return 0;
}

int apollo_read_register(unsigned char regAddress, unsigned char *data) {
  send_command(APOLLO_READ_REGISTER);
  send_data(regAddress);
  *data = apollo_read();
  return 0;
}

int ReadFromRam (unsigned long Address, unsigned long *Data) {
  unsigned char b0, b1, b2, b3;

  Address = Address >> 2;

  apollo_write_register(0x16, (Address & 0xff));
  apollo_write_register(0x15, (unsigned char) ((Address >> 8) & 0xff));
  apollo_write_register(0x14, (unsigned char) ((Address >> 16) & 0xff));
  apollo_write_register(0x17, 1);

  apollo_read_register(0x1d, &b0);
  apollo_read_register(0x1c, &b1);
  apollo_read_register(0x1b, &b2);
  apollo_read_register(0x1a, &b3);

  *Data = (b3 << 24) | (b2 << 16) | (b1 << 8) | b0;

  return 0;
}

int WriteToRam(unsigned long Address, unsigned long Data) {

  Address = Address >> 2;

  apollo_write_register(0x16, (Address & 0xff));
  apollo_write_register(0x15, (unsigned char) ((Address >> 8) & 0xff));
  apollo_write_register(0x14, (unsigned char) ((Address >> 16) & 0xff));


  apollo_write_register(0x1d, Data & 0xff);
  apollo_write_register(0x1c, (Data >> 8) & 0xff);
  apollo_write_register(0x1b, (Data >> 16) & 0xff);
  apollo_write_register(0x1a, (Data >> 24) & 0xff);

  apollo_write_register(0x17, 2);

  return 0;

}

int ProgramRam(unsigned long start_addr, unsigned char *buffer, unsigned long blen) {
  unsigned long i;
  unsigned long *lptr;
  lptr = (unsigned long *) buffer;
  for(i = start_addr; i < start_addr + blen; i+=4) {
    WriteToRam(i, *lptr);
    lptr++;
  }
  return 0;
}



int VerifyByte (unsigned long Address, char Data) {
  int Check;
  char Toggle;

  Check = ReadFromFlash (Address, &Toggle);

  if ((Check) && (Data != Toggle))
      Check = 0;

  return(Check);
}



unsigned long SectorSize (unsigned long Sector) {
  if (Sector < 63)
    return(64*1024);
  else
    return(8*1024);
}



unsigned long SectorAddress (unsigned long Sector) {;
  switch (Sector) {
  case 8:
    return(0x78000);
  case 9:
    return(0x7a000);
  case 10:
    return(0x7c000);
  default:
    return(Sector << 16);
  }

}



int EraseSector (unsigned long Sector) {
  unsigned long Address;
  int   Busy;
  int   Check;
  char  Toggle;
  char  SaveToggle;

  Address = SectorAddress (Sector);
  Check   = WriteToFlash (0x000AAA, (char)0xAA);
  Check   = (Check && WriteToFlash (0x000555, (char)0x55));
  Check   = (Check && WriteToFlash (0x000AAA, (char)0x80));
  Check   = (Check && WriteToFlash (0x000AAA, (char)0xAA));
  Check   = (Check && WriteToFlash (0x000555, (char)0x55));
  Check   = (Check && WriteToFlash (Address, (char)0x30));
  Check   = (Check && ReadFromFlash (Address, &SaveToggle));

  Busy = 1;

  while ((Check) && (Busy)) {
    Check = (Check && ReadFromFlash (Address, &Toggle));

    if (Check) { 
      Busy = (((Toggle & 0x40) != (SaveToggle & 0x40))); 
      SaveToggle = Toggle;
      if (Busy && ((Toggle & 0x20) != 0x00)) {
	Check = (Check && ReadFromFlash (Address, &Toggle));
	if (Check) {
	  Busy = (((Toggle & 0x40) != (SaveToggle & 0x40)) && ((Toggle & 0x80) == 0x00));
	  SaveToggle = Toggle;
	  if (Busy) 
	    Check = 0;
	}
      }
    }
  }
  return(Check);
}



int ProgramByte (unsigned long Address, unsigned char Data) {
  int Busy;
  int Check;
  char Toggle;
  char SaveToggle;

  Check = WriteToFlash (0x000AAA, (char)0xAA);
  Check = (Check && WriteToFlash(0x000555, (char)0x55));
  Check = (Check && WriteToFlash(0x000AAA, (char)0xA0));
  Check = (Check && WriteToFlash(Address, Data));
  Check = (Check && ReadFromFlash(Address, &SaveToggle));

  Busy = 1;

  while ((Check) && (Busy)) {
    Check = (Check && ReadFromFlash (Address, &Toggle));
    if (Check) {     
      Busy = ((Toggle & 0x40) != (SaveToggle & 0x40));
      SaveToggle = Toggle;
      if ((Busy) && ((Toggle & 0x20) != 0x00)) {
	Check = (Check && ReadFromFlash (Address, &Toggle));
	if (Check) {
	  Busy = ((Toggle & 0x40) != (SaveToggle & 0x40));
	  SaveToggle = Toggle;
	  if (Busy)    
	    Check = 0;  
	}
      }
    }
  } 
  return(Check);
}

#if 0
int ProgramFlash(unsigned long FlashLength, char *FlashFile) {
  unsigned long FlashPtr = 0;
  unsigned long NextSector = 0;

  EraseSector(0);	
  NextSector = 1;

  for (FlashPtr = 0; FlashPtr < FlashLength; FlashPtr++) {

    /* If we have crossed into the next sector, erase it */
    
    if (FlashPtr >= SectorAddress(NextSector)) {
      
      printk("Erasing next sector.\n");
      EraseSector(NextSector);
      NextSector++;
    }

    /* Repeatedly try to program and verify the byte until you get it */

    if (FlashFile[FlashPtr] != 0xFF) {
      if (!ProgramByte(FlashPtr, FlashFile[FlashPtr]))
	return(0);

      printk("Programmed byte %u with %x\n", FlashPtr, FlashFile[FlashPtr]); 
      if (!VerifyByte(FlashPtr, FlashFile[FlashPtr]))
	return(0);
    }

    if (((FlashPtr % 1024) == 0) && (FlashPtr != 0)) {
      printk("%u...",FlashPtr);
      if ((FlashPtr % (1024*10)) == 0)
	printk("\n");			
    }	
  }

  printk("done\n");
  return(1);
}

#endif 

int ProgramFlash(unsigned long start_addr, unsigned char *buffer, unsigned long blen) {
  static unsigned long nextSector = 0;
  static unsigned long nextSectorAddress = 0;

  unsigned long bptr = 0;
  unsigned long curraddr = 0;

  if(start_addr == 0) {
    printk("resetting sector & address\n");
    nextSector = 0;
    nextSectorAddress = 0;
  }
  printk("programming chunk size %d at %d\n", (int) blen, (int) start_addr);
  for(bptr = 0, curraddr = start_addr; bptr < blen; bptr++, curraddr++) {

    if(curraddr >= nextSectorAddress) {
      printk("erase sector %d\n", (int) nextSector);
      EraseSector(nextSector);
      nextSector++;
      nextSectorAddress = SectorAddress(nextSector);
    }
    
    if(buffer[bptr] != 0xFF) {
      if(!ProgramByte(curraddr, buffer[bptr])) {
	return 0;
      }

      if(!VerifyByte(curraddr, buffer[bptr])) {
	return 0;
      }

    }
  } 
  return 1;
}
