/*************************************************************
 * Philips ISP1761 peripheral controller driver
 *
 * (c) 2002 Koninklijke Philips Electronics N.V., All rights reserved
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * File Name:	pdc.c	(Based on usb_pdc.c - heavily reworked to work
 *			 in the USB gadget framework.)
 *
 * Refering linux kernel version 2.6.9
 *
 * History:	
 *
 * Version	Date		Author		Comments
 * -----------------------------------------------------------
 * 1.0 	 	09/08/04     SRIRAL PAULRAJ	Initial Creation
 * 1.1      	02/09/04     SRIRAM PAULRAJ	Modified for 1761 reg access
 *
 *************************************************************/


#undef DEBUG		/* enable additional debug code and messages	*/
#undef DEBUG_VERBOSE	/*   add additional informational messages	*/
#undef DEBUG_ENTRY	/*   add function entry/exit messages		*/
#undef DEBUG_TRACE	/*   add USB protocol and data trace info	*/

#define DEBUG_CHIP	/* enable chip register display in sysfs	*/


#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/device.h>
#include <linux/sem.h>
#include <linux/suspend.h>
#include <linux/syscalls.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/usb_ch9.h>
#include <linux/usb_gadget.h>

#include <asm/arch-pxa/fiona_eventmap.h>
#include <asm/arch-pxa/pxa-regs.h>
#include <asm/arch-pxa/boot_globals.h>

#include "pdc.h"


static struct pdc_globals globals;

#ifdef CONFIG_FIONA_PM_USB_DEVICE
// Static global holding fpow device entry returned from fpow_register()
extern struct fpow_component *g_usb_device_fpow_component_ptr;
extern int usb_device_register_with_fpow(void *private, struct fpow_component **fpow_dev);
extern void usb_device_unregister_with_fpow(struct fpow_component *fpow_dev);
#endif

#define DRIVER_AUTHOR 		"Philips Semiconductors/Lab126"
#define	DRIVER_DESC		"ISP1761 Peripheral Controller Driver"
#define DRIVER_VERSION 		"1.0"
#define	DRIVER_NAME		"pdc"

#define USB_BULK_RESET_REQUEST	0xFF


/* -----  globals  ----- */

static struct usb_ep_ops pdc_ep_ops = {
	.enable		= pdc_ep_enable,
	.disable	= pdc_ep_disable,

	.alloc_request	= pdc_ep_alloc_request,
	.free_request	= pdc_ep_free_request,

	.alloc_buffer	= pdc_ep_alloc_buffer,
	.free_buffer	= pdc_ep_free_buffer,

	.queue		= pdc_ep_queue,
	.dequeue	= pdc_ep_dequeue,

	.set_halt	= pdc_ep_set_halt,
	.fifo_status	= pdc_ep_fifo_status,
	.fifo_flush	= pdc_ep_fifo_flush,
};


static const struct usb_gadget_ops pdc_ops = {
	.get_frame		= pdc_get_frame,
	.wakeup			= pdc_wakeup,
	// .set_selfpowered	= ... current versions must always be self-powered
	// .vbus_session	= ...
	.vbus_draw		= pdc_vbus_draw,
	// .pullup		= ...
	// .ioctl		= ...
};


static struct platform_device pdc_device = {
	.name		= "pdc",
	.id		= -1,
	.dev.release	= nop_release,
};


static struct device_driver pdc_driver = {
	.name		= "pdc",
	.bus		= &platform_bus_type,
	.probe		= pdc_probe,
	.remove		= pdc_remove,	// __exit_p(pdc_remove) ???
	// power mgmt support
	// .suspend	= pdc_suspend,
	// .resume	= pdc_resume,
};


static const char ep0name[] = "ep0";


static struct pdc_globals globals = {
	.gadget = {
		.ops		= &pdc_ops,
		.ep0		= &globals.ep[0].ep,
		.name		= DRIVER_NAME,
		.is_dualspeed	= 1,
		.dev = {
			.bus_id		= "gadget",
			.release	= nop_release,
		},
	},
	
	.ep[0] = {
		.ep = {
			.name		= ep0name,
			.ops		= &pdc_ep_ops,
			.maxpacket	= EP0_FIFO_SIZE,
		},
		.dev		= &globals,
		.ep_number	= 0,
		.ep_index	= EP0TX,   // makes GET_DESCRIPTOR work...
		.bEndptAddress	= 0,
	},
	
	.ep[1] = {
		.ep = {
			.name		= "ep1in-bulk",
			.ops		= &pdc_ep_ops,
			.maxpacket	= BULK_FIFO_SIZE_FS,
		},
		.dev		= &globals,
		.ep_number	= 1,
		.ep_index	= EP1TX,
		.bEndptAddress	= USB_DIR_IN | 1,
		.bmAttributes	= USB_ENDPOINT_XFER_BULK,
	},

	.ep[2] = {
		.ep = {
			.name		= "ep2out-bulk",
			.ops		= &pdc_ep_ops,
			.maxpacket	= BULK_FIFO_SIZE_FS,
		},
		.dev		= &globals,
		.ep_number	= 2,
		.ep_index	= EP2RX,
		.bEndptAddress	= 2,
		.bmAttributes	= USB_ENDPOINT_XFER_BULK,
	},
};

static struct pdc_globals	*pdc = &globals;

static char pdc_test_pkt[] =
{	
	/* Philips' test packet contents */
	0xC3, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA,
	0xAA, 0xAA, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE, 0xEE,
	0xEE, 0xEE, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0xBF,
	0xDF, 0xEF, 0xF7, 0xFB, 0xFD, 0xFC, 0x7E, 0xBF,
	0xDF, 0xEF, 0xF7, 0xFB, 0xFD, 0x7E, 0xB6, 0xCE,  
};


static const char *state_names[] = { "DETACHED", "ATTACHED",
				     "POWERED", "POWERED_SUSPENDED",
				     "DEFAULT", "DEFAULT_SUSPENDED",
				     "ADDRESS", "ADDRESS_SUSPENDED",
				     "CONFIGURED", "CONFIGURED_SUSPENDED" };

/* When the pdc controller is connected to a USB charger or a USB hub port that
   isn't generating any traffic, a continuous stream of VBUS interrupts is
   generated.  To work around this problem, a different set of interrupts need
   to be enabled when we're in the POWERED and POWERED_SUSPENDED states.
   This table provides the appropriate interrupt enables for all the USB bus
   states.								*/

static const u32  intr_enables_table[] =
		{ IE_STATE_DETACHED,	IE_STATE_ATTACHED,
		  IE_STATE_POWERED,	IE_STATE_POWERED_SUSPENDED,
		  IE_STATE_DEFAULT,	IE_STATE_DEFAULT_SUSPENDED,
		  IE_STATE_ADDRESS,	IE_STATE_ADDRESS_SUSPENDED,
		  IE_STATE_CONFIGURED,	IE_STATE_CONFIGURED_SUSPENDED };


#ifdef CONFIG_FIONA_PM_USB_DEVICE
#include "usb_device_pm.c"
#endif


static void pdc_dump(void)
{
#ifdef DEBUG_VERBOSE
	DBG_MSG("0x200: Address          = 0x%04X\n", pdc_read16(DC_ADDRESS_REG));
	DBG_MSG("0x204: Endpt MaxPktSize = 0x%04X\n", pdc_read16(DC_MAX_PKT_SIZE_REG));
	DBG_MSG("0x208: Endpt Type       = 0x%04X\n", pdc_read16(DC_EP_TYPE_REG));
	DBG_MSG("0x20C: Mode             = 0x%04X\n", pdc_read16(DC_MODE_REG));
	DBG_MSG("0x210: Intr Config      = 0x%04X\n", pdc_read16(DC_INTR_CONFIG_REG));
	DBG_MSG("0x212: Debug            = 0x%04X\n", pdc_read16(DC_DEBUG_REG));
	DBG_MSG("0x214: Intr Enable      = 0x%08X\n", pdc_read32(DC_INTR_ENABLE_REG));
	DBG_MSG("0x218: Interrupt        = 0x%08X\n", pdc_read32(DC_INTERRUPT_REG));
	DBG_MSG("0x21C: Buffer Length    = 0x%04X\n", pdc_read16(DC_BUFFER_LENGTH_REG));
	DBG_MSG("0x21E: Buffer Status    = 0x%04X\n", pdc_read16(DC_BUFFER_STATUS_REG));
	DBG_MSG("0x220: Data Port        = 0x%04X\n", pdc_read16(DC_DATAPORT_REG));
	DBG_MSG("0x228: Ctrl Func        = 0x%04X\n", pdc_read16(DC_CTRLFN_REG));
	DBG_MSG("0x22C: Endpt Index      = 0x%04X\n", pdc_read16(DC_EP_INDEX_REG));
	DBG_MSG("0x230: DMA Cmd          = 0x%04X\n", pdc_read16(DC_DMA_COMMAND_REG));
	DBG_MSG("0x234: DMA Transfer Ctr = 0x%08X\n", pdc_read32(DC_DMA_TRANSFER_CTR_REG));
	DBG_MSG("0x238: DMA Config       = 0x%04X\n", pdc_read16(DC_DMA_CONFIG_REG));
	DBG_MSG("0x23C: DMA Hardware     = 0x%04X\n", pdc_read16(DC_DMA_HARDWARE_REG));
	DBG_MSG("0x250: DMA Intr Reason  = 0x%04X\n", pdc_read16(DC_DMA_INTR_REASON_REG));
	DBG_MSG("0x254: DMA Intr Enable  = 0x%04X\n", pdc_read16(DC_DMA_INTR_ENABLE_REG));
	DBG_MSG("0x258: DMA Endpt        = 0x%04X\n", pdc_read16(DC_DMA_EP_INDEX_REG));
	DBG_MSG("0x264: DMA Burst Ctr    = 0x%04X\n", pdc_read16(DC_DMA_BURST_CTR_REG));
	DBG_MSG("0x270: Chip ID          = 0x%08X\n", pdc_read32(DC_CHIP_ID_REG));
	DBG_MSG("0x274: Frame Number     = 0x%04X\n", pdc_read16(DC_FRAME_NUMBER_REG));
	DBG_MSG("0x278: Scratch          = 0x%04X\n", pdc_read16(DC_SCRATCH_REG));
	DBG_MSG("0x27C: Unlock Device    = 0x%04X\n", pdc_read16(DC_UNLOCK_DEVICE_REG));
	DBG_MSG("0x284: Test Mode        = 0x%04X\n", pdc_read16(DC_TESTMODE_REG));
	DBG_MSG("0x300: Hw Mode Control  = 0x%08X\n", pdc_read32(HC_HW_MODE_REG));
	DBG_MSG("0x374: OTG Control      = 0x%08X\n", pdc_read32(OTG_CONTROL_REG));
#endif
}

#ifdef DEBUG	//_VERBOSE
static void pdc_dump_request(struct pdc_request *req)
{
	struct usb_request	*r = &req->req;

	DBG_MSG("  struct pdc_request at 0x%p\n", req);
	DBG_MSG("   struct usb_request at 0x%p\n", r);
	DBG_MSG("    buf       = 0x%p\n", r->buf);
	DBG_MSG("    length    = %d\n",   r->length);
	//G_MSG("    dma       = 0x%p\n", r->dma);
	DBG_MSG("    no_intr   = %d\n",   r->no_interrupt);
	DBG_MSG("    zero      = %d\n",   r->zero);
	DBG_MSG("    short_bad = %d\n",   r->short_not_ok);
	DBG_MSG("    complete  = 0x%p\n", r->complete);
	DBG_MSG("    context   = 0x%p\n", r->context);
	//G_MSG("    list      = 0x%p\n", r->list);
	DBG_MSG("    status    = %d\n",   r->status);
	DBG_MSG("    actual    = %d\n",   r->actual);
	DBG_MSG("   timestamp  = 0x%08X\n", req->timestamp);
}
#endif

static void pdc_dump_ep(struct pdc_ep *ep)
{
#ifdef DEBUG	//_VERBOSE
	struct pdc_request	*r;
	struct list_head	*p;

	DBG_MSG("struct pdc_ep ep[%s]:\n", ep->ep.name);
	//struct usb_ep		ep;
//	DBG_MSG("  ep.driver_data    = 0x%p\n",ep->ep.driver_data);
//	DBG_MSG("  ep.name           = \"%s\"\n", ep->ep.name);
//	DBG_MSG("  ep.ops            = 0x%p\n", ep->ep.ops);
	//struct list_head	ep_list;
	
	DBG_MSG("  ep.maxpacket      = 0x%04x\n", ep->ep.maxpacket);

//	DBG_MSG("  dev               = 0x%p\n", ep->dev);
	
//	if (ep->desc) {
//	//struct usb_endpoint_descriptor	*desc;
//	DBG_MSG("  desc->bLength          = 0x%02x\n", ep->desc->bLength);
//	DBG_MSG("  desc->bDescriptorType  = 0x%02x\n", ep->desc->bDescriptorType);
//	DBG_MSG("  desc->bEndpointAddress = 0x%02x\n", ep->desc->bEndpointAddress);
//	DBG_MSG("  desc->bmAttributes     = 0x%02x\n", ep->desc->bmAttributes);
//	DBG_MSG("  desc->wMaxPacketSize   = 0x%04x\n", ep->desc->wMaxPacketSize);
//	DBG_MSG("  desc->bInterval        = 0x%02x\n", ep->desc->bInterval);
//	} else {
//	DBG_MSG("  desc                   = NULL\n");
//	}
	
	//struct list_head	queue;
	list_for_each(p, &ep->queue) {
		r = list_entry(p, struct pdc_request, queue);
		pdc_dump_request(r);
	}
	
//	DBG_MSG("  irqs                   = 0x%08lx\n", ep->irqs);
	
//	DBG_MSG("  ep_index               = 0x%02x\n", ep->ep_index);
//	DBG_MSG("  bmAttributes           = 0x%02x\n", ep->bmAttributes);
	DBG_MSG("  stopped                = 0x%02x\n", ep->stopped);
#endif
}

#ifdef DEBUG_VERBOSE
static void pdc_dump_gadget(struct usb_gadget *gadget)
{
	DBG_MSG("struct usb_gadget at: 0x%p\n", gadget);
	DBG_MSG("  ops               = 0x%p\n", gadget->ep0->ops);
	DBG_MSG("  ep0->driver_data  = 0x%p\n", gadget->ep0->driver_data);
	DBG_MSG("  ep0->name         = \"%s\"\n", gadget->ep0->name);
	DBG_MSG("  ep0->ops          = 0x%p\n", gadget->ep0->ops);
		//struct list_head	ep_list;
	DBG_MSG("  ep0->maxpacket    = 0x%04x\n", gadget->ep0->maxpacket);
	//struct list_head		ep_list;	/* of usb_ep */
	DBG_MSG("  speed             = %d\n", gadget->speed);
	DBG_MSG("  is_dualspeed      = %d\n", gadget->is_dualspeed);
	DBG_MSG("  is_otg            = %d\n", gadget->is_otg);
	DBG_MSG("  is_a_peripheral   = %d\n", gadget->is_a_peripheral);
	DBG_MSG("  b_hnp_enable      = %d\n", gadget->b_hnp_enable);
	DBG_MSG("  a_hnp_support     = %d\n", gadget->a_hnp_support);
	DBG_MSG("  a_alt_hnp_support = %d\n", gadget->a_alt_hnp_support);
	DBG_MSG("  name              = \"%s\"\n", gadget->name);
	//struct device			dev;
};
#endif

static void pdc_dump_globals(struct pdc_globals *g)
{
#ifdef DEBUG_VERBOSE
	int	i;
	DBG_MSG("struct pdc_globals at: 0x%p\n", g);
	pdc_dump_gadget(&g->gadget);
	DBG_MSG("driver        = 0x%p\n", g->driver);
	DBG_MSG("base_address  = 0x%p\n", g->base_address);
	DBG_MSG("intr_enables  = 0x%08x\n", g->intr_enables);
	DBG_MSG("device_state  = 0x%02x\n", g->device_state);
	DBG_MSG("got_irq       = 0x%02x\n", g->got_irq);
	//DBG_MSG("got_disc      = 0x%02x\n", g->got_disc);
	//DBG_MSG("req_pending   = 0x%02x\n", g->req_pending);
	//DBG_MSG("req_std       = 0x%02x\n", g->req_std);
	//DBG_MSG("req_config    = 0x%02x\n", g->req_config);
	// struct timer_list	timer;
	// struct device	*dev;
	DBG_MSG("dev           = 0x%p\n", g->dev);
	for (i = 0 ; i < PDC_NUM_ENDPOINTS ; i++)
		pdc_dump_ep(&g->ep[i]);
#endif
}


static void pdc_enable_ep_interrupt(struct pdc_globals *dev, struct pdc_ep *ep)
{
	u32	int_enables = pdc_read32(DC_INTR_ENABLE_REG);
	u32	ep_enable   = (IE_EP0RX << ep->ep_index);
	
	dev->intr_enables_bulk	|= ep_enable;
	int_enables		|= ep_enable;
	pdc_write32(DC_INTR_ENABLE_REG, int_enables);
	DBG_MSG("%s: DC_INTR_ENABLE_REG = 0x%x\n", __FUNCTION__,
			pdc_read32(DC_INTR_ENABLE_REG));
}

static void pdc_disable_ep_interrupt(struct pdc_globals *dev, struct pdc_ep *ep)
{
	u32	int_enables = pdc_read32(DC_INTR_ENABLE_REG);
	u32	ep_enable   = (IE_EP0RX << ep->ep_index);

	dev->intr_enables_bulk	&= ~(ep_enable);
	int_enables		&= ~(ep_enable);
	pdc_write32(DC_INTR_ENABLE_REG, int_enables);
	DBG_MSG("%s: DC_INTR_ENABLE_REG = 0x%x\n", __FUNCTION__,
			pdc_read32(DC_INTR_ENABLE_REG));
}

static void pdc_enable_endpoint(struct pdc_globals *dev, struct pdc_ep *ep)
{
	u16	max_pkt_size;
	
	if (dev->gadget.speed == USB_SPEED_FULL)
		max_pkt_size = BULK_FIFO_SIZE_FS;
	else
		max_pkt_size = BULK_FIFO_SIZE_HS;
		
	// Issue "clear buffer" command twice

	pdc_write16(DC_EP_INDEX_REG, ep->ep_index);

	pdc_write16(DC_CTRLFN_REG, CNTRLFN_CLBUF);
	
	pdc_write16(DC_EP_INDEX_REG, ep->ep_index);

	pdc_write16(DC_CTRLFN_REG, CNTRLFN_CLBUF);
	
	// Disable endpoint

	pdc_write16(DC_EP_INDEX_REG, ep->ep_index);

	pdc_write16(DC_EP_TYPE_REG, 0);
	
	// Set MaxPacketSize for endpoint
	
	pdc_write16(DC_EP_INDEX_REG, ep->ep_index);
	
	pdc_write16(DC_MAX_PKT_SIZE_REG, max_pkt_size);
	
	if (ep->ep_index & 0x01)	// IN endpoint
		pdc_write16(DC_BUFFER_LENGTH_REG, 42);	// something nonzero
	
	// Set endpoint type...	[FIXME might want to set DBLBUF, too]
	
	pdc_write16(DC_EP_TYPE_REG, EP_TYP_ENDPTYP_BULK);

	//    and enable endpoint
	
	pdc_write16(DC_EP_TYPE_REG, EP_TYP_ENDPTYP_BULK | EP_TYP_ENABLE);
}

static void pdc_disable_endpoint(struct pdc_globals *dev, struct pdc_ep *ep)
{
	pdc_write16(DC_EP_INDEX_REG, ep->ep_index);
	pdc_write16(DC_EP_TYPE_REG, 0);
}


static void pdc_disable_all_endpts(struct pdc_globals *dev)
{
	u16	i;
	
	for (i = EP0RX ; i <= EP7TX ; i++) {
		pdc_write16(DC_EP_INDEX_REG, i);
		pdc_write16(DC_EP_TYPE_REG, 0);	// Disable all endpoints
	}
}

static u16 pdc_read_ep(u8 endpoint, u8 *buffer, u16 length)
{
	u32			*wbuf = (u32 *) buffer;
	u16			available;
	u16			received = 0;
	struct pdc_globals	*dev = &globals;
	
	DBG_VMSG("%s(ep=0x%x, buf=0x%p, length=%d)\n", __FUNCTION__, endpoint,
			buffer, length);
	
	if (length == 0) {
		DBG_MSG("length is zero\n");
	}
	
	if ((u32) wbuf & 3) {
		DBG_MSG("buffer is not word aligned!\n");
	}
	
	if (endpoint == EP0RX) {
		DBG_MSG("%s: trying to read EP0RX!\n", __FUNCTION__);
	}
	
	pdc_write32(DC_EP_INDEX_REG, endpoint);

	if (pdc_read16(DC_BUFFER_STATUS_REG) == 0) {
	
	// Looks like we can get numerous "spurious" OUT interrupts where
	// there is no data available to be read.  If so, just exit.

		DBG_VMSG("Spurious OUT\n");
		goto exit;
	}
	
	pdc_write32(DC_EP_INDEX_REG, endpoint); // try setting ep index again

	available = pdc_read16(DC_BUFFER_LENGTH_REG);
	
	// We occasionally get 4-byte runt CBW packets out of the 1761, even
	// though we ACK'ed a valid 31-byte packet from the host.
	//
	// If we see this case, we set the cbw_error flag, and refuse to clear
	// STALL conditions until either Mass Storage Bulk Reset or USB Bus
	// Reset occurs.  This allows a recovery in about 6-30 seconds.
	
	if ((length == 31) && (available != 31)) {
		DBG_MSG("length=%d, available=%d and %d\n", length, available,
  				pdc_read16(DC_BUFFER_LENGTH_REG));
		DBG_PRINTK("Short CBW returned by pdc!!! (%d of %d bytes)\n",
  				available, length);
		memset(buffer, 0, length);
		dev->cbw_error = 1;
	}

	DBG_VMSG("r%d\n", available);
	
	if (available == 0) {
		DBG_MSG("%s: zero-length packet!\n", __FUNCTION__);
		pdc_write32(DC_EP_INDEX_REG, endpoint);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_CLBUF);
		goto exit;
	}

	DBG_VMSG("%s: %d bytes available\n", __FUNCTION__, available);

	if (length < available) {
		DBG_MSG("%s: length (%d) < available (%d)\n", __FUNCTION__,
				length, available);
		available = length;
	}
		
	received = available;

	while (available >= 4) {
		*wbuf = pdc_read32(DC_DATAPORT_REG);
		wbuf++;
		available -= 4;
	}
	if (available > 0) {
		u32	w = pdc_read32(DC_DATAPORT_REG);
		u8	*bbuf = (u8 *) wbuf;
		
		w = (w << 24) | ((w & 0x0000FF00) << 8) |
				((w & 0x00FF0000) >> 8) |
				((w & 0xFF000000) >> 24);
		while (available > 0) {
			*bbuf = w & 0xFF;
			bbuf++;
			w >>= 8;
			available--;
		}
	}
exit:
	if (dev->cbw_error)
		received = length;

	return received;
}

static void pdc_write_ep(u8 endpoint, u8 *buffer, u16 length)
{
	u32	*wbuf = (u32 *) buffer;
	int	len = length;
	
	DBG_VMSG("%s(ep=0x%x, buf=0x%p, length=%d)\n", __FUNCTION__, endpoint,
			buffer, length);
	if ((u32) wbuf & 3)
		DBG_MSG("buffer is not word aligned!\n");
	
	DBG_VMSG("w%d\n", length);

	pdc_write32(DC_EP_INDEX_REG, endpoint);
	
	if (endpoint == EP0TX) {
		DBG_VMSG("%s: writing DSEN for EP0TX\n", __FUNCTION__);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_DSEN);
	}
	
	pdc_write16(DC_BUFFER_LENGTH_REG, length);
	
	while (len > 0) {
		pdc_write32(DC_DATAPORT_REG, *wbuf);
		wbuf++;
		len -= 4;
	}
	
	udelay(150);
}

/*
 *	pdc_complete - retire a request; caller blocked irqs
 */
static void pdc_complete(struct pdc_ep *ep, struct pdc_request *req, int status)
{
	struct pdc_globals	*dev = &globals;
	unsigned		stopped = ep->stopped;

	pdc_log_entry("\n");
	
	list_del_init(&req->queue);

	if (likely (req->req.status == -EINPROGRESS))
		req->req.status = status;
	else
		status = req->req.status;

	if (status && status != -ESHUTDOWN)
		DBG_MSG("complete %s req %p stat %d len %u/%u\n",
			ep->ep.name, &req->req, status,
			req->req.actual, req->req.length);

	/* don't modify queue heads during completion callback */
	ep->stopped = 1;
	spin_unlock(&dev->lock);
	req->req.complete(&ep->ep, &req->req);
	spin_lock(&dev->lock);
	ep->stopped = stopped;
	
	pdc_log_exit();
}


static u8 pdc_vbus_state(void)
{
	if (pdc_read16(DC_MODE_REG) & MODE_VBUSSTAT)
		return 1;
	else
		return 0;
}


/* -----  api related to endpoint operations  ----- */

static int pdc_ep_enable (struct usb_ep *_ep,
		const struct usb_endpoint_descriptor *desc)
{
	struct pdc_ep		*ep;
	struct pdc_globals	*dev = &globals;
	unsigned long		flags;
	u16			mps = 0;
	u8			attr = 0;
	
	pdc_log_entry("\n");

	DBG_MSG("%s(\"%s\",maxpkt=%d)\n", __FUNCTION__, _ep->name,
			_ep->maxpacket);
	
	pdc_dump_globals(dev);
	
	if (desc) {
		mps = le16_to_cpu(desc->wMaxPacketSize);
		attr = desc->bmAttributes;
	} else {
		return -EINVAL;
	}
		
	DBG_MSG("ep_descriptor: len=%02X type=%02X addr=%02X attr=%02X "
			"mxpkt=%04X intrv=%02X\n",
			desc->bLength, desc->bDescriptorType,
			desc->bEndpointAddress, desc->bmAttributes, mps,
			desc->bInterval);

	ep = container_of(_ep, struct pdc_ep, ep);
	if (!_ep || !desc || ep->desc || _ep->name == ep0name
			|| desc->bDescriptorType != USB_DT_ENDPOINT
		//	|| ep->bEndpointAddress != desc->bEndpointAddress
			|| ep->ep.maxpacket < mps) {
		DBG_MSG("%s: bad ep or descriptor\n", __FUNCTION__);
		return -EINVAL;
	}

	/* xfer types must match, except that interrupt ~= bulk */
	if (ep->bmAttributes != desc->bmAttributes
			&& ep->bmAttributes != USB_ENDPOINT_XFER_BULK
			&& desc->bmAttributes != USB_ENDPOINT_XFER_INT) {
		DBG_MSG("%s, %s type mismatch\n", __FUNCTION__, _ep->name);
		return -EINVAL;
	}

	/* hardware _could_ do smaller, but driver doesn't */
	if ((attr == USB_ENDPOINT_XFER_BULK && (mps != ep->ep.maxpacket)) 
			|| !desc->wMaxPacketSize) {
		DBG_MSG("%s, bad %s maxpacket\n", __FUNCTION__, _ep->name);
		return -ERANGE;
	}

	dev = ep->dev;
	if (!dev->driver || dev->gadget.speed == USB_SPEED_UNKNOWN) {
		DBG_MSG("%s, bogus device state\n", __FUNCTION__);
		return -ESHUTDOWN;
	}

	spin_lock_irqsave(&dev->lock, flags);
	
	ep->desc = desc;
	//ep->dma = -1;
	ep->stopped = 0;
	//ep->pio_irqs = ep->dma_irqs = 0;
	ep->ep.maxpacket = mps;
	
	pdc_enable_endpoint(dev, ep);
	
	/* flush fifo (mostly for OUT buffers) */
	
	if ((ep->ep_index & 0x01) == 0) 	// OUT endpt
		pdc_locked_do_ep_fifo_flush(_ep);

	pdc_locked_do_ep_clear_halt(_ep);
	
	pdc_enable_ep_interrupt(dev, ep);

	DBG_MSG("enabled %s\n", _ep->name);

	spin_unlock_irqrestore(&dev->lock, flags);

	pdc_log_exit();
	return 0;
}

static int pdc_ep_disable (struct usb_ep *_ep)
{
	struct pdc_globals	*dev = &globals;
	struct pdc_ep		*ep;
	unsigned long		flags;
	
	pdc_log_entry("%s\n", _ep->name);

	DBG_MSG("%s(\"%s\")\n", __FUNCTION__, _ep->name);

	ep = container_of(_ep, struct pdc_ep, ep);
	
	spin_lock_irqsave(&dev->lock, flags);
	
	pdc_disable_ep_interrupt(dev, ep);

	ep->stopped = 1;
	pdc_nuke(ep, -ESHUTDOWN);

	ep->desc = NULL;
	
	pdc_log_exit();

	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}

static struct usb_request *
pdc_ep_alloc_request (struct usb_ep *_ep, int gfp_flags)
{
	struct pdc_request	*req;

	pdc_log_entry("\n");
	
	req = kmalloc (sizeof *req, gfp_flags);
	if (!req)
		return 0;

	memset (req, 0, sizeof *req);
	INIT_LIST_HEAD (&req->queue);

	pdc_log_exit();
	
	return &req->req;
}

static void
pdc_ep_free_request (struct usb_ep *_ep, struct usb_request *_req)
{
	struct pdc_request	*req;

	pdc_log_entry("\n");
	
	req = container_of (_req, struct pdc_request, req);
	WARN_ON (!list_empty (&req->queue));
	kfree(req);

	pdc_log_exit();
}


static void *
pdc_ep_alloc_buffer(struct usb_ep *_ep, unsigned bytes,
	dma_addr_t *dma, int gfp_flags)
{
	char			*retval;

	pdc_log_entry("\n");
	
	retval = kmalloc (bytes, gfp_flags & ~(__GFP_DMA|__GFP_HIGHMEM));
	if (retval)
		*dma = /*virt_to_bus*/(dma_addr_t) (retval);
		// FIXME - deprecated

	pdc_log_exit();
	
	return retval;
}

static void
pdc_ep_free_buffer(struct usb_ep *_ep, void *buf, dma_addr_t dma,
		unsigned bytes)
{
	pdc_log_entry("\n");
	
	kfree (buf);

	pdc_log_exit();
}

static int pdc_ep_queue(struct usb_ep *_ep, struct usb_request *_req,
				int gfp_flags)
{
	struct pdc_request	*req;
	struct pdc_ep		*ep;
	struct pdc_globals	*dev;
	unsigned long		flags;
	u16			len, rcvd;

	pdc_log_entry("\n");
	
	DBG_TRACE("ep: \"%s\", buf: 0x%p, len: 0x%x\n", _ep->name, _req->buf,
			_req->length);
	//DBG_MSG("_req->buf          = 0x%p\n", _req->buf);
	//DBG_MSG("_req->length       = 0x%x\n", _req->length);
	//DBG_MSG("_req->dma          = 0x%x\n", _req->dma);
	//DBG_MSG("_req->no_interrupt = %d\n", _req->no_interrupt);
	//DBG_MSG("_req->zero         = %d\n", _req->zero);
	//DBG_MSG("_req->short_not_ok = %d\n", _req->short_not_ok);
	//DBG_MSG("_req->complete     = 0x%p\n", _req->complete);
	//DBG_MSG("_req->context      = 0x%p\n", _req->context);
	//DBG_MSG("_req->list         = ???\n");
	//DBG_MSG("_req->status       = %d\n", _req->status);
	//DBG_MSG("_req->actual       = %d\n", _req->actual);

	req = container_of(_req, struct pdc_request, req);
	if (unlikely (!_req || !_req->complete || !_req->buf
			|| !list_empty(&req->queue))) {
		DBG_MSG("%s, bad params\n", __FUNCTION__);
		if (!list_empty(&req->queue)) {
			DBG_MSG("%s - queue non-empty!\n", __FUNCTION__);
			ep = container_of(_ep, struct pdc_ep, ep);
			pdc_dump_ep(ep);
		} else {
			DBG_MSG("%s - !_req=%d, !_req->complete=%d, "
				"!_req->buf=%d\n", __FUNCTION__, !_req,
				!_req->complete, !_req->buf);
		}
		return -EINVAL;
	}
	
	ep = container_of(_ep, struct pdc_ep, ep);
	if (unlikely (!_ep || (!ep->desc && ep->ep.name != ep0name))) {
		DBG_MSG("%s, bad ep\n", __FUNCTION__);
		return -EINVAL;
	}

	//DBG_MSG("ep->ep.name          = \"%s\"\n", ep->ep.name);
	//DBG_MSG("ep->ep.maxpacket     = %d\n", ep->ep.maxpacket);
	//DBG_MSG("ep->ep_index         = 0x%x\n", ep->ep_index);
	//DBG_MSG("ep->bmAttributes     = 0x%x\n", ep->bmAttributes);
	//DBG_MSG("ep->stopped          = 0x%x\n", ep->stopped);
	
	dev = ep->dev;
	if (unlikely (!dev || !dev->driver
			|| dev->gadget.speed == USB_SPEED_UNKNOWN)) {
		DBG_MSG("%s, bogus device state\n", __FUNCTION__);
		return -ESHUTDOWN;
	}
	
	spin_lock_irqsave(&dev->lock, flags);
	
	DBG_VMSG("ep%d: %d\n", ep->ep_number, _req->length);
	
	req->timestamp = OSCR;

	_req->status = -EINPROGRESS;
	_req->actual = 0;
	
	/* Start up this I/O queue, if necessary */
	
	if (list_empty(&ep->queue)) {
		
	    if (likely(!ep->stopped)) {
	    
		if (EP_DIRECTION_IS_IN(ep)) {		// IN endpoint
		
			// Send the first packet to get things rolling.
		
			len = min((u16) _req->length, (u16) ep->ep.maxpacket);
		
			pdc_write_ep(ep->ep_index, _req->buf, len);
			
			_req->actual += len;

		} else {				// OUT endpoint
			
			len = min((u16) _req->length, (u16) ep->ep.maxpacket);
			
			rcvd = pdc_read_ep(ep->ep_index, _req->buf, len);
			DBG_VMSG("rcvd = %d\n", rcvd);
			
			_req->actual += rcvd;
			
			if (rcvd &&
			    ((rcvd < (u16) ep->ep.maxpacket) ||
			     (rcvd == _req->length))) {
				pdc_complete(ep, req, 0);
				req = NULL;	// done - no need to enqueue
			}
			
		}
		
	    } else {

	    	// The queue is empty, but the endpt is currently stopped.
		// This usually occurs because the endpt is stalled, but it
		// also happens while a request's completion routine is being
		// run.  When the endpt is unstopped, we need to start up
		// this request.

		DBG_VMSG("setting needs_kick\n");

		ep->needs_kick = 1;

	    }

	}
	
	if (req)
		list_add_tail(&req->queue, &ep->queue);
	
	pdc_log_exit();

	spin_unlock_irqrestore(&dev->lock, flags);
	
	return 0;
}

/* dequeue JUST ONE request */
static int pdc_ep_dequeue(struct usb_ep *_ep, struct usb_request *_req)
{
	struct pdc_globals	*dev = &globals;
	struct pdc_ep		*ep;
	struct pdc_request	*req;
	unsigned long		flags;
	int			retval = 0;
	
	pdc_log_entry("\n");

	ep = container_of(_ep, struct pdc_ep, ep);
	if (!ep || ep->ep.name == ep0name)
		return -EINVAL;

	spin_lock_irqsave(&dev->lock, flags);

	// make sure it is currently queued on this endpoint
	
	list_for_each_entry(req, &ep->queue, queue) {
		if (&req->req == _req)
			break;		// found it
	}

	if (&req->req != _req) {
		DBG_MSG("%s: request to dequeue not found\n", __FUNCTION__);
		retval = -EINVAL;
		goto exit;
	}

	DBG_MSG("%s: request dequeue completed\n", __FUNCTION__);
	pdc_complete(ep, req, -ECONNRESET);  // complete request with error
	retval = 0;
exit:
	pdc_log_exit();

	spin_unlock_irqrestore(&dev->lock, flags);

	return retval;
}

static int pdc_ep_set_halt(struct usb_ep *_ep, int value)
{
	struct pdc_globals	*dev = &globals;
	struct pdc_ep		*ep = container_of(_ep, struct pdc_ep, ep);
	unsigned long		flags;
	int			retval = 0;

	pdc_log_entry("(%s,%d)\n", _ep->name, value);
	DBG_MSG("%s(%s, %d)\n", __FUNCTION__, _ep->name, value);

	if (unlikely(!_ep)) {
		DBG_MSG("%s: bad ep\n", __FUNCTION__);
		return -EINVAL;
	}
	
	spin_lock_irqsave(&dev->lock, flags);

	if (!list_empty(&ep->queue)) {
		DBG_MSG("halt on non-empty ep queue\n");
		retval = -EAGAIN;
		goto exit;
	}

	if (value) {	// set halt/stall

		// FIXME - check for stall of IN endpoint with queued requests
		// outstanding, or bytes left to be transmitted.  If this is
		// the case, return -EAGAIN.

		if (EP_DIRECTION_IS_IN(ep)) {
			pdc_write32(DC_EP_INDEX_REG, ep->ep_index);
			if (pdc_read16(DC_BUFFER_STATUS_REG) & 
			    (BUF0_FILLED | BUF1_FILLED)) {
			    	// if either (or both) buffers have data in
				// them, then return an error
				DBG_MSG("halt on non-empty ep\n");
				retval = -EAGAIN;
				goto exit;
			}
		}

		// FIXME - pdc_ep_fifo_flush(_ep); here ???

		pdc_locked_do_ep_set_halt(_ep);

	} else {	// clear halt/stall

		if (dev->cbw_error) {
			
			// don't clear the halt condition until we get
			// a bus reset
			DBG_MSG("Ignoring clear HALT request\n");

		} else {
		
			// FIXME - pdc_ep_fifo_flush(_ep); here ???

			pdc_locked_do_ep_clear_halt(_ep);

		}
		
	}
exit:
	spin_unlock_irqrestore(&dev->lock, flags);
	pdc_log_exit();
	return retval;
}

static int pdc_ep_fifo_status(struct usb_ep *_ep)
{
	pdc_log_entry("\n");
	DBG_MSG("%s: returning -EOPNOTSUPP\n", __FUNCTION__);
	pdc_log_exit();
	return -EOPNOTSUPP;
}

static void pdc_ep_fifo_flush(struct usb_ep *_ep)
{
	struct pdc_globals	*dev = &globals;
	unsigned long		flags;
	
	pdc_log_entry("(%s)\n", _ep->name);
	DBG_MSG("%s(%s)\n", __FUNCTION__, _ep->name);

	spin_lock_irqsave(&dev->lock, flags);

	pdc_locked_do_ep_fifo_flush(_ep);

	spin_unlock_irqrestore(&dev->lock, flags);
	pdc_log_exit();
}


/* -----  api related to the usb controller hardware  ----- */


static int pdc_get_frame(struct usb_gadget *_gadget)
{
	return (pdc_read16(DC_FRAME_NUMBER_REG) & 0x07FF);
}


static int pdc_wakeup(struct usb_gadget *_gadget)
{
	DBG_MSG("pdc_wakeup called\n");

	return -EOPNOTSUPP;
}


static int pdc_vbus_draw(struct usb_gadget *_gadget, unsigned mA)
{
	struct pdc_globals	*dev = &globals;
	unsigned long		flags;
	
	DBG_MSG("%s(%d)\n", __FUNCTION__, mA);

	spin_lock_irqsave(&dev->lock, flags);
	
	// Limit to discrete values: 100 or 500 mA

	if (mA > 100)
		mA = 500;
	else
		mA = 100;

	dev->configured_vbus_draw = mA;

	pdc_set_current_draw(dev, mA);

	spin_unlock_irqrestore(&dev->lock, flags);
	
	return 0;
}


/* "function" sysfs attribute */
static ssize_t
show_function (struct device *_dev, char *buf)
{
	struct pdc_globals	*dev = dev_get_drvdata (_dev);

	if (!dev->driver
			|| !dev->driver->function
			|| strlen (dev->driver->function) > PAGE_SIZE)
		return 0;
	return scnprintf (buf, PAGE_SIZE, "%s\n", dev->driver->function);
}
static DEVICE_ATTR (function, S_IRUGO, show_function, NULL);


/* "vbus" sysfs attribute */
static ssize_t
show_vbus (struct device *_dev, char *buf)
{
	u16	vbus_state;

	if (pdc_read16(DC_MODE_REG) & MODE_VBUSSTAT)
		vbus_state = 1;
	else
		vbus_state = 0;

	return scnprintf(buf, PAGE_SIZE, "%d\n", vbus_state);
}
static DEVICE_ATTR (vbus, S_IRUGO, show_vbus, NULL);


/* "configured" sysfs attribute */
static ssize_t
show_configured (struct device *_dev, char *buf)
{
	struct pdc_globals	*dev = dev_get_drvdata(_dev);
	int			configured = 0;

	if (dev->device_configuration)
		configured = 1;

	return scnprintf(buf, PAGE_SIZE, "%d\n", configured);
}
static DEVICE_ATTR (configured, S_IRUGO, show_configured, NULL);


#ifdef DEBUG_CHIP

/* "chip" sysfs attribute */
static ssize_t
show_chip (struct device *_dev, char *buf)
{
        int     i = 0;

        i += scnprintf(buf+i, PAGE_SIZE-i,
                "200 Address    %04X\n", pdc_read16(DC_ADDRESS_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "204 MaxPktSz   %04X\n", pdc_read16(DC_MAX_PKT_SIZE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "208 Ep Type    %04X\n", pdc_read16(DC_EP_TYPE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "20C Mode       %04X\n", pdc_read16(DC_MODE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "210 Intr Cfg   %04X\n", pdc_read16(DC_INTR_CONFIG_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "212 Debug      %04X\n", pdc_read16(DC_DEBUG_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "214 Int Enable %08X\n", pdc_read32(DC_INTR_ENABLE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "218 Interrupt  %08X\n", pdc_read32(DC_INTERRUPT_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "21C Buf Len    %04X\n", pdc_read16(DC_BUFFER_LENGTH_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "21E Buf Status %04X\n", pdc_read16(DC_BUFFER_STATUS_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "220 Data Port  xxxx\n");
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "228 Ctrl Func  %04X\n", pdc_read16(DC_CTRLFN_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "22C Ep Index   %04X\n", pdc_read16(DC_EP_INDEX_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "230 DMA Cmd    %04X\n", pdc_read16(DC_DMA_COMMAND_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "234 DMA Tfr Ct %08X\n", pdc_read32(DC_DMA_TRANSFER_CTR_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "238 DMA Config %04X\n", pdc_read16(DC_DMA_CONFIG_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "23C DMA Hw     %04X\n", pdc_read16(DC_DMA_HARDWARE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "250 DMA Int    %04X\n", pdc_read16(DC_DMA_INTR_REASON_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "254 DMA Int En %04X\n", pdc_read16(DC_DMA_INTR_ENABLE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "258 DMA Endpt  %04X\n", pdc_read16(DC_DMA_EP_INDEX_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "264 DMA Burst  %04X\n", pdc_read16(DC_DMA_BURST_CTR_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "270 Chip ID    %08X\n", pdc_read32(DC_CHIP_ID_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "274 Frame Nmbr %04X\n", pdc_read16(DC_FRAME_NUMBER_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "278 Scratch    %04X\n", pdc_read16(DC_SCRATCH_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "27C Unlk Dvc   %04X\n", pdc_read16(DC_UNLOCK_DEVICE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "284 Test Mode  %04X\n", pdc_read16(DC_TESTMODE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "300 Hw Mode Ct %08X\n", pdc_read32(HC_HW_MODE_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "374 OTG Ctrl   %08X\n", pdc_read32(OTG_CONTROL_REG));
        i += scnprintf(buf+i, PAGE_SIZE-i,
                "total: %d\n", i+8+3);
        return i;
}
DEVICE_ATTR(chip, S_IRUGO, show_chip, NULL);
#endif


/* "globals" sysfs attribute */
static ssize_t
show_globals(struct device *_dev, char *buf)
{
        struct pdc_globals      *dev = dev_get_drvdata(_dev);
        int i = 0;

        i += scnprintf(buf+i, PAGE_SIZE-i, "addr = %02X  ",
                        dev->device_address);
        i += scnprintf(buf+i, PAGE_SIZE-i, "cnfg = %02X  ",
                        dev->device_configuration);
        i += scnprintf(buf+i, PAGE_SIZE-i, "state = %02X [%s]\n",
                        dev->device_state, state_names[dev->device_state]);
	i += scnprintf(buf+i, PAGE_SIZE-i, "inten = %08X  inten_blk = %08X\n",
			dev->intr_enables, dev->intr_enables_bulk);
        return i;
}
DEVICE_ATTR(globals, S_IRUGO, show_globals, NULL);


/* "logging" sysfs attribute */
static ssize_t
show_logging (struct device *_dev, char *buf)
{
	struct pdc_globals	*dev = dev_get_drvdata(_dev);

	return scnprintf(buf, PAGE_SIZE, "%lu\n", dev->logging_flags);
}

static ssize_t
store_logging (struct device *_dev, const char *buf, size_t count)
{
	struct pdc_globals	*dev = dev_get_drvdata(_dev);
	unsigned long		value;
	
	if (sscanf(buf, "%lu", &value) > 0) {
		dev->logging_flags = value;
		return strlen(buf);
	}
	return -EINVAL;
}

static DEVICE_ATTR (logging, S_IRUGO|S_IWUSR, show_logging, store_logging);


/* when a driver is successfully registered, it will receive
 * control requests including set_configuration(), which enables
 * non-control requests.  then usb traffic follows until a
 * disconnect is reported.  then a host may connect again, or
 * the driver might get unbound.
 */
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	struct pdc_globals	*dev = &globals;
	int			retval;
	
	pdc_log_entry("\n");
	
#ifdef DEBUG_VERBOSE
	DBG_MSG("driver             = 0x%p\n", driver);
	DBG_MSG("driver->function   = \"%s\"\n", driver->function);
	DBG_MSG("driver->speed      = 0x%x\n", driver->speed);
	DBG_MSG("driver->bind       = 0x%p\n", driver->bind);
	DBG_MSG("driver->unbind     = 0x%p\n", driver->unbind);
	DBG_MSG("driver->setup      = 0x%p\n", driver->setup);
	DBG_MSG("driver->disconnect = 0x%p\n", driver->disconnect);
	DBG_MSG("driver->suspend    = 0x%p\n", driver->suspend);
	DBG_MSG("driver->resume     = 0x%p\n", driver->resume);
#endif

	if (!driver	|| driver->speed == USB_SPEED_UNKNOWN
			|| driver->speed == USB_SPEED_LOW
			|| !driver->bind
			|| !driver->unbind
			|| !driver->disconnect
			|| !driver->setup)
		return -EINVAL;
	if (!dev)
		return -ENODEV;
	if (dev->driver)
		return -EBUSY;

	/* first hook up the driver ... */
	dev->driver = driver;
	dev->gadget.dev.driver = &driver->driver;

	device_add (&dev->gadget.dev);
	
	retval = driver->bind(&dev->gadget);
	if (retval) {
		DBG_MSG("bind to driver %s --> error %d\n",
				driver->driver.name, retval);
		device_del (&dev->gadget.dev);

		dev->driver = 0;
		dev->gadget.dev.driver = 0;
		return retval;
	}
	
	device_create_file(dev->dev, &dev_attr_function);

	/* ... then enable host detection and ep0; and we're ready
	 * for set_configuration as well as eventual disconnect.
	 */
	DBG_MSG("registered gadget driver '%s'\n", driver->driver.name);
	pdc_enable(dev);
	dump_state(dev);
	
	pdc_log_exit();

	return 0;
}
EXPORT_SYMBOL(usb_gadget_register_driver);

static void
stop_activity(struct pdc_globals *dev, struct usb_gadget_driver *driver)
{
	int i;

	/* don't disconnect drivers more than once */
	if (dev->gadget.speed == USB_SPEED_UNKNOWN)
		driver = 0;
	dev->gadget.speed = USB_SPEED_UNKNOWN;

	/* prevent new request submissions, kill any outstanding requests  */
	for (i = 0; i < PDC_NUM_ENDPOINTS; i++) {
		struct pdc_ep *ep = &dev->ep[i];

		ep->stopped = 1;
		pdc_nuke(ep, -ESHUTDOWN);
	}

	del_timer_sync(&dev->timer);
	del_timer_sync(&dev->vbus_timer);

	/* report disconnect; the driver is already quiesced */
	if (driver && driver->disconnect) {
		spin_unlock(&dev->lock);
		driver->disconnect(&dev->gadget);
		spin_lock(&dev->lock);
	}

	/* re-init driver-visible data structures */
	pdc_reinit(dev);
}


static void pdc_nuke(struct pdc_ep *ep, int status)
{
	struct pdc_globals	*dev = &globals;
	struct pdc_request	*req;

	/* called with irqs blocked */

	while (!list_empty(&ep->queue)) {
		req = list_entry(ep->queue.next, struct pdc_request, queue);
		pdc_complete(ep, req, status);
	}
	if (ep->desc)
		pdc_disable_ep_interrupt(dev, ep);
}


int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
	struct pdc_globals	*dev = &globals;
	
	pdc_log_entry("\n");

	if (!dev)
		return -ENODEV;
	if (!driver || driver != dev->driver)
		return -EINVAL;

	local_irq_disable();
	pdc_disable(dev);
	stop_activity(dev, driver);
	local_irq_enable();

	driver->unbind(&dev->gadget);
	dev->driver = 0;

	device_remove_file(dev->dev, &dev_attr_function);

	device_del (&dev->gadget.dev);

	DBG_MSG("unregistered gadget driver '%s'\n", driver->driver.name);
	dump_state(dev);

	pdc_log_exit();

	return 0;
}
EXPORT_SYMBOL(usb_gadget_unregister_driver);



/* -----  Hardware access  ----- */


static u16 pdc_read16(u16 reg)
{
	//if (pdc->controller_suspended) DBG_MSG("r16(%x)\n", reg);
	return readw(pdc->base_address + reg);
}

static void pdc_write16(u16 reg, u16 data)
{
	//if (pdc->controller_suspended) DBG_MSG("w16(%x,%x)\n", reg, data);
	writew(data, pdc->base_address + reg);
}

static u32 pdc_read32(u16 reg)
{
	//if (pdc->controller_suspended) DBG_MSG("r32(%x)\n", reg);
	return readl(pdc->base_address + reg);
}

static void pdc_write32(u16 reg, u32 data)
{
	//if (pdc->controller_suspended) DBG_MSG("w32(%x,%x)\n", reg, data);
	writel(data, pdc->base_address + reg);
}


/* The "pdc_locked_..." routines must be called with dev->lock held and
 * local interrupts disabled.  These are mostly primitives used by the
 * exported gadget API entry points as well internal gadget code.  They
 * perform no locking, in order to prevent spinlock deadlocks.
 */

static void pdc_locked_do_ep_fifo_flush(struct usb_ep *_ep)
{
	struct pdc_ep	*ep = container_of(_ep, struct pdc_ep, ep);
	short		mode;
	
	if (EP_DIRECTION_IS_IN(ep)) {
		DBG_MSG("flushing IN ep %d\n", ep->ep_index);
		// point ep index at an unused ep
		pdc_write16(DC_EP_INDEX_REG, EP6TX);

		// turn on dma clock for a moment
		
		mode = pdc_read16(DC_MODE_REG);
		pdc_write16(DC_MODE_REG, mode | MODE_DMACLKON);
		mdelay(2);

		// point dma ep index at this ep
		pdc_write16(DC_DMA_EP_INDEX_REG, ep->ep_index);
		pdc_write16(DC_DMA_COMMAND_REG, DMA_CMD_CLEAR_BUFFER);
		
		pdc_write16(DC_DMA_EP_INDEX_REG, ep->ep_index);
		pdc_write16(DC_DMA_COMMAND_REG, DMA_CMD_CLEAR_BUFFER);

		// point dma ep index back at an unused ep
		pdc_write16(DC_DMA_EP_INDEX_REG, EP7TX);

		mdelay(1);
		pdc_write16(DC_MODE_REG, mode);
	} else {
		DBG_MSG("flushing OUT ep %d\n", ep->ep_index);
		// make sure dma ep index points at an unused ep
		pdc_write16(DC_DMA_EP_INDEX_REG, EP7TX);
		
		pdc_write16(DC_EP_INDEX_REG, ep->ep_index);
		pdc_write16(DC_CTRLFN_REG, CNTRLFN_CLBUF);
		
		pdc_write16(DC_EP_INDEX_REG, ep->ep_index);
		pdc_write16(DC_CTRLFN_REG, CNTRLFN_CLBUF);
	}
}


static void pdc_locked_do_ep_clear_halt(struct usb_ep *_ep)
{
	struct pdc_ep	*ep = container_of(_ep, struct pdc_ep, ep);
	u32		ep_type, scratch;

	DBG_MSG("%s(%s)\n", __FUNCTION__, ep->ep.name);

	pdc_write32(DC_EP_INDEX_REG, ep->ep_index);
	scratch = pdc_read32(DC_CTRLFN_REG);
	pdc_write32(DC_CTRLFN_REG, 0);

	if ((ep->ep_index == EP0RX) || (ep->ep_index == EP0TX)) {
		pdc_write32(DC_EP_INDEX_REG, ep->ep_index ^ 1);
		pdc_write32(DC_CTRLFN_REG, 0);
	}

	if (ep->ep_index > EP0TX) {
		if (EP_DIRECTION_IS_IN(ep)) {
			//pdc_write32(DC_EP_INDEX_REG, ep->ep_index);
			//ep_type = pdc_read32(DC_CTRLFN_REG);
			//pdc_write32(DC_CTRLFN_REG, 0);

			// disable and enable ep to reset DATA PID

			ep_type = pdc_read32(DC_EP_TYPE_REG);
			scratch = ep_type & ~EP_TYP_ENABLE;
			pdc_write32(DC_EP_TYPE_REG, scratch);
			pdc_write32(DC_EP_TYPE_REG, ep_type);
			
			ep_type = pdc_read32(DC_EP_TYPE_REG);

			//FIXME tx_data(pdc,handle,2);<= kicks cntrlr somehow...
		} else {
			pdc_write32(DC_EP_INDEX_REG, ep->ep_index);
			pdc_write32(DC_CTRLFN_REG, 0);

			// disable and enable ep to reset DATA PID

			ep_type = pdc_read32(DC_EP_TYPE_REG);
			scratch = ep_type & ~EP_TYP_ENABLE;
			pdc_write32(DC_EP_TYPE_REG, scratch);
			pdc_write32(DC_EP_TYPE_REG, ep_type);
			
			//rx_data(pdc,handle); <= kicks cntrlr somehow...
		}
	}
	
	ep->stopped = 0;	
}


static void pdc_locked_do_ep_set_halt(struct usb_ep *_ep)
{
	struct pdc_ep	*ep = container_of(_ep, struct pdc_ep, ep);

	DBG_VMSG("%s(%s)\n", __FUNCTION__, ep->ep.name);

	pdc_write32(DC_EP_INDEX_REG, ep->ep_index);
	pdc_write32(DC_CTRLFN_REG, CNTRLFN_STALL);
		
	// stall both sides of endpt 0

	if ((ep->ep_index == EP0RX) || (ep->ep_index == EP0TX)) {
		pdc_write32(DC_EP_INDEX_REG, ep->ep_index ^ 1);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_STALL);
	}

	ep->stopped = 1;
}


/*
	There are three acceptable values that can be passed to this routine:
	
		1	- bus suspend, disconnect, or module unload
		100	- default current draw (or configured current draw)
		500	- high-power configured current draw (if enabled)

	In addition, there is some additional logic to avoid bouncing between
	suspend and normal current levels, since we often get spurious VBUS
	interrupts mixed in with other interrupts.  To work around this, we
	keep track of the current power level to avoid unnecessary changes.
*/
		
#define GPIO_USBHP	55
#define GPIO_SUSPEND	56

static void pdc_set_current_draw(struct pdc_globals *dev, u32 mA)
{
	DBG_MSG("%s(%d)\n", __FUNCTION__, mA);
	
	if ((mA != 1) && (mA != 100) && (mA != 500)) {
		DBG_PRINTK("bogus current draw value: %d\n", mA);
		mA = 1;
	}

	// If it ain't broke, don't fix it
	
	if (mA == dev->current_vbus_draw)
		return;
	
	dev->current_vbus_draw = mA;	// remember new vbus draw level

	if (mA == 1) {
		// set SUSPEND gpio to stop drawing current, USBHP unchanged
		DBG_VMSG("drawing   1mA\n");
		GPSR(GPIO_SUSPEND) = GPIO_bit(GPIO_SUSPEND);
		return;
	}

	if (mA == 500) {
		// set USBHP gpio and make sure SUSPEND gpio is cleared
		DBG_VMSG("drawing 500mA\n");
		GPSR(GPIO_USBHP) = GPIO_bit(GPIO_USBHP);
		GPCR(GPIO_SUSPEND) = GPIO_bit(GPIO_SUSPEND);
	} else {	
		// clear USBHP and make sure SUSPEND is cleared
		DBG_VMSG("drawing 100mA\n");
		GPCR(GPIO_USBHP) = GPIO_bit(GPIO_USBHP);
		GPCR(GPIO_SUSPEND) = GPIO_bit(GPIO_SUSPEND);
	}
}


/* Called by pdc_probe and bus reset handler */

void pdc_init(struct pdc_globals *dev)
{
	u32	temp1;
	u16	mode;
	
	pdc_log_entry("\n");

	/* set the ISP1761 Device controller Hardware Configuration */
	
	/* Set DMA mode (no dma) */
	//pdc_write32(CMD_WRITEDMACONFIG,0);

	// Point DMA Endpoint register at an unused endpoint
	
	mode = pdc_read16(DC_MODE_REG);
	pdc_write16(DC_MODE_REG, mode | MODE_DMACLKON);
	mdelay(2);
	
	pdc_write16(DC_DMA_EP_INDEX_REG, EP7RX);	// Endpt 7 unused

	mdelay(1);
	pdc_write16(DC_MODE_REG, mode);
	
	/*Disable all interrupts */
	DBG_VMSG("-- disabling all interrupts --\n");
	dev->intr_enables = 0;
	pdc_write32(DC_INTR_ENABLE_REG, dev->intr_enables);
	
	/* Set default address */

	pdc_write32(DC_ADDRESS_REG, 0);
		
	DBG_VMSG("chip id = 0x%08X \n", pdc_read32(DC_CHIP_ID_REG));
	pdc_write16(DC_SCRATCH_REG, 0xa5a5);
	temp1 = pdc_read16(DC_SCRATCH_REG);
	DBG_VMSG("scratch 0xa5a5 = 0x%08X \n", temp1);
	
	pdc_log_exit();
	
	return;
}


/* Actually connect to the bus */
void pdc_connect(struct pdc_globals *dev)
{
	pdc_log_entry("\n");
	
	/* Set device address to zero and enable */
	pdc_write32(DC_ADDRESS_REG, ADDRESS_DEVEN | 0);


	// // wr32(374, 0x00000081) // //
	// // wr32(374, 0x00160000) // //
	// // wr32(MODE_REG, WKUPCS | CLKAON | GLINTENA) // //
	
	
	// Default pulse width is 1us (0x1E).  (Minimum pulse is 33.3ns)
	pdc_write32(DC_INTR_PULSE_SIZE_REG, DEFAULT_INT_PULSE_SIZE);

	// paul set int config reg
	pdc_write16(DC_INTR_CONFIG_REG, DEFAULT_INT_CONFIG);
	
	/* Enable interrupts */
	DBG_VMSG("++ enabling all interrupts ++\n");
	dev->intr_enables = IE_BASIC_CTRL_EP_MASK;
	pdc_write32(DC_INTR_ENABLE_REG, dev->intr_enables);

	/* Connect to the bus */
	pdc_write16(DC_MODE_REG, MODE_GLINTENA);

	pdc_dump();
	
	pdc_log_exit();
	
	return;
}


static void pdc_enable (struct pdc_globals *dev)
{
	unsigned long	flags;
	
	spin_lock_irqsave(&dev->lock, flags);
	
	pdc_log_entry("\n");
	
	//ep0_idle(dev);
	
	dev->gadget.speed = USB_SPEED_UNKNOWN;
	
	pdc_locked_go_to_state(STATE_DETACHED, 0, 0);
	
	//dev->stats.irqs = 0;
	
	/* enable suspend/resume and reset irqs */
	
	/* enable ep0 irqs */
	
	/* if hardware supports it, connect to usb and wait for host */


	pdc_connect(dev);

	pdc_locked_go_to_state(STATE_ATTACHED, 0, 0);

	pdc_log_exit();

	spin_unlock_irqrestore(&dev->lock, flags);
}


/*
 * 	pdc_disable - disable USB device controller
 */
static void pdc_disable(struct pdc_globals *dev)
{
	unsigned long	flags;

	spin_lock_irqsave(&dev->lock, flags);
	
	pdc_log_entry("\n");
	
	// FIXME - Probably more needs to be done here...

	
	pdc_locked_go_to_state(STATE_DETACHED, 0, 0);
	
	pdc_log_exit();

	spin_unlock_irqrestore(&dev->lock, flags);
}


static void pdc_reinit(struct pdc_globals *dev)
{
	u32	i;

	pdc_log_entry("\n");
	
	/* device/ep0 records init */
	INIT_LIST_HEAD (&dev->gadget.ep_list);
	INIT_LIST_HEAD (&dev->gadget.ep0->ep_list);
//	dev->ep0state = EP0_IDLE;

	/* basic endpoint records init */
	for (i = 0; i < PDC_NUM_ENDPOINTS; i++) {
		struct pdc_ep *ep = &dev->ep[i];

		if (i != 0)
			list_add_tail (&ep->ep.ep_list, &dev->gadget.ep_list);

		ep->desc = 0;
		ep->stopped = 0;
		INIT_LIST_HEAD (&ep->queue);
		//ep->pio_irqs = ep->dma_irqs = 0;
		ep->irqs = 0;
	}

	/* the rest was statically initialized, and is read-only */
	
	pdc_log_exit();
}



/* -----  struct device support functions  ----- */


static void nop_release (struct device *dev)
{
	pdc_log_entry("\n");
	pdc_log_exit();
}


static void  dump_state(struct pdc_globals *dev)
{
}


static void pdc_watchdog(unsigned long _dev)
{
#if 0
	struct pxa2xx_udc	*dev = (void *)_dev;

	local_irq_disable();
	if (dev->ep0state == EP0_STALL
			&& (UDCCS0 & UDCCS0_FST) == 0
			&& (UDCCS0 & UDCCS0_SST) == 0) {
		UDCCS0 = UDCCS0_FST|UDCCS0_FTF;
		DBG(DBG_VERBOSE, "ep0 re-stall\n");
		start_watchdog(dev);
	}
	local_irq_enable();
#endif
}


#if 0
static void pdc_dump_stats(struct pdc_globals *dev)
{
	DBG_MSG("jiffies = 0x%x\n", jiffies);

}
#endif


// Called with dev->lock held.
// On exit, global interrupt is disabled.

static void pdc_wake_controller(void)
{
	u32	m;

	DBG_MSG("%s\n", __FUNCTION__);

	// issue a read to make sure controller is awake
	
	m = pdc_read32(DC_MODE_REG);

	mdelay(3);

	pdc_write32(DC_UNLOCK_DEVICE_REG, 0xAA37);
	
	m = pdc_read32(DC_MODE_REG);	// dummy read

	pdc_write32(DC_MODE_REG, MODE_SFRESET);
	mdelay(2);
	pdc_write32(DC_MODE_REG, 0);
	mdelay(5);
}


static void pdc_suspend_controller(int force_off)
{
	struct pdc_globals	*dev = &globals;
	unsigned long		flags;
	u32			m;

	// If we should never suspend, just return

	if (!dev->suspend_enabled) {
		DBG_MSG("%s: suspend disabled!\n", __FUNCTION__);
		return;
	}
		
	// Is it already suspended?

	if (dev->controller_suspended)
		DBG_MSG("%s: already suspended\n", __FUNCTION__);
	
	spin_lock_irqsave(&dev->lock, flags);
	DBG_MSG("%s(%s)\n", __FUNCTION__, (force_off ? "off" : "sleep"));
	
	// make sure controller is awake
	
	pdc_wake_controller();

	// restore interrupt enables
	
	if (force_off) {

		if (dev->device_state != STATE_DETACHED) {
			// We're currently connected at some level over USB

			DBG_PRINTK("## Forcing USB disconnect ##\n");

			pdc_locked_go_to_state(STATE_DETACHED, 0, 0);

			// signal higher-level software like VBUS is gone

			if (dev->gadget.speed != USB_SPEED_UNKNOWN &&
			    dev->driver && dev->driver->disconnect) {
			    	spin_unlock(&dev->lock);
				dev->driver->disconnect(&dev->gadget);
				spin_lock(&dev->lock);
				dev->gadget.speed = USB_SPEED_UNKNOWN;
			}

			// Mask interrupts

			pdc_write32(DC_INTR_ENABLE_REG, 0);
			pdc_write32(DC_MODE_REG, 0);

			// Disable pdc

			pdc_write32(DC_ADDRESS_REG, 0);
			
			// Disconnect D+ pullup resistor
#ifndef CONFIG_USB_OTG
			pdc_write32(OTG_CONTROL_REG, 0x00010006);
#endif

			// Soft-reset the device controller

			pdc_write32(DC_MODE_REG, MODE_SFRESET);
			mdelay(2);
			pdc_write32(DC_MODE_REG, 0);
			mdelay(5);

			//DBG_PRINTK("Sending USB_OUT event\n");
			//handle_fiona_event(USB_OUT_EVENTCODE);
		}
		
		// Don't wake on anything.  When the power switch is turned
		// back on, we'll set up to wake up on VBUS or SUSPEND intrs.

		pdc_write32(DC_INTR_ENABLE_REG, 0);
		pdc_write32(DC_MODE_REG, 0);

	} else {
		// wake on VBUS detect or SUSPEND
		// (work around for two-plug wake problem - VBUS interrupt
		//  restarts clocks, but doesn't generate an interrupt to
		//  the processor, but SUSPEND interrupt gets through to the
		//  processor.)

		pdc_write32(DC_INTR_ENABLE_REG, IE_VBUS | IE_SUSP);
		pdc_write32(DC_MODE_REG, MODE_GLINTENA);
	}

	// suspend the controller

 	m = pdc_read32(DC_MODE_REG) | MODE_WKUPCS;
	pdc_write32(DC_MODE_REG, m | MODE_GOSUSP);
	pdc_write32(DC_MODE_REG, m);

	dev->controller_suspended = 1;

	spin_unlock_irqrestore(&dev->lock, flags);
}


static void pdc_vbus_timer_proc(unsigned long _dev)
{
	struct pdc_globals	*dev = (struct pdc_globals *) _dev;
	unsigned long		flags;
#if 0
	int			i;
#endif

	DBG_VMSG("%s: fired\n", __FUNCTION__);

	spin_lock_irqsave(&dev->lock, flags);
	
	if (pdc_vbus_state()) {
	
#if 0
		if (!dev->stats_counter) {
			dev->stats_counter = 4;
			pdc_dump_stats(dev);
		} else
			dev->stats_counter--;
#else
		if (!list_empty(&dev->ep[1].queue)) {
			DBG_MSG("ep1 queue non-empty at %08X\n", OSCR);
			pdc_dump_ep(&dev->ep[1]);
		}
#endif
		
		// check vbus again in a little while
		
		mod_timer(&dev->vbus_timer, jiffies + VBUS_TIMER_INTERVAL);

	} else {

		// VBUS is gone, so either we've been detached,
		// or the host has gone to sleep w/ VBUS off

		DBG_PRINTK("## USB disconnected ##\n");

		// Disconnect D+ pullup resistor
		
#ifndef CONFIG_USB_OTG
		pdc_write32(OTG_CONTROL_REG, 0x00170000);
		pdc_write32(OTG_CONTROL_REG, 0x00000080);
#endif

		pdc_locked_go_to_state(STATE_DETACHED, 0, 0);
		
		pdc_write32(DC_ADDRESS_REG, ADDRESS_DEVEN | 0);
#if 0
		/* prevent new request submissions, kill outstanding requests */
		for (i = 0; i < PDC_NUM_ENDPOINTS; i++) {
			struct pdc_ep *ep = &dev->ep[i];

			ep->stopped = 1;
			pdc_nuke(ep, -ECONNRESET);
		}
#endif

		// signal higher-level software that VBUS is gone

		if (dev->gadget.speed != USB_SPEED_UNKNOWN &&
		    dev->driver && dev->driver->disconnect) {
		    	spin_unlock(&dev->lock);
			dev->driver->disconnect(&dev->gadget);
			spin_lock(&dev->lock);
			dev->gadget.speed = USB_SPEED_UNKNOWN;
		}

		DBG_PRINTK("Sending USB_OUT event\n");
		handle_fiona_event(USB_OUT_EVENTCODE);
	}
	
	spin_unlock_irqrestore(&dev->lock, flags);
}


static void pdc_handle_ep_in_intr(struct pdc_globals *pdc, struct pdc_ep *ep)
{
	struct pdc_request	*req;
	struct usb_request	*_req;
	u16			remaining, length;
	
	/* An IN packet may have been received before we have any data to
	   transmit, so we need to check for an empty I/O queue.  If the
	   queue is empty, the pdc_ep_queue code will start transmitting the
	   data immediately, which will satisfy the interrupt.
	 */
	
	if (list_empty(&ep->queue)) {
		DBG_VMSG("%s: %s request queue is empty!\n", __FUNCTION__,
				ep->ep.name);
		goto exit;
	}

	req = container_of(ep->queue.next, struct pdc_request, queue);
	_req = &req->req;
	
	remaining = _req->length - _req->actual;
	
	if (remaining == 0) {	// transfer completed
		DBG_VMSG("%s: request completed\n", __FUNCTION__);

		pdc_complete(ep, req, 0);

		if (!list_empty(&ep->queue)) {
			req = container_of(ep->queue.next, struct pdc_request,
					queue);
			_req = &req->req;
					
			length = min((u16)_req->length, (u16)ep->ep.maxpacket);

			pdc_write_ep(ep->ep_index, _req->buf, length);

			_req->actual += length;

			ep->needs_kick = 0;	// make sure flag is clear
		}
		goto exit;
	}
	
	length = min(remaining, (u16) ep->ep.maxpacket);
	
	pdc_write_ep(ep->ep_index, _req->buf + _req->actual, length);

	_req->actual += length;
exit:
	return;
}
				
static void pdc_handle_ep_out_intr(struct pdc_globals *pdc, struct pdc_ep *ep)
{
	struct pdc_request	*req;
	struct usb_request	*_req;
	u16			remaining, length;
	u16			received;
	
	/* An OUT packet may have been received before a request was posted
	   to handle it.  If the request queue is empty, the pdc_ep_queue code
	   will attempt to read data immediately, which will satisfy this
	   interrupt.
	 */
	
	if (list_empty(&ep->queue)) {
		DBG_VMSG("%s: %s request queue is empty!\n", __FUNCTION__,
				ep->ep.name);
		goto exit;
	}

	req = container_of(ep->queue.next, struct pdc_request, queue);
	_req = &req->req;
	
	remaining = _req->length - _req->actual;
	
	if (remaining == 0) {	// transfer completed
		DBG_VMSG("%s: request completed\n", __FUNCTION__);

		pdc_complete(ep, req, 0);

		if (list_empty(&ep->queue))
			goto exit;
		
		// try to start up the next request
		
		req = container_of(ep->queue.next, struct pdc_request, queue);
		_req = &req->req;
		remaining = _req->length - _req->actual;

		// fall through to request handling code...
	}
	
	length = min(remaining, (u16) ep->ep.maxpacket);
	
	received = pdc_read_ep(ep->ep_index, _req->buf + _req->actual, length);
	_req->actual += received;
	
	if (received &&
	    ((received < (u16) ep->ep.maxpacket) ||
	     (_req->actual == _req->length))) {		// transfer completed
		DBG_VMSG("%s: request completed\n", __FUNCTION__);

		pdc_complete(ep, req, 0);
	}

exit:	
	return;
}


// Called by main interrupt handler - device spinlock is held

static void pdc_handle_ep0setup_intr(struct pdc_globals *dev)
{
	u8			command[8];
	u8			response[4];
	u32			*buf;
	struct usb_ctrlrequest	*ctrl;
	u16			len, addr;
	u16			ep_num;
	u8			new_state, new_config;
	u8			call_gadget_driver = 0;
	u8			post_event;
	u8			mode;
	int			rc = 0;

	// If ep0 was stalled, clear the stall condition

	if (dev->ep[0].stopped) {
		DBG_MSG("clearing ep0 STALL condition\n");
		pdc_locked_do_ep_clear_halt(&dev->ep[0].ep);
	}
		
	// Read SETUP packet
	
	pdc_write32(DC_EP_INDEX_REG, EP0SETUP);
	len = pdc_read16(DC_BUFFER_LENGTH_REG);
	if (len != 8) {
		pdc_write32(DC_EP_INDEX_REG, EP0TX);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_STALL);
		pdc_write32(DC_EP_INDEX_REG, EP0RX);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_STALL);
		DBG_MSG("Short SETUP packet! (%d bytes)\n", len);
		goto exit;
	}
	
	buf = (u32 *) command;
	buf[0] = pdc_read32(DC_DATAPORT_REG);
	buf[1] = pdc_read32(DC_DATAPORT_REG);
	
	DBG_TRACE("SETUP pkt: %02X %02X %02X %02X "
			"%02X %02X %02X %02X\n", command[0],
			command[1], command[2], command[3],
			command[4], command[5], command[6],
			command[7]);
	
	ctrl = (struct usb_ctrlrequest *) command;

	switch (ctrl->bRequest) {
	
	case USB_REQ_SET_ADDRESS:
	
		if ((ctrl->bRequestType != (USB_TYPE_STANDARD | USB_RECIP_DEVICE)) ||
		    (ctrl->wIndex) || (ctrl->wLength) ||
		    (ctrl->wValue & ~0x7F) ) {
			DBG_MSG("Bad SET_ADDRESS request\n");
			goto bad_request;
		}

		addr = ctrl->wValue & 0x007F;

		DBG_MSG("SET_ADDRESS %d\n", addr);
		
		pdc_locked_go_to_state(STATE_ADDRESS, addr, 0);
		
		addr |= ADDRESS_DEVEN;	// set "device enable" bit"
		pdc_write16(DC_ADDRESS_REG, addr);
		
		pdc_write32(DC_EP_INDEX_REG, EP0TX);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_STATUS_EM_ACK);

		break;
	
	case USB_REQ_SET_FEATURE:

		if ((ctrl->bRequestType == USB_DIR_OUT) &&
		    (ctrl->wValue == USB_DEVICE_REMOTE_WAKEUP) &&
		    (ctrl->wIndex == 0) && (ctrl->wLength == 0)) {

			DBG_MSG("SET_FEATURE DEVICE REMOTE_WAKEUP\n");
			
			// remote_wakeup is ready to be bit-or'ed into the
			// GET_STATUS response byte

			dev->remote_wakeup = (1 << USB_DEVICE_REMOTE_WAKEUP);
			
		} else if ((ctrl->bRequestType == USB_RECIP_ENDPOINT) &&
			   (ctrl->wValue == USB_ENDPOINT_HALT) &&
			   (ctrl->wIndex == 0x0081 || ctrl->wIndex == 0x0002)
			   && (ctrl->wLength == 0)) {
			   
			DBG_MSG("SET_FEATURE ENDPOINT STALL\n");

			ep_num = ctrl->wIndex & 0x000F;

			pdc_locked_do_ep_set_halt(&dev->ep[ep_num].ep);

		} else if ((ctrl->bRequestType == USB_DIR_OUT) &&
			   (ctrl->wValue == 0x0002) &&
			   ((ctrl->wIndex & 0x00FF) == 0) &&
			   (ctrl->wLength == 0) ) {

			u32	i, testmode = 0;
			u32	*tpp = (u32 *) pdc_test_pkt;
			
			DBG_MSG("SET_FEATURE TEST_MODE\n");

			mode = (ctrl->wIndex >> 8) & 0x00FF;
			
			switch (mode) {
			case 0x01:	// Test J
			    DBG_MSG("TEST_J\n");
			    // Foxconn code or's in 0xFF00 with the lower
			    // bits - may be superstition, or a necessary
			    // trick...
			    	testmode = TESTMODE_FORCEHS | TESTMODE_JSTATE;
				break;
			case 0x02:	// Test K
				DBG_MSG("TEST_K\n");
				testmode = TESTMODE_FORCEHS | TESTMODE_KSTATE;
				break;
			case 0x03:	// Test SE0_NAK
				DBG_MSG("TEST_SE0_NAK\n");
				testmode = TESTMODE_FORCEHS | TESTMODE_SE0_NAK;
				break;
			case 0x04:	// Test Packet
				DBG_MSG("TEST_PACKET\n");
				// Foxconn also writes 52 bytes of junk into
				// the control endpoint - may be necessary, or
				// just cruft...
				pdc_write32(DC_EP_INDEX_REG, EP0TX);
				pdc_write32(DC_BUFFER_LENGTH_REG,
						sizeof(pdc_test_pkt));
				for (i = 0 ; i < sizeof(pdc_test_pkt) ; i += 4)
				    pdc_write32(DC_DATAPORT_REG, *tpp++);
				testmode = TESTMODE_FORCEHS | TESTMODE_PRBS;
				break;
			case 0x05:	// Test Force Enable
				DBG_MSG("TEST_FORCE_ENABLE not supported\n");
				break;
			default:
				DBG_MSG("TEST_MODE %d not supported\n", mode);
				break;
			}

			// Okay, this is a little convoluted - if a valid test
			// mode was requested, we ACK the request first, give
			// it a little to time to get back to the host, then
			// switch into the test mode.

			if (testmode) {
			    // ACK the request
			    pdc_write32(DC_EP_INDEX_REG, EP0TX);
			    pdc_write32(DC_CTRLFN_REG, CNTRLFN_STATUS_EM_ACK);
			    // allow the ACK to get back to the host
			    mdelay(1);
			    // enter the test mode
			    pdc_write32(DC_TESTMODE_REG, testmode);
			    // break out so we don't try to send an extra ACK
			    break;
			}
		} else {
			DBG_MSG("Bad SET_FEATURE request\n");
			goto bad_request;
		}
		
		pdc_write32(DC_EP_INDEX_REG, EP0TX);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_STATUS_EM_ACK);

		break;

	case USB_REQ_CLEAR_FEATURE:

		ep_num = 86;	// if altered, we need to check if an endpoint
				// needs to be kickstarted
		
		if ((ctrl->bRequestType == USB_DIR_OUT) &&
		    (ctrl->wValue == USB_DEVICE_REMOTE_WAKEUP) &&
		    (ctrl->wIndex == 0) && (ctrl->wLength == 0)) {
			
			DBG_MSG("CLEAR_FEATURE DEVICE REMOTE_WAKEUP\n");

			dev->remote_wakeup = 0;

		} else if ((ctrl->bRequestType == USB_RECIP_ENDPOINT) &&
			   (ctrl->wValue == USB_ENDPOINT_HALT) &&
			   (ctrl->wIndex == 0x0081 || ctrl->wIndex == 0x0002)
			   && (ctrl->wLength == 0) ) {

			DBG_MSG("CLEAR_FEATURE ENDPOINT STALL\n");

			ep_num = ctrl->wIndex & 0x000F;

			if (dev->cbw_error) {
			    if (dev->cbw_error == 1)
				DBG_MSG("Ignoring CLEAR_FEATURE EP STALL\n");
			    dev->cbw_error |= 2;
			} else {
			    pdc_locked_do_ep_clear_halt(&dev->ep[ep_num].ep);
			}

		} else {
			DBG_MSG("Bad CLEAR_FEATURE ENDPOINT_HALT request\n");
			goto bad_request;
		}
		
		pdc_write32(DC_EP_INDEX_REG, EP0TX);
		pdc_write32(DC_CTRLFN_REG, CNTRLFN_STATUS_EM_ACK);

		// We may have received a request for ep1 (typically a CSW)
		// while the endpt was stalled.  If so, it was simply queued
		// up, but not started.  We need to start it now.

		if ((ep_num != 86) && (dev->ep[ep_num].needs_kick)) {
			struct pdc_ep		*ep = &dev->ep[ep_num];
			struct pdc_request	*req;
			u16			len;

			DBG_VMSG("handling needs_kick\n");

			ep->needs_kick = 0;
			
			req = container_of(ep->queue.next, struct pdc_request,
					queue);

			if (!req) {
#ifdef DEBUG
				BUG();
#else
				break;
#endif
			}

			len = min((u16) req->req.length,(u16) ep->ep.maxpacket);

			pdc_write_ep(ep->ep_index, req->req.buf, len);

			req->req.actual += len;
			
		}
		
		break;
	
	case USB_REQ_GET_STATUS:

		if ((ctrl->bRequestType & USB_DIR_IN) != USB_DIR_IN) {
			DBG_MSG("Bad GET_STATUS direction\n");
			goto bad_request;
		}

		if ((ctrl->bRequestType & USB_TYPE_MASK) != USB_TYPE_STANDARD){
			DBG_MSG("Bad GET_STATUS type\n");
			goto bad_request;
		}

		if (ctrl->wValue) {
			DBG_MSG("Bad GET_STATUS wValue\n");
			goto bad_request;
		}
		
		if (ctrl->wLength != 0x0002) {
			DBG_MSG("Bad GET_STATUS wLength\n");
			goto bad_request;
		}
		
		response[0] = 0x00;	// Standard response
		response[1] = 0x00;

		switch (ctrl->bRequestType & USB_RECIP_MASK) {
		    case USB_RECIP_DEVICE:
			if (ctrl->wIndex == 0) {
				DBG_MSG("GET_STATUS DEVICE\n");
				// indicate that we're self-powered
				response[0] = (1 << USB_DEVICE_SELF_POWERED);
				// reflect remote wakeup setting
				response[0] |= dev->remote_wakeup;

			} else {
				DBG_MSG("Bad wIndex: %d\n", ctrl->wIndex);
				goto bad_request;
			}
			break;	
		    case USB_RECIP_INTERFACE:
		    	DBG_MSG("GET_STATUS INTERFACE %d\n", ctrl->wIndex);
			break;
		    case USB_RECIP_ENDPOINT:
			if (ctrl->wIndex == 0x0081)
				ep_num = 1;
			else if (ctrl->wIndex == 0x0002)
				ep_num = 2;
			else {
				DBG_MSG("Bad GET_STATUS endpoint: %d\n",
						ctrl->wIndex & 0x0F);
				goto bad_request;
			}
		
			DBG_MSG("GET_STATUS EP %d\n", ep_num);

			if (dev->ep[ep_num].stopped)
				response[0] = 0x01;
			break;

		    default:
		    	DBG_MSG("Bad GET_STATUS recipient\n");
			goto bad_request;
			break;
		}
		pdc_write_ep(dev->ep[0].ep_index, response, 2);

		break;

	case USB_REQ_SET_CONFIGURATION:

		if ((ctrl->bRequestType != 0x00) ||
		    (ctrl->wIndex) || (ctrl->wLength) ) {
			DBG_MSG("Bad SET CONFIG request\n");
			goto bad_request;
		}

		DBG_MSG("SET_CONFIGURATION %d\n", ctrl->wValue);
		
		post_event = 0;

		if (ctrl->wValue == 0) {
			// device being unconfigured
			new_state = STATE_ADDRESS;
			new_config = 0;
		} else if (ctrl->wValue == 1) {
			// device being configured
			DBG_PRINTK("## USB configured ##\n");
			new_state = STATE_CONFIGURED;
			new_config = 1;
			post_event = 1;
		} else {
			// bogus configuration value = no change
			DBG_MSG("Bad config value %d\n", ctrl->wValue);
			new_state = dev->device_state;
			new_config = dev->device_configuration;
		}

		pdc_locked_go_to_state(new_state,
				dev->device_address, new_config);

		if (post_event) {
			DBG_PRINTK("Sending USB_IN event\n");
			handle_fiona_event(USB_IN_EVENTCODE);
		}

		/* When we get configured, we need to enable the bulk endpt
		   interrupts.  On the first connection, the gadget client
		   hasn't enabled any bulk endpts yet, so this code has no
		   effect.  When the endpts are enabled, the additional intr
		   enable bits will get set at that time, and we'll note what
		   bits will need to be enabled on subsequent SET_CONFIGURATION
		   events.  When the endpts are disabled, the corresponding
		   interrupt enable bits in the controller and our notes will
		   get cleared.						*/

		if (dev->device_state == STATE_CONFIGURED) {
			u32	int_enables = pdc_read32(DC_INTR_ENABLE_REG);

			int_enables |= dev->intr_enables_bulk;
			pdc_write32(DC_INTR_ENABLE_REG, int_enables);
			DBG_MSG("new intr enables = 0x%08X\n",
					pdc_read32(DC_INTR_ENABLE_REG));
		}

		call_gadget_driver = 1;

		break;
	
	case USB_BULK_RESET_REQUEST:

		// clear flag, so that CLEAR_FEATURE(ENDPT_HALT) will start
		// working again.

		DBG_MSG("USB Bulk-Only Mass Storage Reset received\n");

		dev->cbw_error = 0;
		call_gadget_driver = 1;

		break;

	default:

		call_gadget_driver = 1;
		
		break;
	}

	if (call_gadget_driver) {
		
		// pass through to gadget driver

		if (dev->driver && dev->driver->setup) {
			spin_unlock(&dev->lock);
			rc = dev->driver->setup(&dev->gadget,
					(struct usb_ctrlrequest *) command);
			spin_lock(&dev->lock);
		}

		if (rc < 0) {
			DBG_MSG("SETUP pkt: %02X %02X %02X %02X "
				"%02X %02X %02X %02X\n", command[0],
				command[1], command[2], command[3],
				command[4], command[5], command[6],
				command[7]);

			DBG_MSG("Gadget driver returned rc = %d\n", rc);

			// generate STALL
			pdc_locked_do_ep_set_halt(&dev->ep[0].ep);

			// start watchdog
		}
	}
	
exit:
	return;

bad_request:
	{
		DBG_MSG("SETUP pkt: %02X %02X %02X %02X "
			"%02X %02X %02X %02X\n", command[0],
			command[1], command[2], command[3],
			command[4], command[5], command[6],
			command[7]);

		// FIXME - Clear all previous transfers on the IN endpt
		//	   (Terminate every I/O request with an error.)
	
		if (dev->driver && dev->driver->setup) {
			spin_unlock(&dev->lock);
			rc = dev->driver->setup(&dev->gadget,
					(struct usb_ctrlrequest *) command);
			spin_lock(&dev->lock);
		}
	
		if (rc < 0) {
			
			DBG_MSG("Protocol STALL - err %d\n", rc);
			
			// generate STALL
			pdc_locked_do_ep_set_halt(&dev->ep[0].ep);
			
			// start watchdog
		}

		return;
	}
}


static void pdc_handle_ep0tx_intr(struct pdc_globals *dev)
{
	struct pdc_request	*req;
	struct usb_request	*_req;
	struct pdc_ep		*ep = &dev->ep[0];
	u16			remaining;
	u16			length;
	
	if (list_empty(&ep->queue)) {
		DBG_MSG("%s: ep request queue is empty!\n", __FUNCTION__);
		return;
	}
	
	req = container_of(ep->queue.next, struct pdc_request, queue);
	_req = &req->req;

	remaining = _req->length - _req->actual;

	if (remaining) {

		// Send next packet of current request
	
		length = min(remaining, (u16) ep->ep.maxpacket);

		pdc_write_ep(ep->ep_index, _req->buf + _req->actual, length);
		
		_req->actual += length;

	} else {
		
		// Current request has completed
	
		DBG_VMSG("%s: request completed\n", __FUNCTION__);
		
		pdc_complete(ep, req, 0);

		if (!list_empty(&ep->queue)) {
			
			req = container_of(ep->queue.next, struct pdc_request,
					queue);
			_req = &req->req;

			length = min((u16) _req->length,(u16) ep->ep.maxpacket);

			pdc_write_ep(ep->ep_index, _req->buf, length);

			_req->actual += length;

		}
	}
}


static void pdc_handle_bus_reset_intr(struct pdc_globals *dev)
{
	int	i;

	DBG_PRINTK("## USB bus reset ##\n");

	pdc_locked_go_to_state(STATE_DEFAULT, 0, 0);
			
	// // nxp code kills all pending requests at this point // //

	// Reset speed to full speed
	
	dev->gadget.speed = USB_SPEED_FULL;
			
	// Reset endpoint max pkt sizes to full-speed values
		
	for (i = 1 ; i < PDC_NUM_ENDPOINTS ; i++) {
		dev->ep[i].ep.maxpacket = BULK_FIFO_SIZE_FS;
		dev->ep[i].desc = NULL;
		pdc_disable_endpoint(dev, &dev->ep[i]);
	}
	
	// Bus reset clears the CBW error condition

	dev->cbw_error = 0;
	
	pdc_init(dev);
	pdc_connect(dev);
			
	dev->intr_enables = IE_BASIC_CTRL_EP_MASK;

	/* make sure ep0's queue is cleaned up  */
	{
		struct pdc_ep *ep = &dev->ep[0];

		pdc_nuke(ep, -ECONNRESET);
	}

	DBG_VMSG("bus reset: calling gadget\'s disconnect handler...\n");
	if (dev->driver && dev->driver->disconnect) {
		spin_unlock(&dev->lock);
		dev->driver->disconnect(&dev->gadget);
		spin_lock(&dev->lock);
	}
}
			

static void pdc_handle_ep0rx_intr(struct pdc_globals *dev)
{
	u32	len;

	// Read the data out of the endpoint - should be a
	// zero length packet.
			
	pdc_write32(DC_EP_INDEX_REG, EP0RX); // point to EP0 OUT
	len = pdc_read32(DC_BUFFER_LENGTH_REG);
	if (unlikely (len != 0)) {
		DBG_MSG("%d-byte OUT pkt on ep 0!\n", len);
		// FIXME - no way to return this to gadget driver
	}
	// ACK the OUT packet
	pdc_write32(DC_CTRLFN_REG, CNTRLFN_STATUS_EM_ACK);
}


static void pdc_locked_go_to_state(u8 state, u8 address, u8 configuration)
{
	struct pdc_globals	*dev = &globals;
	u32			mA;
	u32			enables;

	pdc_log_state_change("%s(adr=%d, cfg=%d, state=%2d [%s])\n",
			__FUNCTION__, address, configuration,
			state, state_names[state]);
	
	dev->device_state = state;
	dev->device_address = address;
	dev->device_configuration = configuration;

	/* When we transition to any of the non-configured states, we want to
	   make sure that the configured vbus draw value is reset to a valid
	   default value.  Later, when we're configured, the value will be
	   updated with the correct value, and we'll transition to a higher
	   level, if necessary.						*/

	switch (dev->device_state) {
	case STATE_DETACHED:
	case STATE_ATTACHED:
	case STATE_POWERED_SUSPENDED:
	case STATE_DEFAULT_SUSPENDED:
	case STATE_ADDRESS_SUSPENDED:
		mA = 1;
		dev->configured_vbus_draw = 100;  // reset vbus draw value
		break;
	case STATE_POWERED:
	case STATE_DEFAULT:
	case STATE_ADDRESS:
		mA = 100;
		dev->configured_vbus_draw = 100;  // reset vbus draw value
		break;
	case STATE_CONFIGURED:
		mA = dev->configured_vbus_draw;	// use configuration value
		// don't alter configured vbus draw value
		break;
	case STATE_CONFIGURED_SUSPENDED:
		mA = 1;
		// don't reset configured vbus draw value
		break;
	default:
		DBG_PRINTK("bogus state: %d\n", dev->device_state);
		mA = 100;
	}
	pdc_set_current_draw(dev, mA);

	/* Set interrupt enables based on state.  This allows us to ignore
	   the continuous stream of VBUS interrupts that are generated while
	   we're in the POWERED and POWERED_SUSPENDED states.		*/
	
	dev->intr_enables = intr_enables_table[state];
	enables = dev->intr_enables | dev->intr_enables_bulk;
	pdc_write32(DC_INTR_ENABLE_REG, enables);
}


static irqreturn_t pdc_isr(int irq, void *data, struct pt_regs *r)
{
	struct pdc_globals	*dev = (struct pdc_globals *) data;
	
	unsigned int		event, non_ctrl_ep_event;
	u32			ints, masked, enables;
	int			i;
	
	pdc_log_entry("\n");
	
	spin_lock(&dev->lock);

	while (1) {

		// make sure device is unlocked for writing

		pdc_write32(DC_UNLOCK_DEVICE_REG, 0xAA37);
		
		/* Get a snapshot of the interrupt sources,
		 * then clear all the interrupt sources.
		 */
		
		ints = pdc_read32(DC_INTERRUPT_REG);	// save intr sources
		pdc_write32(DC_INTERRUPT_REG, ints);	// clear intr sources

		/* Mask out invalid/unsupported IRQs */
				
		masked = ints & (IE_EP0SETUP | IE_EP0RX | IE_EP0TX | \
				 IE_NON_CTRL_EP_MASK | IE_BRST |      \
				 IE_SUSP | IE_VBUS | IE_HS_STA |     \
				 IE_RESM | IE_DMA);

		// If suspend and resume interrupts are set at the same time,
		// they should be ignored  (Philips AN10046 sect. 2.4)

		if ((masked & (IE_SUSP | IE_RESM)) == (IE_SUSP | IE_RESM)) {
			DBG_VMSG("suspend *and* resume: masked=0x%x\n", masked);
			masked &= ~(IE_SUSP | IE_RESM);
		}
		
		// We shadow the contents of the interrupt enable register,
		// because it gets cleared to 0x0001 when a bus reset occurs.
		// The high-speed status interrupt can be reported at the same
		// time, and would be lost if and'ed against the new bus reset
		// interrupt enable register value.

		enables = dev->intr_enables | dev->intr_enables_bulk;

		event = masked & enables;

		pdc_log_intr("i %x m %x en %x ev %x\n", ints, masked,
				enables, event);
		DBG_VMSG("i %x e %x\n", ints, event);
		
		/* Exit if there is nothing left to handle */
		if (event == 0)
			goto exit;
		
		// First, handle endpoint interrupts, if any
		
		if (event & IE_EP1TX) {
			DBG_VMSG("ep1 IN intr\n");
			pdc_handle_ep_in_intr(dev, &dev->ep[1]);
		}

		if (event & IE_EP2RX) {
			DBG_VMSG("ep2 OUT intr\n");
			pdc_handle_ep_out_intr(dev, &dev->ep[2]);
		}
		
		non_ctrl_ep_event = event & IE_NON_CTRL_EP_MASK;
		non_ctrl_ep_event &= ~(IE_EP2RX | IE_EP1TX);
		if (non_ctrl_ep_event)
			DBG_MSG("  intr on unconfigured endpoint!\n");

		if (event & IE_EP0SETUP) {
			DBG_VMSG("SETUP intr\n");
			pdc_handle_ep0setup_intr(dev);
#ifdef CONFIG_USB_OTG
			nctrl_msg++;
#endif
		}
		
		if (event & IE_EP0TX) {
			DBG_VMSG("ep0 IN intr\n");  // ACK after IN pkt
			pdc_handle_ep0tx_intr(dev);
		}

		if (event & IE_EP0RX) {
			DBG_VMSG("ep0 OUT intr\n"); // ep0 OUT pkt received
			pdc_handle_ep0rx_intr(dev);
		}

		if (event & IE_BRST) {
			DBG_VMSG("###  USB bus reset interrupt  ###\n");
			pdc_handle_bus_reset_intr(dev);
		}

		if (event & IE_VBUS) {
			DBG_MSG("VBUS interrupt\n");

			if ((dev->device_state == STATE_ATTACHED) ||
			    (dev->device_state == STATE_DETACHED))
				DBG_PRINTK("## USB connected ##\n");

			dev->controller_suspended = 0;

			// make sure controller is awake
			
			pdc_wake_controller();

			// re-enable interrupts
			
			pdc_write32(DC_INTR_ENABLE_REG, dev->intr_enables);
			pdc_write32(DC_MODE_REG, MODE_GLINTENA);
			
			// start watching for VBUS removal
			mod_timer(&dev->vbus_timer,
					jiffies + VBUS_TIMER_INTERVAL);

			pdc_locked_go_to_state(STATE_POWERED, 0, 0);

			/* Connect to the bus - connect D+ pullup resistor */
#ifndef CONFIG_USB_OTG
			pdc_write32(OTG_CONTROL_REG, 0x00160000);
			pdc_write32(OTG_CONTROL_REG, 0x00000081);
#endif
		}
		
		if (event & IE_SUSP) {
			DBG_MSG("USB suspend interrupt\n");
			/* Inform the gadget driver that we're suspended */
			if (dev->gadget.speed != USB_SPEED_UNKNOWN &&
			    dev->driver && dev->driver->suspend) {
			    	spin_unlock(&dev->lock);
			    	dev->driver->suspend(&dev->gadget);
				spin_lock(&dev->lock);
			}

			if ((dev->device_state != STATE_DETACHED) &&
			    (dev->device_state != STATE_ATTACHED)) {
				pdc_locked_go_to_state(
					dev->device_state | STATE_SUSPENDED,
					dev->device_address,
					dev->device_configuration);
			}
		}

		if (event & IE_RESM) {
			DBG_MSG("USB resume interrupt\n");
			/* Inform the configuration driver  */
			if (dev->gadget.speed != USB_SPEED_UNKNOWN &&
			     dev->driver && dev->driver->resume) {
			     	spin_unlock(&dev->lock);
			 	dev->driver->resume(&dev->gadget);
				spin_lock(&dev->lock);
			}

			if ((dev->device_state != STATE_DETACHED) &&
			    (dev->device_state != STATE_ATTACHED)) {
				pdc_locked_go_to_state(
					dev->device_state & ~STATE_SUSPENDED,
					dev->device_address,
					dev->device_configuration);
			}
		}

		if (event & IE_HS_STA) {
			DBG_MSG("High speed state change interrupt\n");
			/* Reconfigure device endpoints if device was FS */
			if ((dev->gadget.speed == USB_SPEED_FULL) ||
			    (dev->gadget.speed == USB_SPEED_UNKNOWN)) {
			   	DBG_VMSG("...going to high speed...\n");
				dev->gadget.speed = USB_SPEED_HIGH;
			
				for (i = 1 ; i < PDC_NUM_ENDPOINTS ; i++) {
					dev->ep[i].ep.maxpacket =
							BULK_FIFO_SIZE_HS;
					dev->ep[i].desc = NULL;
				}
			}
		}
		
		if (event & IE_DMA) {
			DBG_MSG("Unexpected Device DMA interrupt!!!\n");
			/* Handle DMA Int source register  */
		}	

	}

exit:
	spin_unlock(&dev->lock);

	pdc_log_exit();

	return IRQ_HANDLED;
}


/* -----  probe and remove code  ----- */


/*
 * pdc_probe is called in response to the "driver_register" call made by
 * pdc_module_init.  (As a matter of fact, pdc_probe is called *during* the
 * driver_register call.)
 * This routine finishes initializing the driver's global data.
 * It also makes sure that all interrupts are disabled, and that
 * the device is electrically disconnected from the USB bus.
 */

static int __init pdc_probe(struct device *device)
{
	struct pdc_globals	*dev = &globals;
	struct resource		*rsrc;
	int			retval;

	pdc_log_entry("\n");

	/* make sure we have an ISP1761 */
	
	if (0) {
		printk(KERN_ERR "%s: ISP1761 not found!\n", DRIVER_NAME);
		return -ENODEV;
	}

	printk(KERN_ERR "%s: IRQ %d\n", DRIVER_NAME, PDC_IRQ);

	/* other non-static parts of init */

	dev->logging_flags = 0;

	spin_lock_init(&dev->lock);

	dev->dev = device;

	init_timer(&dev->timer);
	dev->timer.function = pdc_watchdog;
	dev->timer.data = (unsigned long) dev;

	init_timer(&dev->vbus_timer);
	dev->vbus_timer.function = pdc_vbus_timer_proc;
	dev->vbus_timer.data = (unsigned long) dev;

	device_initialize(&dev->gadget.dev);
	dev->gadget.dev.parent = device;
	dev->gadget.dev.dma_mask = device->dma_mask;
	dev->gadget.speed = USB_SPEED_FULL;

	dev_set_drvdata(device, dev);

	dev->remote_wakeup = 0;

	dev->configured_vbus_draw = 100;	// safe default
	dev->current_vbus_draw = 0;

	dev->controller_suspended = 0;
	
	dev->cbw_error = 0;

#ifdef CONFIG_ARCH_FIONA
	
	// Check whether the user disabled suspending the device controller.
	
	dev->suspend_enabled = !DISABLE_PDC_SUSPEND();
#else
	dev->suspend_enabled = 1;
#endif


	rsrc = request_mem_region(PDC_PHYS_BASE,PDC_PHYS_LENGTH,DRIVER_NAME);
	if (!rsrc) {
		pdc_log_err("request_mem_region returned NULL\n");
		return -EBUSY;
	}
	
	dev->base_address = ioremap_nocache(PDC_PHYS_BASE, PDC_PHYS_LENGTH);
	if (dev->base_address == NULL) {
		pdc_log_err("Failed to ioremap the pdc!\n");
		release_mem_region(PDC_PHYS_BASE, PDC_PHYS_LENGTH);
		return -ENOMEM;
	}
	
	pdc_log_debug("dev->base_address = 0x%p\n", dev->base_address);
	
	dev->intr_enables = IE_BASIC_CTRL_EP_MASK; // default interrupt events

	dev->intr_enables_bulk = 0;  // bits are set when endpts are enabled

	pdc_locked_go_to_state(STATE_DETACHED, 0, 0);	// initial state
	
	/*for device only  */
	
	pdc_dump();

#ifndef CONFIG_USB_OTG
 #if 0	// this code doesn't look right...
	pdc_write32(OTG_CONTROL_REG,0x00000081);
	pdc_write32(OTG_CONTROL_REG,0x00160000);

	pdc_write32(OTG_CONTROL_REG,0xFFFF0000); // clear all otg interrupt
	pdc_write32(OTG_CONTROL_REG,0x00000481);
 #else
	pdc_write32(OTG_CONTROL_REG, 0xFF790086); // restore to reset values
	pdc_write32(OTG_CONTROL_REG, 0x00060400); // ISP1761 sect. 9.5.2.1
 #endif
#endif

	pdc_wake_controller();

	// Init Mode register
	// Init Interrupt Configuration register
	// Init Interrupt Enable register
	// Init DMA Configuration register
	// Init DMA Hardware register
	
	DBG_VMSG("Intr Enable reg = 0x%x\n", pdc_read32(DC_INTR_ENABLE_REG));
	
	// Init endpoints
	
	pdc_disable_all_endpts(dev);

	pdc_init(dev);		// first-time initialization
	
	pdc_disable(dev);	// disconnect from bus, and force off
	
	pdc_reinit(dev);	// reset data structures

	/* irq setup after old hardware state is cleaned up */
	
	set_irq_type(PDC_IRQ, IRQT_FALLING);
	
	retval = request_irq(PDC_IRQ, pdc_isr,
			SA_INTERRUPT, DRIVER_NAME, dev);
	if (retval != 0) {
		printk(KERN_ERR "%s: can't get irq %i, err %d\n",
			DRIVER_NAME, PDC_IRQ, retval);
		return -EBUSY;
	}
	dev->got_irq = 1;

	/* Add "vbus" and "configured" files to device's sysfs directory */
	
	device_create_file(dev->dev, &dev_attr_vbus);
	device_create_file(dev->dev, &dev_attr_configured);
#ifdef DEBUG_CHIP
	device_create_file(dev->dev, &dev_attr_chip);
#endif
	device_create_file(dev->dev, &dev_attr_globals);
	device_create_file(dev->dev, &dev_attr_logging);

#ifdef CONFIG_FIONA_PM_USB_DEVICE
	// Register USB Device driver with Fiona Power Manager,
	// we're ignoring result here as we'll still allow driver to
	// work even if we can't do efficient power management...
	if (usb_device_register_with_fpow((void*) dev,
			&g_usb_device_fpow_component_ptr))
		printk("USB Device: Failed to register with fpow !!\n");
#endif

	pdc_log_exit();

	return 0;
}


static int __exit pdc_remove(struct device *device)
{
	struct pdc_globals	*dev = &globals;

	pdc_log_entry("\n");

	pdc_disable(dev);

	usb_gadget_unregister_driver(dev->driver);

#ifdef CONFIG_FIONA_PM_USB_DEVICE
	// Unregister USB Device driver with Fiona Power Manager
	usb_device_unregister_with_fpow(g_usb_device_fpow_component_ptr);
	g_usb_device_fpow_component_ptr  = NULL;
#endif

	if (dev->got_irq) {
		free_irq(PDC_IRQ, dev);
		dev->got_irq = 0;
	}

	dev_set_drvdata(dev->dev, 0);

	iounmap(dev->base_address);

	release_mem_region(PDC_PHYS_BASE, PDC_PHYS_LENGTH);

	pdc_log_exit();

	return 0;
}


/* -----  module init and exit code  ----- */

static int __init pdc_module_init(void)
{
	int	retval;
	
	pdc_log_entry("\n");
	
	printk(KERN_INFO "%s: version %s\n", DRIVER_NAME, DRIVER_VERSION);

	retval = driver_register(&pdc_driver);
	
	if (retval)
		DBG_MSG("driver_register returned %d\n", retval);
	
	/* Add 1761 device controller to the list of platform devices */
	/* NOTE: This will cause a call to the probe routine. */

	retval = platform_device_register(&pdc_device);

	if (retval) {
		DBG_MSG("platform_register_device returned %d\n", retval);
		goto exit;
	}

	pdc_dump();
exit:
	pdc_log_exit();

	return retval;
}

module_init(pdc_module_init);


static void __exit pdc_module_exit(void)
{
	struct pdc_globals	*dev = &globals;

	pdc_log_entry("\n");

	device_remove_file(dev->dev, &dev_attr_vbus);
	device_remove_file(dev->dev, &dev_attr_configured);
#ifdef DEBUG_CHIP
	device_remove_file(dev->dev, &dev_attr_chip);
#endif
	device_remove_file(dev->dev, &dev_attr_globals);
	
	platform_device_unregister(&pdc_device);

	driver_unregister(&pdc_driver);
	
	pdc_log_exit();
}

module_exit(pdc_module_exit);


MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR(DRIVER_AUTHOR);

MODULE_LICENSE("GPL");

