/*
 * fiona.c  --  SoC audio for Fiona
 *
 * Copyright 2005, 2006 Lab126, Inc.
 * 
 * Author: Kenneth Kiraly <kiraly@lab126.com>
 *
 * based on sound/soc/pxa/mainstone.c by Liam Girdwood
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
 
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dpm.h>
#include <sound/hpdetect.h>
#include <linux/proc_fs.h>
#include <linux/syscalls.h>
#include <linux/sem.h>
#include <linux/ipc.h>
#include <linux/delay.h>
#include <linux/suspend.h>
#include <linux/interrupt.h>

#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/fiona.h>
#include <asm/arch/fpow.h>
#include <asm/arch/fiona_eventmap.h>
#include <asm/arch/audio.h>

#include "../codecs/wm8971.h"

#include "fiona-audio-pm.h"

#ifdef CONFIG_FIONA_PM_AUDIO
extern struct fpow_component *g_audio_fpow_component_ptr;
static int audio_fpow_configured = 0;
int		audio_register_with_fpow(struct fpow_component **fpow_dev, void *private);
void	audio_unregister_with_fpow(struct fpow_component **fpow_dev);
#endif


#undef ALLOW_NOTIFY_SCRIPT

// Define to handle headphone events locally, 
//   undef to pass event up to framework
#undef CONFIG_SYSTEM_SWITCH_CONTROL

#define info(format, arg...) printk("hpdet: " format "\n" , ## arg)

#define hp_detect_status()		((GPLR(GPIO_FIONA_HPDET_EVT3) & GPIO_bit(GPIO_FIONA_HPDET_EVT3)) ? 0 : 1)


static struct proc_dir_entry *proc_hpdet_parent = NULL;

static wait_queue_head_t hpdet_event_wq;
static int hpdet_event = 0;

static pid_t hpdet_thread_pid = 0;
static DECLARE_COMPLETION(hpdet_thread_exited);


static void
hpdet_notify_sem(
    void)
{
	int sem_id;

	// retrieve the notification semaphore
	sem_id = sys_semget(HPDET_NOTIFY_SEM_KEY, 1, S_IRUGO | S_IWUGO);
	if (sem_id >= 0) {
		struct sembuf sem_operations[1];

		// signal the semaphore
		sem_operations[0].sem_num = 0;		// which semaphore to use
		sem_operations[0].sem_op = 1;		// add one from the semaphore value
		sem_operations[0].sem_flg = 0;		// set wait flag

		// do the signal    
		sys_semop(sem_id, sem_operations, 1);

	} else {
		// semaphore doesn't exist; ignore as nobody may be waiting on it

	}
}


#ifdef ALLOW_NOTIFY_SCRIPT
static void
hpdet_notify_script(
	int inserted)
{
	int err;

	char *envp[] = {
		"HOME=/",
		"TERM=linux",
		"PATH=/sbin:/usr/sbin:/bin:/usr/bin",
		NULL
	};

	char *argv[] = {
		HPDET_NOTIFY_SCRIPT,
		inserted ? "1" : "0",
		NULL
	};

	if ((err = call_usermodehelper(argv[0], argv, envp, 0)) < 0) {
	    info("notify script execution failed (%d)", err);
	}
}
#endif


static int
hpdet_thread(
	void *unused)
{
	int hp_status = 0;

	// initialize our global state of headphones
	hp_status = hp_detect_status();

	daemonize("hpdetd");
	allow_signal(SIGKILL);

	while (1) {
		wait_event_interruptible(hpdet_event_wq, hpdet_event);

#if defined(CONFIG_PM) || defined(CONFIG_FIONA_PM)
		if (current->flags & PF_FREEZE)  {
			refrigerator(PF_FREEZE);

			// we just woke up, check to see if the headphones transitioned...
			// if it hasn't changed, wait for next event, otherwise drop thru
			// and detect/report headphone status
			if (hp_status == hp_detect_status()) {
				hpdet_event = 0;
				continue;
			}
		}
#endif

		if (signal_pending(current)) {
			break;
		}

		// wait 500ms (debounce - interrupt source hits several times
		// while user is pushing in jack )
		msleep(500);

		hpdet_event = 0;

		hp_status = hp_detect_status();

		enable_irq(FIONA_HPDET_IRQ);

		// notify any waiting processes
		hpdet_notify_sem();

#ifdef ALLOW_NOTIFY_SCRIPT
		// execute user-level notify script
		hpdet_notify_script(hp_status);
#endif

#ifndef CONFIG_SYSTEM_SWITCH_CONTROL
		handle_fiona_event( hp_status ? \
			HEADPHONES_IN_EVENTCODE : HEADPHONES_OUT_EVENTCODE);
#endif

	}

	complete_and_exit(&hpdet_thread_exited, 0);
}


static void
hpdet_start_thread(
	void)
{
	pid_t pid;

	init_waitqueue_head(&hpdet_event_wq);

	pid = kernel_thread(hpdet_thread, NULL, CLONE_KERNEL);

	hpdet_thread_pid = (pid >= 0) ? pid : 0;
}


static void
stop_hpdet_thread(
	void)
{
	int ret;

	if (hpdet_thread_pid > 0) {
		ret = kill_proc(hpdet_thread_pid, SIGKILL, 1);
		if (ret == 0) {
			wait_for_completion(&hpdet_thread_exited);
		}
	}
}


static irqreturn_t
hpdet_irq_handler(
	int irq,
	void *data,
	struct pt_regs *regs)
{
	disable_irq(FIONA_HPDET_IRQ);

	hpdet_event = 1;

	wake_up(&hpdet_event_wq);

	return IRQ_HANDLED;
}


static void
hpdet_init_irq(
	void)
{
	int res;

	set_irq_type(FIONA_EVT3_HPDET_IRQ, IRQT_BOTHEDGE);

	res = request_irq(FIONA_EVT3_HPDET_IRQ, hpdet_irq_handler, SA_INTERRUPT, "hpdet", NULL);
	if (res == 0) {
		//info("switch is using IRQ %d", FIONA_EVT3_HPDET_IRQ);

	} else {
		info("error %d requesting IRQ %d", res, FIONA_EVT3_HPDET_IRQ);

	}
}


static void
hpdet_free_irq(
	void)
{
	free_irq(FIONA_P3_WAN_EN_IRQ, NULL);
}


static int 
proc_status_read(
	char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = sprintf(page, "%d\n", hp_detect_status());

	*eof = 1;

	return len;
}


static void
hpdet_proc_status_init(
	void)
{
	proc_hpdet_parent = create_proc_entry(HPDET_PROC_AUDIO, S_IFDIR | S_IRUGO | S_IXUGO, NULL);
	if (proc_hpdet_parent != NULL) {
		struct proc_dir_entry *proc_hpdet = create_proc_entry(HPDET_PROC_HEADPHONE, S_IRUGO, proc_hpdet_parent);

		proc_hpdet->data = NULL;
		proc_hpdet->read_proc = proc_status_read;
		proc_hpdet->write_proc = NULL;
	}
}


static void
hpdet_init(
	void)
{
	// configure headphone detect GPIO
	pxa_gpio_mode(GPIO_FIONA_HPDET_EVT3_MD);

	// create the /proc entry
	hpdet_proc_status_init();

	// start the detect handler thread
	hpdet_start_thread();

	// enable the detect IRQ
	hpdet_init_irq();
}


static void
hpdet_cleanup(
	void)
{
	hpdet_free_irq();

	stop_hpdet_thread();

	if (proc_hpdet_parent != NULL) {
		remove_proc_entry(HPDET_PROC_HEADPHONE, proc_hpdet_parent);
		remove_proc_entry(HPDET_PROC_AUDIO, NULL);
		
		proc_hpdet_parent = NULL;
	}
}


static int
fiona_startup(
	snd_pcm_substream_t *substream)
{
	// power up the output channel
	GPSR1 = GPIO_bit(GPIO52_AUD_OUT_POWER);

	return 0;
}


static void
fiona_shutdown(
	snd_pcm_substream_t *substream)
{
	// power down the output channel
	GPCR1 = GPIO_bit(GPIO52_AUD_OUT_POWER);
}


static struct snd_soc_ops fiona_ops = {
	.startup = fiona_startup,
	.shutdown = fiona_shutdown,
};


static int fiona_suspend(
	struct snd_soc_machine *machine)
{
	return 0;
}


static int fiona_resume(
	struct snd_soc_machine *machine)
{
	return 0;
}


static int fiona_probe(
	struct snd_soc_machine *machine)
{
	return 0;
}


static int fiona_remove(
	struct snd_soc_machine *machine)
{
	return 0;
}


// fiona audio dpm widgets
static const struct snd_soc_dpm_widget wm8971_dpm_widgets[] = {
	SND_SOC_DPM_HP("Headphone Jack"),
	SND_SOC_DPM_OUTPUT("Speaker")
};

// fiona audio interconnections
static const char *fiona_interconnects[][3] = {
	// headphone connected to LOUT1, ROUT1
	{ "Headphone Jack", NULL, "LOUT1" },
	{ "Headphone Jack", NULL, "ROUT1" },

	// speaker connected to LOUT2, ROUT2
	{ "Speaker", NULL, "LOUT2" },
	{ "Speaker", NULL, "ROUT2" },

	// terminator	
	{ NULL, NULL, NULL }
};

// unconnected codec pins on fiona
static const struct snd_soc_dpm_pin wm8971_fiona_pins[] = {
	{ "LINPUT1", 0 },
	{ "RINPUT1", 0 },
	{ "MIC", 0 }
};


// provide headphone detect and speaker phase support
static const char *hd_pol[] = { "Headphone", "Speaker" };
static const char *mono_inv[] = { "Non Inverted", "Inverted" };

static const struct soc_enum wm8971_enum[] = {
    SOC_ENUM_SINGLE(WM8971_ADCTL2, 5, 2, hd_pol),
    SOC_ENUM_SINGLE(WM8971_ADCTL2, 4, 2, mono_inv),
};

static const snd_kcontrol_new_t wm8971_fiona_controls[] = {
	SOC_SINGLE("Headphone Detect", WM8971_ADCTL2, 6, 1, 0),
	SOC_ENUM("Headphone Detect Polarity", wm8971_enum[0]),
	SOC_ENUM("Playback Speaker Phase", wm8971_enum[1]),
};


static int fiona_wm8971_init(
	struct snd_soc_codec *codec)
{
	int i, err;
	
	// set up fiona codec pins
	for (i = 0; i < ARRAY_SIZE(wm8971_fiona_pins); ++i) {
		snd_soc_dpm_set_connection(codec, &wm8971_fiona_pins[i]);
	}
	
	// set up fiona-specific controls
	for (i = 0; i < ARRAY_SIZE(wm8971_fiona_controls); ++i) {
		if ((err = snd_ctl_add(codec->card, snd_soc_cnew(&wm8971_fiona_controls[i],codec, NULL))) < 0) {
			return err;
		}
	}
	
	// add fiona-specific widgets
	for (i = 0; i < ARRAY_SIZE(wm8971_dpm_widgets); ++i) {
		snd_soc_dpm_new_control(codec, &wm8971_dpm_widgets[i]);
	}
	
	// set up fiona-specific audio path interconnects
	for (i = 0; fiona_interconnects[i][0] != NULL; ++i) {
		snd_soc_dpm_connect_input(codec, fiona_interconnects[i][0], fiona_interconnects[i][1], fiona_interconnects[i][2]);
	}

	snd_soc_dpm_sync(codec);

#ifdef CONFIG_FIONA_PM_AUDIO
    // register Audio driver with Fiona Power Manager.  we're ignoring result here as we'll
    // still allow driver to work even if we can't do efficient power management...
    if (!audio_fpow_configured && !audio_register_with_fpow(&g_audio_fpow_component_ptr, codec)) {
    	audio_fpow_configured = 1;

    } else {
        printk("AUDIO: Failed to register with fpow !!\n");

    }
#endif

	return 0;
}


static struct snd_soc_machine_config codecs[] = {
	{
		.name = "WM8971",
		.sname = "WM8971",
		.interface = SND_SOC_I2S,
		.int_num = 0,
		.init = fiona_wm8971_init,
	},
};

static struct snd_soc_machine fiona = {
	.name = "Fiona",
	.probe = fiona_probe,
	.remove = fiona_remove,
	.suspend = fiona_suspend,
	.resume = fiona_resume,
	.ops = &fiona_ops,
	.config = codecs,
	.nconfigs = ARRAY_SIZE(codecs),
};

static int __init snd_soc_init(
	void) 
{
	hpdet_init();

	return snd_soc_register_machine(&fiona);
}


static void __exit snd_soc_exit(
	void) 
{
	hpdet_cleanup();

#ifdef CONFIG_FIONA_PM_AUDIO
	if (audio_fpow_configured) {
		audio_fpow_configured = 0;
		// unregister with Fiona Power Management System
		audio_unregister_with_fpow(&g_audio_fpow_component_ptr);
	}
#endif

	snd_soc_unregister_machine(&fiona);
}


module_init(snd_soc_init); 
module_exit(snd_soc_exit);

// module information
MODULE_AUTHOR("Kenneth Kiraly <kiraly@lab126.com>");
MODULE_DESCRIPTION("ALSA SoC Fiona board");
MODULE_LICENSE("GPL");

