/*
 * soc-core.c  --  ALSA Soc Audio Layer
 *
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Author: Liam Girdwood
 *         liam.girdwood@wolfsonmicro.com or linux@wolfsonmicro.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  Revision history
 *    12th Aug 2005   Initial version.
 *    25th Oct 2005   Working Codec, Interface and Platform registration.
 *
 *  TODO:
 *   o Finish machine layer - move machine stuff out pxa2xx-pcm.c
 *   o Make codec info const and copy on registration (e.g. kcontrols)
 *   o Change away from ALSA pm callbacks and types for 2.6.15
 *   o Add hw rules to enforce rates, etc.
 *   o More testing with other codecs/machines.
 *   o Add more codecs and platforms to ensure good API coverage.
 *   o Support TDM on PCM and I2S
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/pm.h>
#include <linux/bitops.h>
#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dpm.h>
#include <sound/initval.h>

#ifdef CONFIG_FIONA_PM_AUDIO
#include <asm/arch/fpow.h>
#endif

#define SND_SOC_VERSION "0.7"

#define iface_name(x) \
	(((x) == SND_SOC_AC97) ? "ac97" : \
	((x) == SND_SOC_I2S) ? "i2s" : \
	((x) == SND_SOC_SSP) ? "ssp" : NULL)

static DECLARE_MUTEX(soc_sem);
static LIST_HEAD(codecs);
static LIST_HEAD(interfaces);
static struct snd_soc_platform *platform = NULL;
static struct snd_soc_machine *machine = NULL;


int snd_soc_update_bits(struct snd_soc_codec *codec, unsigned short reg,
				unsigned short mask, unsigned short value)
{
	int change;
	unsigned short old, new;

	old = snd_soc_read(codec, reg);
	new = (old & ~mask) | value;
	change = old != new;
	if (change)
		snd_soc_write(codec, reg, new);

	return change;
}

int snd_soc_test_bits(struct snd_soc_codec *codec, unsigned short reg,
				unsigned short mask, unsigned short value)
{
	int change;
	unsigned short old, new;

	old = snd_soc_read(codec, reg);
	new = (old & ~mask) | value;
	change = old != new;

	return change;
}

static int soc_match_hw(struct snd_soc_hw_mode *c, struct snd_soc_hw_mode *i, 
	int rate, int fs)
{
	if (c->rate != i->rate)
		return 0;
	
	if (c->rate != rate)
		return 0;
	
	if (c->fs != i->fs)
		return 0;
	
	if (!(c->hformat & i->hformat & SND_SOC_FORMAT_MASK))
		return 0;
	
	if (!(c->hformat & i->hformat & SND_SOC_CLOCK_MASK))
		return 0;
	
	if (!(c->hformat & i->hformat & SND_SOC_INV_MASK))
		return 0;
	
	if (i->fs < fs)
		return i->fs;
	return 0;
}

/* Check that common hw capabilities exist between the codec and interface for
 * the given hw params */
static int soc_check_hw_params(snd_pcm_substream_t *substream, 
	snd_pcm_hw_params_t *params)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	int fs = 2048, i, j, match = 0;	
	
	/* try and match interface <--> codec table entries */
	for (i = 0; i < pcm_c->hw.num_hmodes; i++) {
		for (j = 0; j < pcm_i->hw.num_hmodes; j++) {
			if (soc_match_hw(&pcm_c->hw.hmodes[i], &pcm_i->hw.hmodes[j], 
					params_rate(params), fs)) {
				pcm_c->hw_runtime = pcm_c->hw.hmodes[i];
				pcm_i->hw_runtime = pcm_i->hw.hmodes[j];
				fs = pcm_c->hw.hmodes[i].fs;
				match = 1;
			}
		}
	}
	
	if (!match) {
		printk(KERN_ERR "soc: can't match codec and interface capabilities\n");
		return -EINVAL;
	}
	
	pcm_i->hw_runtime.hformat =
		1 << (generic_ffs(pcm_i->hw_runtime.hformat & pcm_c->hw_runtime.hformat & SND_SOC_FORMAT_MASK) -1) | 
		1 << (generic_ffs(pcm_i->hw_runtime.hformat & pcm_c->hw_runtime.hformat & SND_SOC_CLOCK_MASK) - 1) |
		1 << (generic_ffs(pcm_i->hw_runtime.hformat & pcm_c->hw_runtime.hformat & SND_SOC_INV_MASK) - 1);

	pcm_i->hw_runtime.hbits = pcm_i->hw_runtime.hbits & pcm_c->hw_runtime.hbits;
	pcm_c->hw_runtime.hformat = pcm_i->hw_runtime.hformat;
	pcm_c->hw_runtime.hbits = pcm_i->hw_runtime.hbits;
	
	/* for debug atm */
	printk(KERN_INFO "soc: codec rate %d fs %d format %x\n", pcm_c->hw_runtime.rate, 
		pcm_c->hw_runtime.fs, pcm_c->hw_runtime.hformat);
	printk(KERN_INFO "soc: iface rate %d fs %d format %x\n", pcm_i->hw_runtime.rate, 
		pcm_i->hw_runtime.fs, pcm_i->hw_runtime.hformat);
	printk(KERN_INFO "soc: audio rate %d chn %d\n", params_rate(params), params_channels(params));	
	
	return 0;
}

/*
 * Initialize the runtime->hw record, private data can be allocated
 */
static int soc_pcm_open(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	snd_pcm_runtime_t *runtime = substream->runtime;
	int ret = 0;

	if(!machine || !platform)
		return -ENODEV;
	
	/* Check that the codec and SoC audio interface's are compatible */
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		runtime->hw.rates = pcm_c->playback.rates & pcm_i->playback.rates;
		runtime->hw.formats = pcm_c->playback.formats & pcm_i->playback.formats;
		runtime->hw.rate_min = max(pcm_c->playback.rate_min, pcm_i->playback.rate_min);
		runtime->hw.rate_max = min(pcm_c->playback.rate_max, pcm_i->playback.rate_max);
		runtime->hw.channels_min = max(pcm_c->playback.channels_min, pcm_i->playback.channels_min);
		runtime->hw.channels_max = min(pcm_c->playback.channels_max, pcm_i->playback.channels_max);
	} else {
		runtime->hw.rates = pcm_c->capture.rates & pcm_i->capture.rates;
		runtime->hw.formats = pcm_c->capture.formats & pcm_i->capture.formats;
		runtime->hw.rate_min = max(pcm_c->capture.rate_min, pcm_i->capture.rate_min);
		runtime->hw.rate_max = min(pcm_c->capture.rate_max, pcm_i->capture.rate_max);
		runtime->hw.channels_min = max(pcm_c->capture.channels_min, pcm_i->capture.channels_min);
		runtime->hw.channels_max = min(pcm_c->capture.channels_max, pcm_i->capture.channels_max);
	}
	
	/* will the codec and interface work together ? */
	snd_pcm_limit_hw_rates(runtime);
	if (!runtime->hw.rates) {
		printk(KERN_ERR "soc: %s <-> %s No matching rates\n", pcm_c->name, pcm_i->name);
		return -ENODEV;
	}
	if(!runtime->hw.formats) {
		printk(KERN_ERR "soc: %s <-> %s No matching formats\n", pcm_c->name, pcm_i->name);
		return -ENODEV;
	}
	if(!runtime->hw.channels_min || !runtime->hw.channels_max) {
		printk(KERN_ERR "soc: %s <-> %s No matching channels\n", pcm_c->name, pcm_i->name);
		return -ENODEV;
	}
	printk(KERN_INFO "soc: %s <-> %s info:\n", pcm_c->name, pcm_i->name);
	printk(KERN_INFO "soc: rate mask 0x%x \nsoc: min ch %d max ch %d\nsoc: min rate %d max rate %d\n",
		runtime->hw.rates, runtime->hw.channels_min,
		runtime->hw.channels_max, runtime->hw.rate_min, runtime->hw.rate_max);

	if(pcm_i->ops.startup) {
		if((ret = pcm_i->ops.startup(substream)) < 0) {
			printk(KERN_ERR "soc: can't open interface %s\n", pcm_i->name);
			goto ops_err;
		}
	}

	if(platform->pcm_ops->open) {
		if((ret = platform->pcm_ops->open(substream)) < 0) {
			printk(KERN_ERR "soc: can't open platform %s\n", platform->name);
			goto platform_err;
		}
	}
	
	if(machine->ops->startup) {
		if((ret = machine->ops->startup(substream)) < 0) {
			printk(KERN_ERR "soc: %s startup failed\n", machine->name);
			goto machine_err;
		}
	}

	if(pcm_c->ops.startup) {
		if((ret = pcm_c->ops.startup(substream)) < 0) {
			printk(KERN_ERR "soc: can't open codec %s\n", pcm_c->name);
			goto pcm_err;
		}
	}

	return ret;

pcm_err:
	if(pcm_c->ops.shutdown)
		pcm_c->ops.shutdown(substream);
	
machine_err:
	if(machine->ops->shutdown)
		machine->ops->shutdown(substream);
	
platform_err:
	if(platform->pcm_ops->close)
		platform->pcm_ops->close(substream);

ops_err:
	if(pcm_i->ops.shutdown)
		pcm_i->ops.shutdown(substream);	
	
	return ret;
}

/* private data can be freed here */
static int soc_pcm_close(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;

	if(pcm_i->ops.shutdown)
		pcm_i->ops.shutdown(substream);
	
	if(pcm_c->ops.shutdown)
		pcm_c->ops.shutdown(substream);
	
	if(machine->ops->shutdown)
		machine->ops->shutdown(substream);
	
	if(platform->pcm_ops->close)
		platform->pcm_ops->close(substream);

	return 0;
}

/* called when pcm is prepared, can set format, sample rate; 
 * atomic, can be called multiple times, can refer to runtime info*/
static int soc_pcm_prepare(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	int ret = 0;

	if(platform->pcm_ops->prepare)
		if((ret = platform->pcm_ops->prepare(substream)) < 0)
			return ret;
	
	if(pcm_c->ops.prepare)
		if((ret = pcm_c->ops.prepare(substream)) < 0)
			return ret;
	
	if(pcm_i->ops.prepare)
		if((ret = pcm_i->ops.prepare(substream)) < 0)
			return ret;

	return ret;
}

static int soc_pcm_trigger(snd_pcm_substream_t *substream, int cmd)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	int ret = 0;

	if(pcm_c->ops.trigger)
		if((ret = pcm_c->ops.trigger(substream, cmd)) < 0)
			return ret;
	
	if(pcm_i->ops.trigger)
		if((ret = pcm_i->ops.trigger(substream, cmd)) < 0)
			return ret;
	
	if(platform->pcm_ops->trigger)
		if((ret = platform->pcm_ops->trigger(substream, cmd)) < 0)
			return ret;

	return ret;
}

/* called when hw params is set by app, can be called multiple times
 * alloc buffers (using snd_pcm_lib_* ), non-atomic, hw setup
 */
static int soc_pcm_hw_params(snd_pcm_substream_t *substream,
				snd_pcm_hw_params_t *params)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	struct snd_soc_codec *codec = (struct snd_soc_codec*)pcm_i->card->private_data;
	int ret = 0;

	/* prepare hw non AC97 only */
	if(pcm_i->type != SND_SOC_AC97) {
		if((ret = soc_check_hw_params(substream, params)) < 0)
			return ret;
	}
	
	if(pcm_c->ops.hw_params) {
		if((ret = pcm_c->ops.hw_params(substream, params)) < 0) {
			printk(KERN_ERR "soc: can't set codec %s hw params\n", pcm_c->name);
			goto codec_err;
		}
	}
	
	if(pcm_i->ops.hw_params) {
		if((ret = pcm_i->ops.hw_params(substream, params)) < 0) {
			printk(KERN_ERR "soc: can't set interface %s hw params\n", pcm_i->name);
			goto interface_err;
		}
	}
	
	if(platform->pcm_ops->hw_params) {
		if((ret = platform->pcm_ops->hw_params(substream, params)) < 0) {
			printk(KERN_ERR "soc: can't set platform %s hw params\n", platform->name);
			goto platform_err;
		}
	}
	
	if(codec->dpm_state == SNDRV_CTL_POWER_D0)
		return 0;
	
	if(codec->dpm_event)
		codec->dpm_event(codec, SNDRV_CTL_POWER_D1);
	
	snd_soc_dpm_codec_event(codec, SNDRV_CTL_POWER_D1);
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		snd_soc_dpm_stream_event(codec, pcm_c->playback.sname, SND_SOC_DPM_STREAM_START);
	else
		snd_soc_dpm_stream_event(codec, pcm_c->capture.sname, SND_SOC_DPM_STREAM_START);
		
	if(codec->dpm_event)
		codec->dpm_event(codec, SNDRV_CTL_POWER_D0);
	snd_soc_dpm_codec_mute(codec, 0);
	codec->dpm_state = SNDRV_CTL_POWER_D0;
	return ret;
	
platform_err:
	if(platform->pcm_ops->hw_free)
		platform->pcm_ops->hw_free(substream);
	
interface_err:
	if(pcm_i->ops.hw_free)
		pcm_i->ops.hw_free(substream);
	
codec_err:
	if(pcm_c->ops.hw_free)
		pcm_c->ops.hw_free(substream);

	return ret;
}

/* free buffers, can be called multiple times */
static int soc_pcm_hw_free(snd_pcm_substream_t *substream)
{
	struct snd_soc_pcm_codec *pcm_c = substream->private_data;
	struct snd_soc_pcm_interface *pcm_i = pcm_c->pcm_i;
	struct snd_soc_codec *codec = (struct snd_soc_codec*)pcm_i->card->private_data;

	if(pcm_c->ops.hw_free)
		pcm_c->ops.hw_free(substream);
	
	if(pcm_i->ops.hw_free)
		pcm_i->ops.hw_free(substream);
	
	if(platform->pcm_ops->hw_free)
		platform->pcm_ops->hw_free(substream);
	
	if(codec->dpm_state == SNDRV_CTL_POWER_D3hot)
		return 0;
	
	snd_soc_dpm_codec_mute(codec, 1);
	snd_soc_dpm_codec_event(codec, SNDRV_CTL_POWER_D3hot);
		
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		snd_soc_dpm_stream_event(codec, pcm_c->playback.sname, SND_SOC_DPM_STREAM_STOP);
	else
		snd_soc_dpm_stream_event(codec, pcm_c->capture.sname, SND_SOC_DPM_STREAM_STOP);
	
	if(codec->dpm_event)
		codec->dpm_event(codec, SNDRV_CTL_POWER_D3hot);
	
	codec->dpm_state = SNDRV_CTL_POWER_D3hot;
	return 0;
}

static snd_pcm_ops_t soc_pcm_ops = {
	.open		= soc_pcm_open,
	.close		= soc_pcm_close,
	.hw_params	= soc_pcm_hw_params,
	.hw_free	= soc_pcm_hw_free,
	.prepare	= soc_pcm_prepare,
	.trigger	= soc_pcm_trigger,
};

#define soc_suspend	NULL
#define soc_resume	NULL

static int soc_register_pcm(int nconfig)
{
	int i, reg = 1;

	/* register when all pcm are created for codec */
	for(i = 0; i < machine->nconfigs; i++) {
		
		if(strcmp(machine->config[i].name, machine->config[nconfig].name))
			continue;
		
		if(!machine->config[i].connect || machine->config[i].card_reg)
			reg = 0;
	}
	
	if (reg) {	
		snd_soc_pcm_register(machine, machine->config[nconfig].name);
		for(i = 0; i < machine->nconfigs; i++) {
			if(!strcmp(machine->config[i].name, machine->config[nconfig].name))
				machine->config[i].card_reg = 1;
		}
	}
	
	return reg;
}

static int soc_sync_objects(void)
{
	int i, ret = 0;
	
	/* only match codecs/interfaces when we have a machine and platform */
	if(machine == NULL || platform == NULL)
		return ret;

	/* check machine config */
	for(i = 0; i < machine->nconfigs; i++) {
		if(machine->config[i].connect)
			continue;
	
		if((ret = snd_soc_pcm_connect(machine, i)) == 0) {
			machine->config[i].connect = 1;
			soc_register_pcm(i);
		}
	}
	
	return 0;
}

static int soc_probe_codec(struct snd_soc_codec *codec)
{
	int ret = 0;
	
	/* register a sound card */
	codec->card = snd_card_new(SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1, codec->owner, 0);
	if (!codec->card) {
		printk(KERN_ERR "soc: can't create sound card for codec %s\n", codec->name);
		return -ENODEV;
	}

	platform_device_register(&codec->pdev);
	codec->card->dev = &codec->pdev.dev;
	codec->card->private_data = codec;
	strncpy(codec->card->driver, codec->name, sizeof(codec->card->driver));

	if(codec->probe && (ret = codec->probe(codec)) < 0)
		goto err;
	codec->probed = 1;
	codec->dpm_state = SNDRV_CTL_POWER_D3hot;
	soc_sync_objects();
	return 0;
	
err:
	snd_card_free(codec->card);
	platform_device_unregister(&codec->pdev);
	codec->card = NULL;
	return ret;
}

static void soc_free_dpm_widgets(struct snd_soc_codec *codec)
{
	struct snd_soc_dpm_widget *w, *lw = NULL;
	struct snd_soc_dpm_path *p, *lp = NULL;
	struct list_head *l;
	
	list_for_each(l, &codec->dpm_widgets)
	{
		w =  list_entry(l, struct snd_soc_dpm_widget, list);
		if (lw)
			kfree(lw);
		lw = w;
	}
	kfree(lw);
	
	list_for_each(l, &codec->dpm_paths)
	{
		p =  list_entry(l, struct snd_soc_dpm_path, list);
		if(p->long_name)
			kfree(p->long_name);
		if (lp)
			kfree(lp);
		lp = p;
	}
	kfree(lp);
}

static int soc_remove_codec(struct snd_soc_codec *codec)
{	
	if(!codec->probed)
		return 0;
	
	if (codec->card) {
		snd_soc_dpm_sys_remove(&codec->pdev.dev);
		snd_card_free(codec->card);
		platform_device_unregister(&codec->pdev);
		codec->card = NULL;
	}
	if(codec->remove)
		codec->remove(codec);

	/* free any DPM objects */
	soc_free_dpm_widgets(codec);
	codec->probed = 0;

	return 0;
}

int snd_soc_register_codec(struct snd_soc_codec *codec)
{
	int ret = 0;

	down(&soc_sem);
	INIT_LIST_HEAD(&codec->list);
	INIT_LIST_HEAD(&codec->dpm_widgets);
	INIT_LIST_HEAD(&codec->dpm_paths);
	list_add(&codec->list, &codecs);
	
	if((ret = soc_probe_codec(codec)) < 0) {
		printk(KERN_ERR "soc: probe of codec %s failed\n", codec->name);
		list_del_init(&codec->list);
		up(&soc_sem);
		return ret;
	}

	up(&soc_sem);
	return ret;
}

void snd_soc_unregister_codec(struct snd_soc_codec *codec)
{
	down(&soc_sem); 
	list_del_init(&codec->list);
	soc_remove_codec(codec);	
	up(&soc_sem);
}

int snd_soc_register_interface(struct snd_soc_pcm_interface *iface)
{	
	int ret = 0;
	
	down(&soc_sem);	
	if(iface->probe && (ret = iface->probe(iface)) < 0) {
		printk(KERN_ERR "soc: probe of interface %s failed\n", iface->name);
		up(&soc_sem);
		return ret;
	}		

	INIT_LIST_HEAD(&iface->list);
	list_add(&iface->list, &interfaces);
	soc_sync_objects();
	up(&soc_sem);
	return 0;
}

void snd_soc_unregister_interface(struct snd_soc_pcm_interface *iface)
{
	down(&soc_sem);
	list_del_init(&iface->list);
	if (iface->remove) 
		iface->remove(iface);
	up(&soc_sem);	
}

static struct device_driver soc_pdrv;

int snd_soc_register_platform(struct snd_soc_platform *p)
{
	int ret = 0;
	
	down(&soc_sem);
	if(platform) {
		up(&soc_sem);
		return -EBUSY;
	}

	platform = p;
	soc_pcm_ops.mmap = p->pcm_ops->mmap;
	soc_pcm_ops.pointer = p->pcm_ops->pointer;
	soc_pcm_ops.ioctl = p->pcm_ops->ioctl;
	soc_pdrv.bus = p->bus;
	soc_pdrv.name = p->name;
	soc_pdrv.probe = p->probe;
	soc_pdrv.remove = p->remove;

	if((ret = driver_register(&soc_pdrv)) < 0) {
		printk(KERN_ERR "soc: failed to register platform %s\n", p->name);
		up(&soc_sem);
		return ret;
	}
	
	soc_sync_objects();
	up(&soc_sem);
	return ret;
}

void snd_soc_unregister_platform(struct snd_soc_platform *p)
{
	struct list_head *l;
	struct snd_soc_codec *codec = NULL;
	
	down(&soc_sem);
	list_for_each(l, &codecs) {
		codec = list_entry(l, struct snd_soc_codec, list);
		if (codec->card) {
			snd_soc_dpm_sys_remove(&codec->pdev.dev);
			snd_card_free(codec->card);
			platform_device_unregister(&codec->pdev);
			codec->card = NULL;
		}
	}
	
	driver_unregister(&soc_pdrv);
	platform = NULL;
	up(&soc_sem);	
}

int snd_soc_register_machine(struct snd_soc_machine *m)
{
	int ret = 0;

	down(&soc_sem);
	if(m->probe) {
		if((ret = m->probe(m)) < 0) {
			up(&soc_sem);
			return ret;
		}
	}
	machine = m;
	soc_sync_objects();
	up(&soc_sem);
	return 0;
}

void snd_soc_unregister_machine(struct snd_soc_machine *m)
{
	down(&soc_sem);
	if(machine->remove)
		machine->remove(m);
	machine = NULL;
	up(&soc_sem);
}

/* find codec from pcm name */
static struct snd_soc_codec * soc_get_codec_from_pcm(char *pcm_name, 
	struct snd_soc_pcm_codec **pcm_c)
{
	struct list_head *l;
	struct snd_soc_codec *c = NULL;

	list_for_each(l, &codecs) {
		int i = 0;
		c = list_entry(l, struct snd_soc_codec, list);

		if(c->npcms == 0)
			continue;

		for(;i < c->npcms; i++) {
			if(!strcmp(pcm_name, c->pcms[i].name)) {
				if(pcm_c)
					*pcm_c = &c->pcms[i];
				return c;
			}
		}
	}
	
	return NULL;
}

static struct snd_soc_codec* soc_get_codec(char *name)
{
	struct list_head *l;
	struct snd_soc_codec *c = NULL;

	list_for_each(l, &codecs) {
		c = list_entry(l, struct snd_soc_codec, list);
			
		if(!strcmp(name, c->name))
			return c;
	}
	
	return NULL;
}

static struct snd_soc_pcm_interface* soc_get_interface(int iface_type, int num)
{
	struct list_head *l;
	struct snd_soc_pcm_interface *i = NULL;
	
	list_for_each(l, &interfaces) {
		i = list_entry(l, struct snd_soc_pcm_interface, list);
		if (i->type == iface_type && i->id == num){
			if (i->pcm_c)
				return NULL;
			else
				return i;
		}
	}
	return NULL;
}

int snd_soc_pcm_connect(struct snd_soc_machine *machine, int nconfig)
{
	snd_pcm_t *pcm;
	struct snd_soc_pcm_codec *pcm_c;
	int ret = 0;
	struct snd_soc_pcm_interface *iface;
	struct snd_soc_codec *codec;
	int iface_type = machine->config[nconfig].interface;
	int num = machine->config[nconfig].int_num;
	char *name = machine->config[nconfig].sname;
	char new_name[32];

	if((codec = soc_get_codec_from_pcm(name, &pcm_c)) == NULL)
		return -ENODEV;
	
	if((iface = soc_get_interface(iface_type, num)) == NULL)
		return -ENODEV;
	
	/* check client and interface hw capabilities */
	sprintf(new_name, "%s-%s-%d", pcm_c->name, iface_name(iface_type), num);
	if((ret = snd_pcm_new(codec->card, new_name, codec->pcm_devs++, pcm_c->nplayback, pcm_c->ncapture, &pcm)) < 0) {
		printk(KERN_ERR "soc: can't create pcm for codec %s\n", codec->name);
		goto err;
	}
	pcm_c->pcm_i = iface;
	pcm_c->codec = codec;
	iface->card = codec->card;
	iface->pcm_c = pcm_c;
	pcm->private_data = pcm_c;
		
	if (pcm_c->nplayback)
		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &soc_pcm_ops);

	if (pcm_c->ncapture)
		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &soc_pcm_ops);

	if((ret = platform->pcm_new(codec->card, pcm_c, pcm)) < 0) {
		printk(KERN_ERR "soc: platform pcm constructor failed\n");
		goto err;
	}
	pcm->private_free = platform->pcm_free;
	if(machine->config[nconfig].init)
		machine->config[nconfig].init(codec);
	
	printk(KERN_INFO "soc: %s <-> %s mapping ok\n", pcm_c->name, iface->name);
	return ret;
	
err:
	pcm_c->pcm_i = NULL;
	iface->card = NULL;
	iface->pcm_c = NULL;
	return ret;
}

#ifdef CONFIG_FIONA_PM_AUDIO
int snd_soc_pcm_fpow_set_mode(struct snd_soc_codec *codec, int fpow_mode)
{
	static int cur_fpow_mode = FPOW_MODE_OFF;
	int ret = -EINVAL;

	if ((cur_fpow_mode == FPOW_MODE_ON && fpow_mode == FPOW_MODE_SLEEP) ||
	    (cur_fpow_mode == FPOW_MODE_SLEEP && fpow_mode == FPOW_MODE_ON)) {
		cur_fpow_mode = fpow_mode;

		ret = 0;

	} else {
		if (codec != NULL) {
			int (*soc_pm_op)(struct snd_soc_codec *codec) = NULL;

			switch (fpow_mode) {
				
				case FPOW_MODE_ON :
				case FPOW_MODE_SLEEP :
					soc_pm_op = codec->resume;
					break;
	
				case FPOW_MODE_OFF :
					soc_pm_op = codec->suspend;
					break;
	
				default :
					break;
	
			}
	
			if (soc_pm_op) {
				ret = soc_pm_op(codec);

				cur_fpow_mode = fpow_mode;
			}
		}
	}

	return ret;
}
#endif

int snd_soc_pcm_register(struct snd_soc_machine *machine, char *name)
{
	struct snd_soc_codec* codec;
	int ret = 0;

	if((codec = soc_get_codec(name)) == NULL)
		return -ENODEV;

	snprintf(codec->card->shortname, sizeof(codec->card->shortname),
		 "%s", codec->name);
	snprintf(codec->card->longname, sizeof(codec->card->longname),
		 "%s (%s)", codec->longname, "SoC");
	
	snd_card_set_pm_callback(codec->card, soc_suspend, soc_resume, codec);
	
	if((ret = snd_card_register(codec->card)) < 0) {
		printk(KERN_ERR "soc: failed to register soundcard for codec %s\n", codec->name);
		return ret;
	}
	
	codec->pdev.dev.driver_data = codec;
	snd_soc_dpm_sys_add(&codec->pdev.dev);

	return ret;	
}

int snd_soc_set_runtime_hw(snd_pcm_substream_t *substream, const snd_pcm_hardware_t *hw)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	runtime->hw.info = hw->info;
	runtime->hw.formats = hw->formats;
	runtime->hw.period_bytes_min = hw->period_bytes_min;
	runtime->hw.period_bytes_max = hw->period_bytes_max;
	runtime->hw.periods_min = hw->periods_min;
	runtime->hw.periods_max = hw->periods_max;
	runtime->hw.buffer_bytes_max = hw->buffer_bytes_max;
	runtime->hw.fifo_size = hw->fifo_size;
	return 0;
}

snd_kcontrol_t *snd_soc_cnew(const snd_kcontrol_new_t *_template, 
	void *data, char *long_name)
{
	snd_kcontrol_new_t template;
	
	memcpy(&template, _template, sizeof(template));
	if(long_name)
		template.name = long_name;
	snd_runtime_check(!template.index, return NULL);
	//template.index = codec->num;
	return snd_ctl_new1(&template, data);
}

/*
 * Controls
 */

int snd_soc_info_enum_double(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t * uinfo)
{
	struct soc_enum *e = (struct soc_enum *)kcontrol->private_value;
	
	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = e->shift_l == e->shift_r ? 1 : 2;
	uinfo->value.enumerated.items = e->mask;
	
	if (uinfo->value.enumerated.item > e->mask - 1)
		uinfo->value.enumerated.item = e->mask - 1;
	strcpy(uinfo->value.enumerated.name, e->texts[uinfo->value.enumerated.item]);
	return 0;
}

int snd_soc_get_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	struct soc_enum *e = (struct soc_enum *)kcontrol->private_value;
	unsigned short val, bitmask;
	
	for (bitmask = 1; bitmask < e->mask; bitmask <<= 1)
		;
	val = snd_soc_read(codec, e->reg);
	ucontrol->value.enumerated.item[0] = (val >> e->shift_l) & (bitmask - 1);
	if (e->shift_l != e->shift_r)
		ucontrol->value.enumerated.item[1] = (val >> e->shift_r) & (bitmask - 1);

	return 0;
}

int snd_soc_put_enum_double(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	struct soc_enum *e = (struct soc_enum *)kcontrol->private_value;
	unsigned short val;
	unsigned short mask, bitmask;
	
	for (bitmask = 1; bitmask < e->mask; bitmask <<= 1)
		;
	if (ucontrol->value.enumerated.item[0] > e->mask - 1)
		return -EINVAL;
	val = ucontrol->value.enumerated.item[0] << e->shift_l;
	mask = (bitmask - 1) << e->shift_l;
	if (e->shift_l != e->shift_r) {
		if (ucontrol->value.enumerated.item[1] > e->mask - 1)
			return -EINVAL;
		val |= ucontrol->value.enumerated.item[1] << e->shift_r;
		mask |= (bitmask - 1) << e->shift_r;
	}
	
	return snd_soc_update_bits(codec, e->reg, mask, val);
}

/* volume and switch controls */
int snd_soc_info_volsw(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t * uinfo)
{
	int mask = (kcontrol->private_value >> 16) & 0xff;
	int shift = (kcontrol->private_value >> 8) & 0x0f;
	int rshift = (kcontrol->private_value >> 12) & 0x0f;

	uinfo->type = mask == 1 ? SNDRV_CTL_ELEM_TYPE_BOOLEAN : SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = shift == rshift ? 1 : 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = mask;
	return 0;
}

int snd_soc_get_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	int reg = kcontrol->private_value & 0xff;
	int shift = (kcontrol->private_value >> 8) & 0x0f;
	int rshift = (kcontrol->private_value >> 12) & 0x0f;
	int mask = (kcontrol->private_value >> 16) & 0xff;
	int invert = (kcontrol->private_value >> 24) & 0x01;

	ucontrol->value.integer.value[0] = (snd_soc_read(codec, reg) >> shift) & mask;
	if (shift != rshift)
		ucontrol->value.integer.value[1] = (snd_soc_read(codec, reg) >> rshift) & mask;
	if (invert) {
		ucontrol->value.integer.value[0] = mask - ucontrol->value.integer.value[0];
		if (shift != rshift)
			ucontrol->value.integer.value[1] = mask - ucontrol->value.integer.value[1];
	}

	return 0;
}

int snd_soc_put_volsw(snd_kcontrol_t * kcontrol, snd_ctl_elem_value_t * ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	int reg = kcontrol->private_value & 0xff;
	int shift = (kcontrol->private_value >> 8) & 0x0f;
	int rshift = (kcontrol->private_value >> 12) & 0x0f;
	int mask = (kcontrol->private_value >> 16) & 0xff;
	int invert = (kcontrol->private_value >> 24) & 0x01;
	int err;
	unsigned short val, val2, val_mask;
	
	val = (ucontrol->value.integer.value[0] & mask);
	if (invert)
		val = mask - val;
	val_mask = mask << shift;
	val = val << shift;
	if (shift != rshift) {
		val2 = (ucontrol->value.integer.value[1] & mask);
		if (invert)
			val2 = mask - val2;
		val_mask |= mask << rshift;
		val |= val2 << rshift;
	}
	err = snd_soc_update_bits(codec, reg, val_mask, val);
	return err;
}

static int __init snd_soc_init(void) 
{
	printk(KERN_INFO "soc: version %s liam.girdwood@wolfsonmicro.com\n", SND_SOC_VERSION);
	return 0;
}

static void __exit snd_soc_exit(void) 
{
}

module_init(snd_soc_init); 
module_exit(snd_soc_exit);

EXPORT_SYMBOL_GPL(snd_soc_register_codec);
EXPORT_SYMBOL_GPL(snd_soc_unregister_codec);
EXPORT_SYMBOL_GPL(snd_soc_register_platform);
EXPORT_SYMBOL_GPL(snd_soc_unregister_platform);
EXPORT_SYMBOL_GPL(snd_soc_pcm_connect);
EXPORT_SYMBOL_GPL(snd_soc_register_interface);
EXPORT_SYMBOL_GPL(snd_soc_unregister_interface);
EXPORT_SYMBOL_GPL(snd_soc_set_runtime_hw);
EXPORT_SYMBOL_GPL(snd_soc_cnew);
EXPORT_SYMBOL_GPL(snd_soc_info_enum_double);
EXPORT_SYMBOL_GPL(snd_soc_get_enum_double);
EXPORT_SYMBOL_GPL(snd_soc_put_enum_double);
EXPORT_SYMBOL_GPL(snd_soc_info_volsw);
EXPORT_SYMBOL_GPL(snd_soc_get_volsw);
EXPORT_SYMBOL_GPL(snd_soc_put_volsw);
EXPORT_SYMBOL_GPL(snd_soc_update_bits);
EXPORT_SYMBOL_GPL(snd_soc_test_bits);
EXPORT_SYMBOL_GPL(snd_soc_register_machine);
EXPORT_SYMBOL_GPL(snd_soc_unregister_machine);
EXPORT_SYMBOL_GPL(snd_soc_pcm_register);

/* Module information */ 
MODULE_AUTHOR("Liam Girdwood, liam.girdwood@wolfsonmicro.com, www.wolfsonmicro.com");
MODULE_DESCRIPTION("ALSA SoC Core");
MODULE_LICENSE("GPL");
