/*
 * (C) Copyright 2006, Lab126, Inc.
 * Jerry Katzung, katzung@lab126.com
 * Mike Puckett, mikepuck@amazon.com
 *
 * Definitions for Samsung OneNAND flash.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#ifndef _ONENAND_H_
#define _ONENAND_H_

#ifndef __COMMON_H_
#include <common.h>
#endif

// The discrete and composite Samsung OneNAND parts differ in how one
// accesses their corresponding (should they be present) registers and
// flash banks.  On discrete parts, one simply accesses the individual
// register base addresses.  But, on composite (or "stacked") parts,
// the register base is the same for each "bank," but one must set (or
// clear) the DFS (Device Flash Select) and DBS (Device BufferRAM Select)
// bits appropriately.
//
// The onenand_ddp_select enum tells onenand_ddp_op which operation to
// to perform.
//
enum onenand_ddp_select
{
    ddp_reg = 0,      // Just return register base address.
    ddp_fba,          // Set or clear XFS (DFS or FFS) bit for FBA.
    ddp_sba,          // Set or clear XFS (DFS or FFS) bit for SBA.
    ddp_dbs           // Set or clear DBS bit.
};
typedef enum onenand_ddp_select onenand_ddp_select;

enum onenand_type
{
    onenand_128MBx2,  // Set/clear DFS & DBS bits to switch banks, 0..1023 blocks each.
    onenand_256MBx1,  // One bank, 0..2047 blocks (no need to set/clear DFS & DBS bits).
    
    onenand_unknown
};
typedef enum onenand_type onenand_type;

// OneNAND common routines for cmd_nand.c and fiona/flash.c.
//
// In flash.c:
//
extern ulong onenand_ddp_op(ulong phys_base, onenand_ddp_select which_ddp);
extern void  onenand_set_fba(ulong phys_base, ushort fba);
extern void  onenand_set_sba(ulong phys_base, ushort sba);
extern void  onenand_set_fxa(ulong phys_base, ushort fpa, ushort fsa);
extern void  onenand_set_bsa(ulong phys_base, ushort bsa, ushort bsc);
extern void  onenand_set_sc1(ulong phys_base, ushort sc1);

extern void  onenand_display(int bank_num);
extern void  onenand_display_bad_blocks(int bank_num);

// In cmd_nand.c:
//
extern bool  onenand_block_is_bad(ulong io_base_addr, ulong block);
extern bool  onenand_block_is_truly_bad(ulong io_base_addr, ulong block);
extern int   onenand_wait_for_int(ulong io_base_addr, ushort int_type, bool check_status);

extern int   onenand_set_protection_state(ulong io_base_addr, ulong block, ushort state);
extern int   onenand_get_protection_state(ulong io_base_addr, ulong block);

extern int   onenand_load_spare(ulong io_base_addr, ulong block, ushort page);
extern int   onenand_write_spare(ulong io_base_addr, ulong block, ushort page);

extern int   onenand_load_page(ulong io_base_addr, ulong block, ushort page);
extern int   onenand_write_page(ulong io_base_addr, ulong block, ushort page);

extern int   onenand_erase_block(ulong io_base_addr, ulong block, bool clean);

// Misc. OneNAND support defines and macros.
//
#define SAMSUNG_MANUFACTURER_ID     0x00EC  /* Returned manufacturer_id value.    */

#define SAMSUNG_DEVICE_ID_128MBx2   0x004C  /* 256MB (2Gb) two-die OneNAND.       */
#define SAMSUNG_DEVICE_ID_256MBx1   0x0044  /* 256MB (2Gb) one-die OneNAND.       */

#define FLASH_ID_SAMSUNG            (SAMSUNG_MANUFACTURER_ID)

#define ONENAND_SAMSUNG(m)          (SAMSUNG_MANUFACTURER_ID == (m))

#define ONENAND_SAMSUNG_128MBx2(m, d)    \
                                    (ONENAND_SAMSUNG(m) && \
                                    (SAMSUNG_DEVICE_ID_128MBx2 == (d)))

#define ONENAND_SAMSUNG_256MBx1(m, d)    \
                                    (ONENAND_SAMSUNG(m) && \
                                    (SAMSUNG_DEVICE_ID_256MBx1 == (d)))

#define ONENAND_BADBLOCK_PAGES      2

#define ONENAND_ERASED_MARKER       ((ushort)0xFFFF)
#define ONENAND_FAKE_BAD_MARKER     ((ushort)0x8888)

#define ONENAND_BOOTRAM_OFFSET      0x00000000
#define ONENAND_DATARAM0_OFFSET     0x00000400
#define ONENAND_DATARAM1_OFFSET     0x00000C00
#define ONENAND_ECC_OFFSET          0x00010000
#define ONENAND_REGS_OFFSET         0x0001E000

#define ONENAND_BOOTRAM_BASE(phys_base)         ((ushort *) (onenand_ddp_op(phys_base, ddp_dbs) + ONENAND_BOOTRAM_OFFSET))
#define ONENAND_DATARAM0_BASE(phys_base)        ((ushort *) (onenand_ddp_op(phys_base, ddp_dbs) + ONENAND_DATARAM0_OFFSET))
#define ONENAND_DATARAM1_BASE(phys_base)        ((ushort *) (onenand_ddp_op(phys_base, ddp_dbs) + ONENAND_DATARAM1_OFFSET))
#define ONENAND_BOOTRAM_ECC_BASE(phys_base)     ((ushort *) (onenand_ddp_op(phys_base, ddp_dbs) + ONENAND_ECC_OFFSET))
#define ONENAND_DATARAM0_ECC_BASE(phys_base)    ((ushort *) (onenand_ddp_op(phys_base, ddp_dbs) + ONENAND_ECC_OFFSET + 0x20))
#define ONENAND_DATARAM1_ECC_BASE(phys_base)    ((ushort *) (onenand_ddp_op(phys_base, ddp_dbs) + ONENAND_ECC_OFFSET + 0x60))
#define ONENAND_REGS_BASE(phys_base)            (onenand_ddp_op(phys_base, ddp_reg))

#define ONENAND_REG_BASE_RAW(phys_base)         ((onenand_regs_t *) ((phys_base) + ONENAND_REGS_OFFSET))
#define ONENAND_REG_BASE(phys_base)             (ONENAND_REG_BASE_RAW(onenand_ddp_op(phys_base, ddp_reg)))

typedef struct {
    volatile ushort manufacturer_id;            /* 0x0000 */
    volatile ushort device_id;                  /* 0x0002 */
    volatile ushort version_id;                 /* 0x0004 */
    volatile ushort data_buffer_size;           /* 0x0006 */
    volatile ushort boot_buffer_size;           /* 0x0008 */
    volatile ushort buffer_count;               /* 0x000A */
    volatile ushort technology;                 /* 0x000C */
    volatile ushort rsrv0[249];
    volatile ushort start_addr_1;               /* 0x0200 */
    volatile ushort start_addr_2;               /* 0x0202 */
    volatile ushort start_addr_3;               /* 0x0204 */
    volatile ushort start_addr_4;               /* 0x0206 */
    volatile ushort rsrv_start_addr_5;
    volatile ushort rsrv_start_addr_6;
    volatile ushort rsrv_start_addr_7;
    volatile ushort start_addr_8;               /* 0x020E */
    volatile ushort rsrv1[248];
    volatile ushort start_buffer;               /* 0x0400 */
    volatile ushort rsrv2[31];
    volatile ushort command;                    /* 0x0440 */
    volatile ushort sys_config_1;               /* 0x0442 */
    volatile ushort sys_config_2;               /* 0x0444 */
    volatile ushort rsrv3[29];
    volatile ushort controller_status;          /* 0x0480 */
    volatile ushort interrupt;                  /* 0x0482 */
    volatile ushort rsrv4[10];
    volatile ushort start_block_addr;           /* 0x0498 */
    volatile ushort rsrv5;
    volatile ushort write_protection_status;    /* 0x049C */
    volatile ushort rsrv6[1201];
    volatile ushort ecc_status;                 /* 0x0E00 */
    volatile ushort ecc_main_1st_sector;        /* 0x0E02 */
    volatile ushort ecc_spare_1st_sector;       /* 0x0E04 */
    volatile ushort ecc_main_2nd_sector;        /* 0x0E06 */
    volatile ushort ecc_spare_2nd_sector;       /* 0x0E08 */
    volatile ushort ecc_main_3rd_sector;        /* 0x0E0A */
    volatile ushort ecc_spare_3rd_sector;       /* 0x0E0C */
    volatile ushort ecc_main_4th_sector;        /* 0x0E0E */
    volatile ushort ecc_spare_4th_sector;       /* 0x0E10 */
    volatile ushort rsrv7[247];
} onenand_regs_t;

typedef struct {
    volatile ushort bi;              /* [0..1] Bad Block Info                 */
    volatile ushort ecc_internal_1;  /* [2..3] Managed by Internal ECC logic  */
    volatile uchar  ecc_internal_2;  /* [4]    Managed by Internal ECC logic  */
    volatile uchar  reserved_1;      /* [5]    Reserved for future use        */
    volatile uchar  reserved_2;      /* [6..7] Reserved for future use        */
    volatile uchar  ecc_main_1;      /* [8]    ECC code for Main  area data 1 */
    volatile uchar  ecc_main_2;      /* [9]    ECC code for Main  area data 2 */
    volatile uchar  ecc_spare_3;     /* [A]    ECC code for Spare area data 3 */
    volatile uchar  ecc_spare_1;     /* [B]    ECC code for Spare area data 1 */
    volatile uchar  ecc_spare_2;     /* [C]    ECC code for Spare area data 2 */
    volatile uchar  reserved_3;      /* [D]    Reserved for future use        */
    volatile ushort free;            /* [E..F] Free Usage                     */
} onenand_spare_t;

typedef struct {
    onenand_spare_t boot_s_0,        /* [00..0F] ONENAND_BOOTRAM_ECC_BASE     */
                    boot_s_1,        /* [10..1F]                              */
                    
                    data_s_0_0,      /* [20..2F] ONENAND_DATARAM0_ECC_BASE    */
                    data_s_0_1,      /* [30..3F]                              */
                    data_s_0_2,      /* [40..4F]                              */
                    data_s_0_3,      /* [50..5F]                              */
                    
                    data_s_1_0,      /* [60..6F] ONENAND_DATARAM1_ECC_BASE    */
                    data_s_1_1,      /* [70..7F]                              */
                    data_s_1_2,      /* [80..8F]                              */
                    data_s_1_3;      /* [90. 9F]                              */
} onenand_spare_area_t;

/* Start Address1 register                                     base+0x1E200   */
#define ONENAND_DFS                     0x8000     /* 0=1GB 1024, 1=2GB 2048  */
#define ONENAND_FFS                     0x0400     /* 0=1GB 1024, 1=2GB 2048  */

/* Start Address2 register                                     base+0x1E202   */
#define ONENAND_DBS                     0x8000     /* 0=1GB 1024, 1=2GB 2048  */

// For the stacked parts, we need to preserve the setting of the DFS bit when
// the FBA part of start_addr_1 is set or cleared.  We also need to set
// the DFS bit when we're setting start_block_addr.
//
#define ONENAND_SET_FBA(phys_base, fba) (onenand_set_fba(phys_base, fba)) 
#define ONENAND_SET_SBA(phys_base, sba) (onenand_set_sba(phys_base, sba))

#define ONENAND_USE_FBA                 true
#define ONENAND_USE_SBA                 false

// We also need to clear or set the DBS bit of start_addr_2 on stacked parts
// when requested.
//
#define ONENAND_SET_DBS(phys_base)      (onenand_ddp_op(phys_base, ddp_dbs))

// On the Samsung OneNAND stacked parts, access to the registers is always at
// the same physical location.  However, there are still multiple banks that
// we must switch between in order to access the part's entire capacity.  To
// help us know which bank we're being asked to address (and to provide some
// compatibility with earlier, non-stacked parts), we define a structure that
// always lists which address is to be used access the registers, reg_base,
// and which area in the part we're being asked to address, phys_base.  On
// non-stacked parts, both addresses will always be the same.  On stacked parts,
// they will both be the same for the lower bank; but, for the higher bank,
// they will differ.  We use the fact that they differ in the higher bank to
// know whether we're supposed to set (higher) or clear (lower) the DFS and
// DBS bits when asked.
//
// Yet another complication:  We now have "monolithic" parts that have the
// the physical capacity of the stacked parts, but are addressed like the
// the non-stacked parts.  Because our underlying code currently assumes
// flash parts are banked and that each bank is 128MBs each, we still treat
// the monolithic parts as though they were "banked."
//
typedef struct {
        onenand_type part_type;
        ulong        regs_base,
                     phys_base;
}
onenand_info_t;

/* Start Address8 register (Flash Page/Sector Address)         base+0x1E20E   */
#define ONENAND_FPA_SHIFT               2    /* FPA is the upper six bits.    */

#define ONENAND_FBA_ZERO                0    /* Start at block zero.          */
#define ONENAND_FPA_ZERO                0    /* Start at page zero.           */
#define ONENAND_FSA_ZERO                0    /* Start at sector zero.         */

#define ONENAND_SET_FXA(b, p, s)        (onenand_set_fxa(b, p, s))

/* Start Buffer register                                       base+0x1E400   */
#define ONENAND_STBUF_BSA_BOOTRAM0      0x0
#define ONENAND_STBUF_BSA_BOOTRAM1      0x1
#define ONENAND_STBUF_BSA_DATARAM0_0    0x8
#define ONENAND_STBUF_BSA_DATARAM0_1    0x9
#define ONENAND_STBUF_BSA_DATARAM0_2    0xA
#define ONENAND_STBUF_BSA_DATARAM0_3    0xB
#define ONENAND_STBUF_BSA_DATARAM1_0    0xC
#define ONENAND_STBUF_BSA_DATARAM1_1    0xD
#define ONENAND_STBUF_BSA_DATARAM1_2    0xE
#define ONENAND_STBUF_BSA_DATARAM1_3    0xF
#define ONENAND_STBUF_BSA_SHIFT         8      /* BSA is upper 8 bits.        */

#define ONENAND_STBUF_BSC_1_SECTOR      0x1
#define ONENAND_STBUF_BSC_2_SECTORS     0x2
#define ONENAND_STBUF_BSC_3_SECTORS     0x3
#define ONENAND_STBUF_BSC_4_SECTORS     0x0

#define ONENAND_SET_BSA(b, s, c)        (onenand_set_bsa(b, s, c))

/* Command register                                            base+0x1E440   */
#define ONENAND_COMMAND_LOAD            0x0000
#define ONENAND_COMMAND_LOAD_SPARE      0x0013
#define ONENAND_COMMAND_PROGRAM         0x0080
#define ONENAND_COMMAND_PROGRAM_SPARE   0x001A
#define ONENAND_COMMAND_PROGRAM_COPY    0x001B
#define ONENAND_COMMAND_UNLOCK          0x0023
#define ONENAND_COMMAND_LOCK            0x002A
#define ONENAND_COMMAND_LOCK_TIGHT      0x002C
#define ONENAND_COMMAND_ERASE_VERIFY    0x0071
#define ONENAND_COMMAND_ERASE           0x0094
#define ONENAND_COMMAND_ERASE_BLOCKS    0x0095
#define ONENAND_COMMAND_ERASE_SUSPEND   0x00B0
#define ONENAND_COMMAND_ERASE_RESUME    0x0030
#define ONENAND_COMMAND_RESET_FLASH     0x00F0
#define ONENAND_COMMAND_RESET_NAND      0x00F3
#define ONENAND_COMMAND_OTP_ACCESS      0x0065

#define ONENAND_MAX_ERASE_ATTEMPTS      3      /* On erase failure, retry.    */
#define ONENAND_NUM_OTP_USER_PAGES      10     /* Pages 0..9 in OTP area.     */

#define ONENAND_ERASE_JFFS2             true   /* Erase "clean" to JFFS2.     */
#define ONENAND_ERASE_RAW               false  /* Just do the erase and stop. */

/* System Configuration 1 register                             base+0x1E442   */
#define ONENAND_RM_SYNC                0x8000  /* Bit 15                      */
#define ONENAND_RM_ASYNC               0x0000 
#define ONENAND_BRL_08_CYCLES          0x0000  /* Bits 12-14                  */
#define ONENAND_BRL_09_CYCLES          0x1000
#define ONENAND_BRL_10_CYCLES          0x2000
#define ONENAND_BRL_03_CYCLES          0x3000
#define ONENAND_BRL_04_CYCLES          0x4000
#define ONENAND_BRL_05_CYCLES          0x5000
#define ONENAND_BRL_06_CYCLES          0x6000
#define ONENAND_BRL_07_CYCLES          0x7000
#define ONENAND_BL_CONTINUOUS          0x0000  /* Bits 9-11                   */
#define ONENAND_BL_04_WORDS            0x0200
#define ONENAND_BL_08_WORDS            0x0400
#define ONENAND_BL_16_WORDS            0x0600
#define ONENAND_BL_32_WORDS            0x0800
#define ONENAND_BL_1K_WORDS            0x0A00
#define ONENAND_ECC_ON                 0x0000 /* Bit 8                        */
#define ONENAND_ECC_OFF                0x0100
#define ONENAND_RDY_HIGH               0x0080 /* Bit 7                        */
#define ONENAND_RDY_LOW                0x0000
#define ONENAND_INT_HIGH               0x0040 /* Bit 6                        */
#define ONENAND_INT_LOW                0x0000
#define ONENAND_IOBE_ENABLED           0x0020 /* Bit 5                        */
#define ONENAND_IOBE_DISABLED          0x0000
#define ONENAND_RDY_ACTIVE_VALID       0x0010 /* Bit 4                        */
#define ONENAND_RDY_ACTIVE_EARLY       0x0000

/* Controller Status register                                  base+0x1E480   */
#define ONENAND_STATUS_ON_GO            0x8000 /* 0=ready      1=busy         */
#define ONENAND_STATUS_LOCKED           0x4000 /* 0=unlocked   1=locked       */
#define ONENAND_STATUS_LOAD             0x2000 /* 0=ready      1=busy | error */
#define ONENAND_STATUS_PROG             0x1000 /* 0=ready      1=busy | error */
#define ONENAND_STATUS_ERASE            0x0800 /* 0=ready      1=busy | error */
#define ONENAND_STATUS_ERROR            0x0400 /* 0=pass       1=fail         */
#define ONENAND_STATUS_SUSPEND          0x0200 /* 0=resume     1=erase suspend*/
#define ONENAND_STATUS_RSTB             0x0080 /* 0=ready      1=busy         */
#define ONENAND_STATUS_OTPL             0x0040 /* 0=unlocked   1=locked       */
#define ONENAND_STATUS_TO               0x0001 /* 0=no timeout 1=timeout      */

#define ONENAND_OTPL(status)            (ONENAND_STATUS_OTPL  & status)
#define ONENAND_ERROR(status)           (ONENAND_STATUS_ERROR & status)

#define ONENAND_CHECK_STATUS            true
#define ONENAND_NO_STATUS               false

/* Interrupt Status register                                   base+0x1E482   */
#define ONENAND_CLEAR_ISR               0x0000 /* Use to start int. polling.  */

#define ONENAND_INT                     0x8000 /* 0=pending    1=complete     */
#define ONENAND_RI                      0x0080 /* 0=pending    1=complete     */
#define ONENAND_WI                      0x0040 /* 0=pending    1=complete     */
#define ONENAND_EI                      0x0020 /* 0=pending    1=complete     */
#define ONENAND_RSTI                    0x0010 /* 0=pending    1=complete     */

/* Write Protection Status register                            base+0x1E49C   */
#define ONENAND_WP_UNLOCKED             4
#define ONENAND_WP_LOCKED               2
#define ONENAND_WP_LOCKED_TIGHT         1

#define ONENAND_LOCKED                  1      /* locked or locked-tight      */
#define ONENAND_UNLOCKED                0      /* unlocked                    */

#endif /* _ONENAND_H_ */
